/*****************************************************************************
 *
 * This software module was originally developed by
 *
 *   Noel O'Connor (Teltec DCU / ACTS-MoMuSyS)
 *
 * and edited by
 * 
 *   Robert Danielsen (Telenor / ACTS-MoMuSyS)
 *   Michael Frater (UNSW0
 *   Jong Deuk Kim (HYUNDAI)
 *   Cheol Soo Park (HYUNDAI)
 *   Osamu Sunohara (SONY)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original
 * developer of this software module and his/her company, the subsequent
 * editors and their companies, and ISO/IEC have no liability for use of this
 * software module or modifications thereof in an implementation. Copyright is
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming
 * products.
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own
 * purpose, assign or donate the code to a third party and to inhibit third
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works.
 *
 * Copyright (c) 1996
 *
 *****************************************************************************/


/***********************************************************HeaderBegin*******
 *                                                                         
 * File:	vm_vop_bound.c
 *
 * Author:	Noel O'Connor Teltec Irl.
 * Created:	3-04-96
 *                                                                         
 * Description: 
 *
 * Notes: 	
 *
 * Modified:	21.04.96 Robert Danielsen: Reformatted. New headers.
 * 		24.10.96 Noel O'Connor: Added SARP for bounding box
 *											extraction
 *              03.01.97 Michael Frater: support for non-8-bit video
 *		07.05.98 Jong Deuk Kim (jdkim97@hei.co.kr, HYUNDAI) 
 *		         Cheol Soo Park (cspark@super5.hyundai.co.kr, HYUNDAI): interlaced tools
 *      	10.15.99 Dae-Sung Cho (Samsung AIT): added support for OBSS
 *              01.04.01 Osamu Sunohara (SONY): Functions for Studio Profile
 *
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/

#include <stdio.h>
#include "momusys.h"
#include "mom_structs.h"
#include "mom_access.h"
#include "mom_vop.h"
#include "vm_common_defs.h"
#include "vm_vop_bound.h"

/* begin: added for OBSS by Samsung AIT (1999-10-15) */
#define PARTIAL_X_MIN	110
#define PARTIAL_X_MAX	240
#define PARTIAL_Y_MIN	42
#define PARTIAL_Y_MAX	176
/* end: added for OBSS by Samsung AIT (1999-10-15) */

/***********************************************************CommentBegin******
 *
 * -- GetVopBounded -- Calculates and extracts the bounding box of this vop
 *
 * Author :		
 *	Noel O'Connor Teltec Irl.
 *
 * Created :		
 *	11-03-96
 *
 * Purpose :		
 *	This function takes as input the vop read from disk and
 * 	calculates and extracts the bounding box of this vop (based
 *	on the vop's alpha plane).
 * 
 * Arguments in : 	
 *	Vop *orig - pointer to original vop
 *	(i.e. vop read from disk),
 *	Int to_nearest_mb  - flag  	( 0 - gives the actual bounded vop )
 *					( 1 - gives the bounded vop extended
 *						to be of size n x 16 )
 *
 * Arguments in/out :	
 *	-
 *
 * Arguments out :	
 *	-
 *
 * Return values :	
 *	Vop *vop_bb - pointer to the bounded vop
 *
 * Side effects :	
 *	Memory is allocated within this function for
 *	the bounded vop
 *
 * Description :	
 *	This function first calculates the vop bounding box co-ordinates
 *	and dimensions. A vop is allocated with these dimensions. The 
 *	required image data is then extracted into this new bounded vop.
 *	Also the alpha plane charateristics (rectangular/arbitrary shape,
 *	binary/grey-level) are determined.
 *
 * See also :
 *	-
 *
 * Modified :		
 *	21.01.97 Added fix for bug VM4-970110-03 by Noel O'Connor
 *      07.05.98 : Jong Deuk Kim (HYUNDAI): modified for field-based chrominance subsampling
 *      10.15.99 Dae-Sung Cho (Samsung AIT): added support for OBSS
 *
 ***********************************************************CommentEnd********/
Vop *
GetVopBounded( Vop *orig,
               Int to_nearest_mb,
               Int even_size,
               Int *vop_has_content,
               VolConfig *vol_config /* HYUNDAI 980507 */)
{
  Vop *vop_bb;
  
  Image *alpha_image,
    *y_image,
    *u_image,
    *v_image;
  
  SInt *alpha,
    *lum,
    *u,
    *v,
    *g;
	
  Int x,
    y,
    hor_spat_ref,
    ver_spat_ref,
    i,
    y_ind,
    x_ind,
    width,
    height,
    multiple_of_mb_size,
    inside_alpha_image,
    mod_dim,
    lines,
    pels,
    aux;	
				
  Int vop_coded; /* MW 30-NOV-1998 */
	
  /* Use pointers to vop image data */			
  alpha_image = GetVopA(orig);
  y_image = GetVopY(orig);
  u_image = GetVopU(orig);
  v_image = GetVopV(orig);
	
  /* Use dimensions of alpha plane read from disk */
  x = GetImageSizeX(alpha_image);
  y = GetImageSizeY(alpha_image);
	
  /* Use pointers to YUVa */
  alpha = (SInt *) GetImageData(alpha_image);
  lum = (SInt *) GetImageData(y_image);
  u = (SInt *) GetImageData(u_image);
  v = (SInt *) GetImageData(v_image);
	
			
  /* 
   *	Call GetBoundingRect() to get the tightest rectangle
   *		around Vop with the top left co-ordinates modified
   *		to be even integers 
   */
  
  /* Get _tightest_ rectangle */
  multiple_of_mb_size = FALSE;
  
  /* Don't limit to inside alpha image 			*/
  /* (actually this is only necessary with 	*/
  /* multiple_of_mb_size set to TRUE) 			*/
  inside_alpha_image = FALSE;		
  
  /* Calc. the bounding box information */
  /* begin: added for OBSS by Samsung AIT (1999-10-15) */
  if( GetVolConfigScalability(vol_config)==1 && GetVolConfigScalType(vol_config)==1 &&
      GetVolConfigEnhanceType(vol_config)==1 && GetVolConfigUseRefShape(vol_config)==0 &&
      GetVolConfigShape(vol_config)==1 ) {
    hor_spat_ref = PARTIAL_X_MIN;
    ver_spat_ref = PARTIAL_Y_MIN;
    hor_spat_ref -= hor_spat_ref%2;
    ver_spat_ref -= ver_spat_ref%2;
    
    pels = PARTIAL_X_MAX - hor_spat_ref + 1;
    lines = PARTIAL_Y_MAX - ver_spat_ref + 1;
  } else 
    /* end: added for OBSS by Samsung AIT (1999-10-15) */
    GetBoundingRect(alpha,x,y,multiple_of_mb_size,
                    inside_alpha_image,&hor_spat_ref,&ver_spat_ref,&lines,&pels);
  
#ifdef _RANDOM_VOP_CODED_ /* MW 30-NOV-1998 */
  if(rand() <= (Int)(RAND_MAX/2.0))
  {
    pels  = 0;
    lines = 0;
  }
#endif
  
  /* Check for vop content */
  if((pels != 0) && (lines != 0))
  { /* MW 30-NOV-1998 */
    vop_coded = 1;
    *vop_has_content = TRUE;
  }
  else
  { /* MW 30-NOV-1998 */
    vop_coded = 0;
    *vop_has_content = FALSE;
  }
  
  if(*vop_has_content)
  {
    /* begin: added for OBSS by Samsung AIT (1999-10-15) */
    if( GetVolConfigScalability(vol_config)!=1 || GetVolConfigScalType(vol_config)!=1 ||
        GetVolConfigEnhanceType(vol_config)!=1 || GetVolConfigUseRefShape(vol_config)!=0 ||
        GetVolConfigShape(vol_config)!=1 ) 
    {
      /* end: added for OBSS by Samsung AIT (1999-10-15) */
      /* Do the Shape Adaptive Region Partitioning if necessary */
      if((hor_spat_ref != 0) || (ver_spat_ref != 0))
        DoSARP(alpha,x,&hor_spat_ref,&ver_spat_ref,&pels,&lines,
               vol_config->interlaced /* HYUNDAI 980507 */);
      /* begin: added for OBSS by Samsung AIT (1999-10-15) */
    }
    /* end: added for OBSS by Samsung AIT (1999-10-15) */
    
    /* Modify the bounding box width and height to be multiples
       of MB_SIZE */
    if((mod_dim = (pels%MB_SIZE)) != 0)
      pels += (MB_SIZE - mod_dim);
    if((mod_dim = (lines%MB_SIZE)) != 0)
      lines += (MB_SIZE - mod_dim);
    
    /* Allocate the bounding box */
    vop_bb = AllocVop(pels,lines,GetVopAuxCompCount(orig));
  }
  else   
    vop_bb = SallocVop();
  
  /* Extract image data into this bounded vop */
  width = pels; height = lines;	
  
  /* Copy data from original vop */
  CopyVopNonImageField(orig,vop_bb);
  
  PutVopHorSpatRef(hor_spat_ref,vop_bb);	/* Spatial ref. of justified 	*/
  PutVopVerSpatRef(ver_spat_ref,vop_bb);	/* vop.	*/
  PutVopWidth(width,vop_bb);		/* Store width and height */
  PutVopHeight(height,vop_bb);		/* as integer multiples of 16 */
  PutVopQuantPrecision(GetVopQuantPrecision(orig), vop_bb);
  PutVopBitsPerPixel(GetVopBitsPerPixel(orig), vop_bb);
  
  /* MW 30-NOV-1998 */
  PutVopCoded(vop_coded,vop_bb);
  
  /* Extract Y and A image data  */
  
  if (*vop_has_content)
  {
    i = 0;
    for(y_ind = ver_spat_ref; y_ind < ver_spat_ref + lines; y_ind++)
      for(x_ind = hor_spat_ref; x_ind < hor_spat_ref + pels; x_ind++)
      {
        if (ValidCoordinate(x_ind,y_ind,x,y)) 
        {
          vop_bb->a_chan->f[i] = alpha[x_ind + (y_ind * x)];
          vop_bb->y_chan->f[i] = lum[x_ind + (y_ind * x)];
        }
        
        i++;
      }
    
    /* Extract U and V image data */
    i = 0;
    for(y_ind = ver_spat_ref/2; y_ind < (ver_spat_ref/2) + lines/2; y_ind++)
      for(x_ind = hor_spat_ref/2; x_ind < (hor_spat_ref/2) + pels/2; x_ind++)
      {
        if (ValidCoordinate(x_ind,y_ind,x/2,y/2)) 
        {
          vop_bb->u_chan->f[i] = u[x_ind + (y_ind * (x/2))];
          vop_bb->v_chan->f[i] = v[x_ind + (y_ind * (x/2))];
        }
        i++;
      }

    /* Extract Aux image data - MAC (SB) 23-Nov-99 */
    for(aux=0;aux<GetVopAuxCompCount(orig);aux++) { 
      g = (SInt *) GetImageData(GetVopG(aux,orig));
      i = 0;
      for(y_ind = ver_spat_ref; y_ind < ver_spat_ref + lines; y_ind++) {
        for(x_ind = hor_spat_ref; x_ind < hor_spat_ref + pels; x_ind++) {
          if (ValidCoordinate(x_ind,y_ind,x,y)) {
            vop_bb->g_chan[aux]->f[i] = g[x_ind + (y_ind * x)];
          }
          i++;
        }
      }
    }
    /* End - MAC (SB) 23-Nov-99 */
  }
  else PutVopArbitraryShape(ARB_SHAPE,vop_bb);
  
  return(vop_bb);
}

/***********************************************************CommentBegin******
 *
 * -- GetBoundingRect -- Obtains the position of the bounding rectangle
 *
 * Author :		
 *	Noel O'Connor Teltec Irl.
 *
 * Created :		
 *	11-03-96
 *
 * Purpose :		
 *	Obtains the position of the bounding rectangle of the shape of a vop.
 * 	Size can includes bottom-left extension to make size MULTIPLE of 16,
 *	and perhaps INSIDE the vop.
 * 
 * Arguments in : 	
 *	SInt	*alpha - pointer to alpha image
 *	Int		vop_width - width of input alpha image
 *	Int		vop_height - height of input alpha image
 *	Int		multiple - when this flag is set the function
 *				returns a BB which is a multiple of MB_SIZE
 *				Otherwise it returns the _exact_ BB.
 *	Int		inside - bounding rectangle inside the alpha plane
 *
 * Arguments in/out :	
 *	-
 *
 * Arguments out :	
 *	Int		*br_x0 - BB horizontal coordinate
 *	Int		*br_y0 - BB vertical coordinate
 *	Int		*br_lines - row size of BB
 *	Int		*br_pels - column size of BB
 *
 * Return values :	
 *	-
 *
 * Side effects :	
 *	-
 *
 * Description :	
 *	-
 *
 * See also :
 *	-
 *
 * Modified :		
 *	
 *
 ***********************************************************CommentEnd********/

Void
GetBoundingRect(
     SInt       *alpha,      /* alpha plane data   */
     Int         vop_width,   /* alpha plane width  */
     Int         vop_height,  /* alpha plane height */
     Int         multiple,    /* 1: dimensions multiple of 16 */
     Int         inside,      /* bounding rectangle (BB) inside 
     				 the alpha plane */
     Int         *br_x0,  	  /* BB horizontal coordinate */
     Int         *br_y0,      /* BB vertical coordinate */
     Int         *br_lines,   /* rows of the BB */
     Int         *br_pels     /* columns of the BB */
     )   
{
#define BIGVALUE 10000
  Int i,j,
    x_min=  BIGVALUE,
    x_max=    -1,
    y_min=  BIGVALUE,
    y_max=    -1;
  
  Int base,kk;


  for(j=0;j<vop_height;j++)
    {
      base=j*vop_width;

      for(i=0;i<vop_width;i++)
	{
	  if(alpha[base+i]!=0)
	    {
	      x_min=(i<x_min)?i:x_min;
	      x_max=(i>x_max)?i:x_max;
	      y_min=(j<y_min)?j:y_min;
	      y_max=(j>y_max)?j:y_max;
	    }
	}
    }

	x_min -= x_min%2;
	y_min -= y_min%2;
	
  if (!((x_min==BIGVALUE)||(x_max==-1)||(y_min==BIGVALUE)||(y_max==-1)))
    {
      *br_x0 = x_min;
      *br_y0 = y_min;
      *br_pels = x_max - x_min +1 ;
      *br_lines  = y_max - y_min +1 ;
     
      /* multiples of 16 */
      if(multiple==1) 
	{
	  if((kk=(*br_lines)%MB_SIZE)!=0)
	    *br_lines+=(MB_SIZE-kk);
	  if((kk=(*br_pels)%MB_SIZE)!=0)
	    *br_pels+=(MB_SIZE-kk);
	}
     
      /* check the bottom side */
      if (inside==1)
	{
	  if ((*br_x0+*br_pels)>vop_width) 
	    {
	      /* fprintf(stderr,"\nGetBoundingRect: fixing right bound!\n\n"); */
	      *br_x0= vop_width - *br_pels ;
	    }
	  /* check the right side */
	  if ((*br_y0+*br_lines)>vop_height)  
	    { 
	      /* fprintf(stderr,"\nGetBoundingRect: fixing bottom bound!\n\n"); */
	      *br_y0= vop_height - *br_lines ;
	    } 
	}
      /* check if position of upper-left corner is inside the VOP */
      if( (*br_x0 <0) || (*br_y0 <0) )
	{
          fprintf(stderr,
		  "\nGetBoundingRect: UNABLE TO FIX POSITION-DIMENSIONS!\n\n");
          fprintf(stderr,
		  "\n                 SETTING br_x0=br_y0=br_pels=br_lines=0 !\n\n");
          *br_x0 = *br_y0 = *br_pels = *br_lines = 0;
	}

    }
  else
    {
      *br_x0 = *br_y0 = *br_pels = *br_lines  = 0;
      fprintf(stderr,"GetBoundingRect: alpha plane without any object!!\n");
    }

 
}/*GetBoundingRect*/


/***********************************************************CommentBegin******
 *
 * -- DoSARP -- Intelligent VOP formation as per VM 4.0 (Shape Adaptive Region
 *							Partitioning)
 *
 * Author :		
 *	Noel O'Connor Teltec Irl.
 *
 * Created :		
 *	11-03-96
 *
 * Purpose :		
 * 	Having found the tightest rectangle for the VOP (with top left coordinates 
 *	suitably adjusted) it is necessary to find the optimal top left co-ordinates
 *	for the bounding box which ensure that the minimum number of macroblocks are 
 *	coded.
 * 
 * Arguments in : 	
 *	SInt 	*alpha - pointer to alpha pixel data (unbounded)
 *	
 * Arguments in/out :	
 *	Int 	*hor_spat_ref - initially x co-ord. of tightest rect. (suitably adjusted)
 *	Int		*ver_spat_ref - initially y co-ord. of tightest rect. (suitably adjusted)
 *	Int		*vop_width - initially width of tightest rectangle
 *	Int		*vop_height - initially height of tightest rectangle
 *
 * Arguments out :	
 *	-
 *
 * Return values :	
 *	none
 *
 * Side effects :	
 *	-
 *
 * Description :	
 *	This code is based on the source code distributed by Gwang Hoon Park of Hyundai
 *	(11/10/96)
 *
 * See also :
 *	-
 *
 * Modified :		
 *   07.05.98 Jong Deuk Kim, Cheol Soo Park (HYUNDAI): 
 *            change the value of ver_spat_ref to multiple of four at interlaced
 *	
 *
 ***********************************************************CommentEnd********/
Void
DoSARP(SInt *alpha,             /* Pointer to alpha plane image dat */
        Int dim_x,              /* X dimension of alpha plane on disk */
        Int *hor_spat_ref,      /* current hor. spatial reference */
        Int *ver_spat_ref,      /* current ver. spatial referance */
        Int *vop_width,         /* current bb width */
        Int *vop_height,        /* current bb height */
        Int Interlaced          /* HYUNDAI 980507 */)
{
	Int		N = MB_SIZE,
				search_step = 2,
				min_block_number = 9999,
				block_number,
				GS_x, GS_y,
				SARP_grid_x=0, SARP_grid_y=0,
				move_SARP_grid_x, move_SARP_grid_y,
				start_x, start_y,
				x,y,xx,yy,
				check;

	/* Get starting co-ordinates for loop */ 
	start_x = (*hor_spat_ref>=0 && *hor_spat_ref<N) ? -(*hor_spat_ref) : -(N-2);
	start_y = (*ver_spat_ref>=0 && *ver_spat_ref<N) ? -(*ver_spat_ref) : -(N-2);
	
	/* Offset for location of tightest rectangle in alpha image */
	start_x += (*hor_spat_ref);
	start_y += (*ver_spat_ref);

	/* For each control point */	
	for(GS_y = start_y; GS_y <= (*ver_spat_ref); GS_y += search_step)
		for(GS_x = start_x; GS_x <= (*hor_spat_ref); GS_x += search_step)
		{
        /* HYUNDAI 980507 : start */
                         if ( Interlaced  )
                         {
                             if( GS_y%4 != 0) break;
                         }
        /* HYUNDAI 980507 : end */
			block_number = 0;
			
			/* For each MB in bb defined by control point */
			for(y = GS_y; y< GS_y + ((*vop_height)+N); y+=N)
				for(x = GS_x; x< GS_x + ((*vop_width)+N); x+=N)
				{
					check = 0;
					      
					/* for upper and lower side boundary pixels of each macroblock*/
				
      		for ( yy = 0; yy < N; yy+=(N-1)) 
					{
        		if (check) break;
        		for ( xx = 0; xx < N; xx++)
          		if((x+xx >= (*hor_spat_ref)) && (y+yy >= (*ver_spat_ref))
								 	&& (x+xx < ((*hor_spat_ref) + (*vop_width))) 
									&& (y+yy < ((*ver_spat_ref) + (*vop_height))))
             		if (alpha[((y+yy) * dim_x) + (x+xx)]) 
								{
                 		block_number ++;
                 		check ++;
                 		break;
                }
         	}
				
      		/* for left and right side boundary pixels of each macroblock*/
				
       		if(check==0) 
					{
         		for ( xx = 0; xx < N; xx+=(N-1)) 
						{
           		if (check) break;
           		for ( yy = 1; yy < N-1; yy++)
             		if((x+xx >= (*hor_spat_ref)) && (y+yy >= (*ver_spat_ref))
									&& (x+xx < ((*hor_spat_ref) + (*vop_width))) 
									&& (y+yy < ((*ver_spat_ref) + (*vop_height))))
                if (alpha[((y+yy) * dim_x) + (x+xx)]) 
								{	
                    block_number ++;
                    check ++;
                    break;
                }
            }
          }
   			}

			if(block_number <= min_block_number)
			{
				SARP_grid_x = GS_x;	
				SARP_grid_y = GS_y;
				min_block_number = block_number;

			}
		}

	/* Calculate relative shift of co-ord. for new bb */
	move_SARP_grid_x = (*hor_spat_ref) - SARP_grid_x;
	move_SARP_grid_y = (*ver_spat_ref) - SARP_grid_y;	

	/* Return bb co-ordinates based on minimum */		
	(*hor_spat_ref) = SARP_grid_x;
	(*ver_spat_ref) = SARP_grid_y;
	(*vop_width) += move_SARP_grid_x;
	(*vop_height) += move_SARP_grid_y;
	
	return;
}
 
/* 01.04.01: Studio Profile */
/***********************************************************CommentBegin******
 *
 * -- GetVopBounded_Studio -- Calculates and extracts the bounding box of this vop
 *
 * Author :		
 *	Noel O'Connor Teltec Irl.
 *
 * Created :		
 *	11-03-96
 *
 * Purpose :		
 *	This function takes as input the vop read from disk and
 * 	calculates and extracts the bounding box of this vop (based
 *	on the vop's alpha plane).
 * 
 * Arguments in : 	
 *	Vop *orig - pointer to original vop
 *	(i.e. vop read from disk),
 *	Int to_nearest_mb  - flag  	( 0 - gives the actual bounded vop )
 *					( 1 - gives the bounded vop extended
 *						to be of size n x 16 )
 *
 * Arguments in/out :	
 *	-
 *
 * Arguments out :	
 *	-
 *
 * Return values :	
 *	Vop *vop_bb - pointer to the bounded vop
 *
 * Side effects :	
 *	Memory is allocated within this function for
 *	the bounded vop
 *
 * Description :	
 *	This function first calculates the vop bounding box co-ordinates
 *	and dimensions. A vop is allocated with these dimensions. The 
 *	required image data is then extracted into this new bounded vop.
 *	Also the alpha plane charateristics (rectangular/arbitrary shape,
 *	binary/grey-level) are determined.
 *
 * See also :
 *	-
 *
 * Modified :		
 *	21.01.97 Added fix for bug VM4-970110-03 by Noel O'Connor
 *      07.05.98 : Jong Deuk Kim (HYUNDAI): modified for field-based chrominance subsampling
 *      01.04.01: Osamu Sunohara (SONY): Functions for Studio Profile
 *
 ***********************************************************CommentEnd********/
/* The input alpha(grayscale) image is used to calculate the bounded box. */
/* Therefore, if there is no alpha auxiliary component for grayscale shape, */
/* all the input alpha components must not be used for the calculation */
/* because a rectanglar binary shape with the same size with the input image. */
Vop *
GetVopsBounded_Studio(Vop *orig, Int to_nearest_mb, Int even_size, Int *vop_has_content, VolConfig *vol_config /* HYUNDAI 980507 */)
{
  Vop *vop_bb;
  
  Image *alpha_image[MAX_AUX_COMP_NUM][2],
        *y_image[2],
        *u_image[2],
        *v_image[2];
	
  SInt *alpha[MAX_AUX_COMP_NUM][2],
       *lum[2],
       *u[2],
       *v[2];
	
  Int x[2],
      y[2],
    hor_spat_ref[2],
    ver_spat_ref[2],
    y_ind,
    x_ind,
    width,
    height,
    multiple_of_mb_size,
    inside_alpha_image,
		mod_dim,
    lines[2],
    pels[2];	
  
  Int vop_coded; /* MW 30-NOV-1998 */

  Int k, vop_num;
  Int vop_structure;
  Int mem_pos_orig[2], mem_pos_bb[2];
  
  Int x_bb, y_bb;
  SInt *ptr_data;

  Int chroma_format = GetVopChromaFormat_st(orig);
  Int cdivx, cdivy;
  Int shape = GetVopShape(orig);
  Int rec_field_line=0;

  Int binary_number;
  Int i;
  
  if(chroma_format == CHROMA444)
    cdivx=cdivy=1;
  else if(chroma_format == CHROMA422)
    {
      cdivx=2;
      cdivy=1;
    }
  else
    cdivx=cdivy=2;
  
  
  if((vop_structure = GetVopVopStructure_st(orig)) == FRAME_VOP)
    vop_num = 1;
  else
    vop_num = 2;
  
  for(k=0;k<2;k++) {
    /* Use pointers to vop image data */
    for(i=0;i<MAX_AUX_COMP_NUM;i++)
      alpha_image[i][k] = GetVopA_st(orig,i,k);
    y_image[k]     = GetVopY_st(orig,k);
    u_image[k]     = GetVopU_st(orig,k);
    v_image[k]     = GetVopV_st(orig,k);
    
    /* Use pointers to YUVa */
    for(i=0;i<MAX_AUX_COMP_NUM;i++)/*AM2PDAM*/
      alpha[i][k] = (SInt *) GetImageData(alpha_image[i][k]);
    lum[k]   = (SInt *) GetImageData(y_image[k]);
    u[k]     = (SInt *) GetImageData(u_image[k]);
    v[k]     = (SInt *) GetImageData(v_image[k]);

    /* Use dimensions of alpha plane read from disk */
    x[k] = GetVopWidth_st (orig, k);
    y[k] = GetVopHeight_st(orig, k);
  }

  /*Decide Binary Shape Number*/
  binary_number = 0;
  if(shape==GREY_SCALE) {
    binary_number = -1;/* This means that all alpha_image(grayscale) must not used to achieve bounded box.*/
    for(i=0;i<GetVopAuxCompCount_st(orig);i++) {
      if(GetVopAuxCompType_st(orig,i)==AUX_COMP_ALPHA) {
	binary_number = i;
	break;
      }
    }
  }
  
        /* 
	 *	Call GetBoundingRect() to get the tightest rectangle
	 *		around Vop with the top left co-ordinates modified
	 *		to be even integers 
	 */

	/* Get _tightest_ rectangle */
  multiple_of_mb_size = FALSE;

	/* Don't limit to inside alpha image 			*/
	/* (actually this is only necessary with 	*/
	/* multiple_of_mb_size set to TRUE) 			*/
	inside_alpha_image = FALSE;		

	/* Calc. the bounding box information */

  if(!(shape=GREY_SCALE && binary_number==-1)) {
    GetBoundingRects_Studio(alpha[binary_number],x,y,
			    chroma_format, vol_config->interlaced, vop_structure,
			    multiple_of_mb_size,inside_alpha_image,
			    hor_spat_ref,ver_spat_ref,lines,pels);
    
  } else {
    /* If alpha auxiliary components don't exist for gray_scale_shape,
       rectangular binary shape with the size with the input image. */
    for(i=0;i<2;i++) {
      hor_spat_ref[i] = 0;
      ver_spat_ref[i] = 0;
      pels[i]  = x[i];
      lines[i] = y[i];
    }
  }
  for(k=0;k<2;k++) fprintf(stderr,"Tightest Recht k%d:x%d,y%d,hor%d,ver%d\n",
			   k,pels[k],lines[k],hor_spat_ref[k],ver_spat_ref[k]);
  
#ifdef _RANDOM_VOP_CODED_ /* MW 30-NOV-1998 */
  if(rand() <= (Int)(RAND_MAX/2.0))
    {
      pels  = 0;
      lines = 0;
    }
#endif

  *vop_has_content = TRUE;
  vop_coded = 1;
  
#if 0
        /* Check for vop content */
	if((pels != 0) && (lines != 0))
	  { /* MW 30-NOV-1998 */
	    vop_coded = 1;
		*vop_has_content = TRUE;
	  }
	else
	  { /* MW 30-NOV-1998 */
	    vop_coded = 0;
		*vop_has_content = FALSE;
	  }
#endif

	if(*vop_has_content)
	{
	  
		/* Do the Shape Adaptive Region Partitioning if necessary */
		if((hor_spat_ref != 0) || (ver_spat_ref != 0))
                        DoSARP_Studio(alpha[binary_number],x,
				      chroma_format, vop_structure,
				      hor_spat_ref,ver_spat_ref,pels,lines,
				      vol_config->interlaced /* HYUNDAI 980507 */);

		/* Modify the bounding box width and height to be multiples
				of MB_SIZE */
		for(k=0;k<vop_num;k++) {
		  if((mod_dim = (pels[k] %MB_SIZE)) != 0)
		    pels[k]  += (MB_SIZE - mod_dim);
		  
		  if(shape == RECTANGULAR && vol_config->interlaced) {
		    if(vop_structure == FRAME_VOP) {
		      lines[k] = 2*((lines[k]+31)/32)*16;
		    } else if(k==0) {
		      if((lines[0]-lines[1])>1) {
			fprintf(stderr,"bad case\n");
			exit(3);
		      }
		      rec_field_line = (((lines[0]+lines[1])+31)/32)*16;
		      lines[k] = rec_field_line;
		    } else {
		      lines[k] = rec_field_line;
		    }
		  } else {
		    if((mod_dim = (lines[k]%MB_SIZE)) != 0)
		      lines[k] += (MB_SIZE - mod_dim);
		  }

		  fprintf(stderr,"Decided rectanglek %d:x%d,y%d,hor%d,ver%d\n",
			  k,pels[k],lines[k],hor_spat_ref[k],ver_spat_ref[k]);
  
		  /* Allocate the bounding box */
		  if(k==0)
		    vop_bb = AllocFrameVop_Studio();
		  AllocVopChannels_Studio(vop_bb, vop_structure, k, pels[k], lines[k], chroma_format);
		}
	      }
	else   
		vop_bb = SallocVop();
               
  /* Extract image data into this bounded vop */
/*  width = pels; height = lines;	*/

  for(k=0;k<vop_num;k++) {
    /* Copy data from original vop */
    CopyVopNonImageField_Studio(orig, vop_structure, k, vop_bb, k);
    
    PutVopHorSpatRef_st(hor_spat_ref[k],vop_bb,k);	/* Spatial ref. of justified 	*/
    PutVopVerSpatRef_st(ver_spat_ref[k],vop_bb,k);	/* vop.	*/
    PutVopWidth_st(pels[k],vop_bb,k);		/* Store width and height */
    PutVopHeight_st(lines[k],vop_bb,k);		/* as integer multiples of 16 */
    PutVopQuantPrecision(GetVopQuantPrecision(orig), vop_bb);
    PutVopBitsPerPixel(GetVopBitsPerPixel(orig), vop_bb);
    
    /* MW 30-NOV-1998 */
    PutVopCoded(vop_coded,vop_bb);
  }
  
  /* Extract Y and A image data  */
	
  if (*vop_has_content)
    {

      /* For a rectangular shape, a region outside a tightest rectangle is filled with 0? */
      for(k=0;k<vop_num;k++) {

	y_bb = 0;
	for(y_ind = ver_spat_ref[k]; y_ind < ver_spat_ref[k] + lines[k]; y_ind++) {
	  x_bb = 0;
	  for(x_ind = hor_spat_ref[k]; x_ind < hor_spat_ref[k] + pels[k]; x_ind++) {

	    if (ValidCoordinate(x_ind,y_ind,x[k],y[k])) 
	      {
		CalcFieldMemoryPosition_Studio(vop_structure, k, x_ind, y_ind, x[k],   mem_pos_orig);
		CalcFieldMemoryPosition_Studio(vop_structure, k, x_bb,  y_bb, pels[k], mem_pos_bb);

		for(i=0;i<MAX_AUX_COMP_NUM;i++) {
		  ptr_data = (SInt *)GetImageData(GetVopA_st(vop_bb,i,mem_pos_bb[0]));
		  *(ptr_data + mem_pos_bb[1]) = alpha[i][mem_pos_orig[0]][mem_pos_orig[1]];
		}
		ptr_data = (SInt *)GetImageData(GetVopY_st(vop_bb,mem_pos_bb[0]));
		*(ptr_data + mem_pos_bb[1]) = lum[mem_pos_orig[0]][mem_pos_orig[1]];
	      }
	    x_bb++;
	  }
	  y_bb++;
	}
	
	/* Extract U and V image data */
	y_bb = 0;
	for(y_ind = ver_spat_ref[k]/cdivy; y_ind < (ver_spat_ref[k]/cdivy) + lines[k]/cdivy; y_ind++) {
	  x_bb = 0;
	  for(x_ind = hor_spat_ref[k]/cdivx; x_ind < (hor_spat_ref[k]/cdivx) + pels[k]/cdivx; x_ind++) {
	    
	    if (ValidCoordinate(x_ind,y_ind,x[k]/cdivx,y[k]/cdivy)) 
	      {
		CalcFieldMemoryPosition_Studio(vop_structure, k, x_ind, y_ind, x[k]/cdivx,   mem_pos_orig);
		CalcFieldMemoryPosition_Studio(vop_structure, k, x_bb,  y_bb, pels[k]/cdivx, mem_pos_bb);
		
		ptr_data = (SInt *)GetImageData(GetVopU_st(vop_bb,mem_pos_bb[0]));
		*(ptr_data + mem_pos_bb[1]) = u[mem_pos_orig[0]][mem_pos_orig[1]];
		ptr_data = (SInt *)GetImageData(GetVopV_st(vop_bb,mem_pos_bb[0]));
		*(ptr_data + mem_pos_bb[1]) = v[mem_pos_orig[0]][mem_pos_orig[1]];
	      }
	    x_bb++;
	  }
	  y_bb++;
	}
	
      }/*k*/
      
    }
  else PutVopArbitraryShape(ARB_SHAPE,vop_bb);
  
  return(vop_bb);
}

/***********************************************************CommentBegin******
 *
 * -- GetBoundingRects_Studio -- Obtains the position of the bounding rectangle
 *
 * Author :		
 *	Noel O'Connor Teltec Irl.
 *
 * Created :		
 *	11-03-96
 *
 * Purpose :		
 *	Obtains the position of the bounding rectangle of the shape of a vop.
 * 	Size can includes bottom-left extension to make size MULTIPLE of 16,
 *	and perhaps INSIDE the vop.
 * 
 * Arguments in : 	
 *	SInt	*alpha - pointer to alpha image
 *	Int		vop_width - width of input alpha image
 *	Int		vop_height - height of input alpha image
 *	Int		multiple - when this flag is set the function
 *				returns a BB which is a multiple of MB_SIZE
 *				Otherwise it returns the _exact_ BB.
 *	Int		inside - bounding rectangle inside the alpha plane
 *
 * Arguments in/out :	
 *	-
 *
 * Arguments out :	
 *	Int		*br_x0 - BB horizontal coordinate
 *	Int		*br_y0 - BB vertical coordinate
 *	Int		*br_lines - row size of BB
 *	Int		*br_pels - column size of BB
 *
 * Return values :	
 *	-
 *
 * Side effects :	
 *	-
 *
 * Description :	
 *	-
 *
 * See also :
 *	-
 *
 * Modified :		
 * 01.04.01 Osamu Sunohara (SONY): Function for Studio Profile
 *	
 *
 ***********************************************************CommentEnd********/

Void
GetBoundingRects_Studio(
     SInt        **alpha,      /* alpha plane data   */
     Int         *vop_width,   /* alpha plane width  */
     Int         *vop_height,  /* alpha plane height */
     Int         chroma_format,
     Int         Interlaced,
     Int         vop_structure, 
     Int         multiple,    /* 1: dimensions multiple of 16 */
     Int         inside,      /* bounding rectangle (BB) inside 
     				 the alpha plane */
     Int         *br_x0,  	  /* BB horizontal coordinate */
     Int         *br_y0,      /* BB vertical coordinate */
     Int         *br_lines,   /* rows of the BB */
     Int         *br_pels     /* columns of the BB */
     )   
{
#define BIGVALUE 10000
  Int i,j,
    x_min=  BIGVALUE,
    x_max=    -1,
    y_min=  BIGVALUE,
    y_max=    -1;
  
  Int base,kk;

  Int vop_num, k;
  Int mem_pos[2];

  
  if(vop_structure == FRAME_VOP)
    vop_num = 1;
  else
    vop_num = 2;

  for(k=0; k<vop_num; k++) {
    x_min=  BIGVALUE;
    x_max=    -1;
    y_min=  BIGVALUE;
    y_max=    -1;
    
    for(j=0;j<vop_height[k];j++)
      {
	for(i=0;i<vop_width[k];i++)
	  {
	    CalcFieldMemoryPosition_Studio(vop_structure, k, i, j, vop_width[k], mem_pos);
	    if(alpha[mem_pos[0]][mem_pos[1]]!=0)
	      {
		x_min=(i<x_min)?i:x_min;
		x_max=(i>x_max)?i:x_max;
		y_min=(j<y_min)?j:y_min;
		y_max=(j>y_max)?j:y_max;
	      }
	  }
      }

    /* horizontal */
    if(chroma_format != CHROMA444)
      x_min -= x_min%2;

    /* vertical */
    if(Interlaced && vop_structure == FRAME_VOP && chroma_format == CHROMA420)
      y_min -= y_min%4;
    else if((!Interlaced && chroma_format == CHROMA420) ||
	    (Interlaced && vop_structure == FRAME_VOP && chroma_format != CHROMA420) ||
	    (Interlaced && vop_structure == FIELD_VOP && chroma_format == CHROMA420))
      y_min -= y_min%2;
	
    if (!((x_min==BIGVALUE)||(x_max==-1)||(y_min==BIGVALUE)||(y_max==-1)))
      {
	br_x0[k] = x_min;
	br_y0[k] = y_min;
	br_pels[k] = x_max - x_min +1 ;
	br_lines[k]  = y_max - y_min +1 ;
	
	/* multiples of 16 */
	if(multiple==1) 
	  {
	    if((kk=(br_lines[k])%MB_SIZE)!=0)
	      br_lines[k]+=(MB_SIZE-kk);
	    if((kk=(br_pels[k])%MB_SIZE)!=0)
	      br_pels[k]+=(MB_SIZE-kk);
	  }
	
	/* check the bottom side */
	if (inside==1)
	  {
	    if ((br_x0[k]+br_pels[k])>vop_width[k]) 
	      {
		/* fprintf(stderr,"\nGetBoundingRect: fixing right bound!\n\n"); */
		br_x0[k]= vop_width[k] - br_pels[k] ;
	      }
	    /* check the right side */
	    if ((br_y0[k]+br_lines[k])>vop_height[k])  
	      { 
		/* fprintf(stderr,"\nGetBoundingRect: fixing bottom bound!\n\n"); */
		br_y0[k]= vop_height[k] - br_lines[k] ;
	      } 
	  }
	/* check if position of upper-left corner is inside the VOP */
	if( (br_x0[k] <0) || (br_y0[k] <0) )
	  {
	    fprintf(stderr,
		    "\nGetBoundingRect: UNABLE TO FIX POSITION-DIMENSIONS!\n\n");
	    fprintf(stderr,
		    "\n                 SETTING br_x0=br_y0=br_pels=br_lines=0 !\n\n");
	    br_x0[k] = br_y0[k] = br_pels[k] = br_lines[k] = 0;
	  }
	
      }
    else
      {
	br_x0[k] = br_y0[k] = br_pels[k] = br_lines[k]  = 0;
	fprintf(stderr,"GetBoundingRect: alpha plane without any object!!\n");
      }
  }
 
}/*GetBoundingRects_Studio*/

/***********************************************************CommentBegin******
 *
 * -- DoSARP_Studio -- Intelligent VOP formation as per VM 4.0 (Shape Adaptive Region
 *							Partitioning)
 *
 * Author :		
 *	Noel O'Connor Teltec Irl.
 *
 * Created :		
 *	11-03-96
 *
 * Purpose :		
 * 	Having found the tightest rectangle for the VOP (with top left coordinates 
 *	suitably adjusted) it is necessary to find the optimal top left co-ordinates
 *	for the bounding box which ensure that the minimum number of macroblocks are 
 *	coded.
 * 
 * Arguments in : 	
 *	SInt 	*alpha - pointer to alpha pixel data (unbounded)
 *	
 * Arguments in/out :	
 *	Int 	*hor_spat_ref - initially x co-ord. of tightest rect. (suitably adjusted)
 *	Int		*ver_spat_ref - initially y co-ord. of tightest rect. (suitably adjusted)
 *	Int		*vop_width - initially width of tightest rectangle
 *	Int		*vop_height - initially height of tightest rectangle
 *
 * Arguments out :	
 *	-
 *
 * Return values :	
 *	none
 *
 * Side effects :	
 *	-
 *
 * Description :	
 *	This code is based on the source code distributed by Gwang Hoon Park of Hyundai
 *	(11/10/96)
 *
 * See also :
 *	-
 *
 * Modified :		
 *   07.05.98 Jong Deuk Kim, Cheol Soo Park (HYUNDAI): 
 *            change the value of ver_spat_ref to multiple of four at interlaced
 *   01.04.01 Osamu Sunohara (SONY): Function for Studio Profile
 *	
 *
 ***********************************************************CommentEnd********/
Void
DoSARP_Studio(SInt **alpha,             /* Pointer to alpha plane image dat */
        Int *dim_x,              /* X dimension of alpha plane on disk */
	Int chroma_format,
        Int vop_structure,
        Int *hor_spat_ref,      /* current hor. spatial reference */
        Int *ver_spat_ref,      /* current ver. spatial referance */
        Int *vop_width,         /* current bb width */
        Int *vop_height,        /* current bb height */
        Int Interlaced          /* HYUNDAI 980507 */)
{
	Int		N = MB_SIZE,
				search_step_x,
	                        search_step_y,
				min_block_number = 999999,
				block_number,
				GS_x, GS_y,
				SARP_grid_x=0, SARP_grid_y=0,
				move_SARP_grid_x, move_SARP_grid_y,
				start_x, start_y,
				x,y,xx,yy,
				check;

	Int k, vop_num, mem_pos[2];


	if(vop_structure == FRAME_VOP)
	  vop_num = 1;
	else
	  vop_num = 2;

	/* search_step_x */
	if(chroma_format==CHROMA444)
	  search_step_x = 1;
	else
	  search_step_x = 2;

	/* search_step_y */
	if(Interlaced && vop_structure == FRAME_VOP && chroma_format == CHROMA420)
	  search_step_y = 4;
	else if((!Interlaced && chroma_format == CHROMA420) ||
		(Interlaced && vop_structure == FRAME_VOP && chroma_format != CHROMA420) ||
		(Interlaced && vop_structure == FIELD_VOP && chroma_format == CHROMA420))
	  search_step_y = 2;
	else
	  search_step_y = 1;

	/* vop roops */
	for(k=0;k<vop_num;k++) {
	  min_block_number = 999999;
	  
	  /* Get starting co-ordinates for loop */ 
	  start_x = (hor_spat_ref[k]>=0 && hor_spat_ref[k]<N) ? -(hor_spat_ref[k]) : -(N-search_step_x);/*-(N-2)*/
	  start_y = (ver_spat_ref[k]>=0 && ver_spat_ref[k]<N) ? -(ver_spat_ref[k]) : -(N-search_step_y);/*-(N-2)*/
	  
	  /* Offset for location of tightest rectangle in alpha image */
	  start_x += (hor_spat_ref[k]);
	  start_y += (ver_spat_ref[k]);
	  
	  /* For each control point */	
	  for(GS_y = start_y; GS_y <= (ver_spat_ref[k]); GS_y += search_step_y)
	    for(GS_x = start_x; GS_x <= (hor_spat_ref[k]); GS_x += search_step_x)
	      {

		block_number = 0;
		
		/* For each MB in bb defined by control point */
		for(y = GS_y; y< GS_y + ((vop_height[k])+N); y+=N)
		  for(x = GS_x; x< GS_x + ((vop_width[k])+N); x+=N)
		    {
		      check = 0;
		      
		      /* for upper and lower side boundary pixels of each macroblock*/
		      
		      for ( yy = 0; yy < N; yy+=(N-1)) 
			{
			  if (check) break;
			  for ( xx = 0; xx < N; xx++)
			    if((x+xx >= (hor_spat_ref[k])) && (y+yy >= (ver_spat_ref[k]))
			       && (x+xx < ((hor_spat_ref[k]) + (vop_width[k]))) 
			       && (y+yy < ((ver_spat_ref[k]) + (vop_height[k])))) {

			      CalcFieldMemoryPosition_Studio(vop_structure, k, x+xx, y+yy, dim_x[k], mem_pos);
			      if(alpha[mem_pos[0]][mem_pos[1]])
				{
				  block_number ++;
				  check ++;
				  break;
				}
			    }
			}
		      
		      /* for left and right side boundary pixels of each macroblock*/
		      
		      if(check==0) 
			{
			  for ( xx = 0; xx < N; xx+=(N-1)) 
			    {
			      if (check) break;
			      for ( yy = 1; yy < N-1; yy++)
				if((x+xx >= (hor_spat_ref[k])) && (y+yy >= (ver_spat_ref[k]))
				   && (x+xx < ((hor_spat_ref[k]) + (vop_width[k]))) 
				   && (y+yy < ((ver_spat_ref[k]) + (vop_height[k])))) {

				  CalcFieldMemoryPosition_Studio(vop_structure, k, x+xx, y+yy, dim_x[k], mem_pos);
				  if(alpha[mem_pos[0]][mem_pos[1]])
				    {	
				      block_number ++;
				      check ++;
				      break;
				    }
				}
			    }
			}
		    }
		
		if(block_number <= min_block_number)
		  {
		    SARP_grid_x = GS_x;	
		    SARP_grid_y = GS_y;
		    min_block_number = block_number;
		  }
	      }
	  
	  /* Calculate relative shift of co-ord. for new bb */
	  move_SARP_grid_x = (hor_spat_ref[k]) - SARP_grid_x;
	  move_SARP_grid_y = (ver_spat_ref[k]) - SARP_grid_y;	
	  
	  /* Return bb co-ordinates based on minimum */		
	  (hor_spat_ref[k]) = SARP_grid_x;
	  (ver_spat_ref[k]) = SARP_grid_y;
	  (vop_width[k])   += move_SARP_grid_x;
	  (vop_height[k])  += move_SARP_grid_y;
	}
	  
	  return;
      }
/* 01.04.01: Studio Profile */


