/*****************************************************************************
 *                                                                          
 * This software module was originally developed by 
 *
 * Martina Eckert (UPM-GTI / ACTS-MoMuSys)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4 
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free 
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original 
 * developer of this software module and his/her company, the subsequent 
 * editors and their companies, and ISO/IEC have no liability for use of this 
 * software module or modifications thereof in an implementation. Copyright is 
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming 
 * products. 
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own 
 * purpose, assign or donate the code to a third party and to inhibit third 
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works. 
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/

/***********************************************************HeaderBegin*******
 *                                                                         
 * File:        rc_upm2a.c
 *
 * Author:      Martina Eckert, UPM-GTI
 *
 * Created:     22-07-98
 *                                                                         
 * Description: UPM2 control functions
 *
 * Notes:       
 *
 * Flags:       -D_RC_DEBUG_  -  RC debugging   
 *
 * Modified:
 *      22.07.98 M. Eckert: Change UPM2 to more a more restrictive 
 *               functionality
 *                                 
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "momusys.h"
#include "mom_access.h"
#include "mom_structs.h"
#include "vm_config.h"
#include "mom_vop.h"
#include "mom_vol.h"
#include "vm_enc_defs.h"
#include "rc.h"


#define MAX_DIFF 1E-3

static RC_HIST      rc_hist_upm[MAX_NUM_VOS][MAX_NUM_VOLS][3];  /* History of 
                                                                   each VOL */
static RC_CFG_PARAM rc_cfg_upm[MAX_NUM_VOS][MAX_NUM_VOLS];   /* RC parameter */



/***********************************************************CommentBegin******
 *
 * -- PriorityOptimization_a --
 *
 * Author : 
 *      Martina Eckert, UPM-GTI     
 *
 * Created :            
 *      22-07-98
 *
 * Purpose :
 *      Calculates a distribution of the global target bitrate between the
 *      different VOs under the constraint to assure a minimum quality for
 *      the most important VO. If possible it is tried to obtain the same 
 *      quality for the rest of VOs according to its importance (position 
 *      in list). The position is provided by the user with the alpha values. 
 *      The difference to algorithm UPM2 is its higher restrictiveness:
 *      No bits are assigned to lower priority VOs, if for the more important 
 *      one the maximum possible distortion D_max still has not been achieved.
 *      Furthermore, there exists the option to a further restriction, which 
 *      is, that the qp of a VO can not be lower than the one of a VO with
 *      higher priority.
 *      
 * 
 * Arguments in :       
 *     Int    vo_id        -  ID of current VO 
 *     Int    vol_id       -  ID of current VOL 
 *     Int    num_vols_tot -  Number of VOPs 
 *     Double R            -  Global target bit rate
 *     Int   (* rate)(RC_HIST *rch, Int vo_i, Int vol_i, Int q, Double mad, 
 *                            Int frame_type)
 *                         -  Rate functions 
 *     Double (* dist)(RC_HIST *rch, Int vo_i, Int vol_i, Int q, Int mode_mod, 
 *                            Int frame_type)
 *                         -  Distortion functions 
 *     Double mad          -  MAD 
 *     LIST   pos[]        -  Priority list 
 *     Double psnr_min     -  Maximal allowed distortion of first VOP in list 
 *     Int    qp[][]       -  --> Decisions 
 *     Int    mode_mod     -  0 = calculation of distortion per pixel,
 *                            1 = calculation of total distortion 
 *     Int    mode_alg     -  0 = lowest qp of one VO can be the minimal 
 *                                possible one
 *                            1 = lowest qp of one VP has to be greater than 
 *                                qp of VO with higher priority
 *     Int    frame_type -  I=0, P=1, B=2
 *
 * Arguments in/out :   
 *      -
 *
 * Arguments out :      
 *      -
 *
 * Return values :      
 *     Double r_sum    -  Predicted bitrate for all VOs
 *
 * Side effects :       
 *      -
 *
 * Description :        
 *
 *
 * See also :
 *
 *
 * Modified :           
 *       08.10.99 M.Eckert: Fix bug in actualization with break flag
 *
 *
 ***********************************************************CommentEnd********/

static Double PriorityOptimization_a(
                                     Int    vo_id,                             
                                     Int    vol_id,                            
                                     Int    num_vols_tot,          
                                     Double R,                                 
                                     Int   (* rate)(RC_HIST *rch,
                                                    Int vo_i, Int vol_i, 
                                                    Int q, Double mad,
                                                    Int frame_type),
                                     Double (* dist)(RC_HIST *rch,
                                                     Int vo_i, Int vol_i, 
                                                     Int q, Int mode_mod,
                                                     Int frame_type),
                                     Double mad,                               
                                     LIST   pos[],                             
                                     Double psnr_min,                          
                                     Int    qp[][MAX_NUM_VOLS],
                                     Int    mode_mod,
                                     Int    mode_alg,
                                     Int    frame_type
                                     )

{
  Int     vo_i, vol_i, p, qp_aux, qp_min;
  Int     break_flag=0;
  Double  r_qp[MAX_NUM_VOS][MAX_NUM_VOLS], d_qp[MAX_NUM_VOS][MAX_NUM_VOLS];
  Double  r_sum = 0., D_max, r_aux, d_aux;
  Double  d_fin, psnr_fin;         
  RC_HIST  *rch_aux;

  /* Calculate maximal distortion */
  /* D_max = 255. / pow (10., psnr_min/20.); with PSNR=20*log10(255/D_max) */
  /* Dist is quadratic: */
  D_max = 65025. / pow (10., psnr_min/10.); /* with PSNR=10*log10(255^2/D_max) */
   

  fprintf(stdout,"Global RC: VOP [%d,%d] has got highest priority, \n \
it's minimal quality should be %f dB \n (intern distortion value %f) \n", 
          pos[0].vo_id, pos[0].vol_id, psnr_min, D_max);
  fflush(stdout);


  /***
   First step:
   We assign bits so that as many VOPs as
   possible have at most distortion per pixel  D_max:
   Starting with the VO of highest priority, the qp is decreased for all
   VOs until D_max (target of highest priority VO) is reached.
   Restrictions:
   In mode_alg=0, the qp has to be greater than the smallest
   possible qp, in mode_alg=1 it also has to be greater than the VO anterior
   in the list (with higher priority). 
   ***/

  for (p = 0; p < num_vols_tot; p++) 
    {    
      vo_i = pos[p].vo_id;
      vol_i = pos[p].vol_id;
      
      /* Set auxiliar variables for VO dependent values: */
      qp_aux  = MAX_QUANT;
      rch_aux = &rc_hist_upm[vo_i][vol_i][frame_type];
      r_aux   = 0;
      d_aux   = (*dist)(rch_aux, vo_i, vol_i, qp_aux, mode_mod, frame_type);
      qp_min  = get_quant(p, qp[pos[p-1].vo_id][pos[p-1].vol_id], mode_alg);

      if (rch_proc_stat(rch_aux) != 0) /* (still not processed) */
      {
        if (break_flag){
	  r_qp[vo_i][vol_i] = r_aux;
	  d_qp[vo_i][vol_i] = d_aux;
	  qp[vo_i][vol_i]   = qp_aux;
	  continue;
	}

        r_aux = (Double) (*rate)(rch_aux, vo_i, vol_i, qp_aux, mad, frame_type);

        while (d_aux > D_max && qp_aux > qp_min)
        {
          double r_new = 
            (Double) (*rate)(rch_aux, vo_i, vol_i, qp_aux-1, mad, frame_type);

          if (r_sum + r_new < R){
            qp_aux--;
            r_aux = r_new; }
          else{
            break_flag = 1;
	    r_qp[vo_i][vol_i] = r_aux;
	    d_qp[vo_i][vol_i] = d_aux;
	    qp[vo_i][vol_i]   = qp_aux;	      
            break; 
	  }
        }
      }
      else {
        qp_aux = rch_get_last_qp(rch_aux);
        r_aux = (Double) (*rate)(rch_aux, vo_i, vol_i, qp_aux, mad, frame_type);
        if (r_aux < 0.) r_aux = 0.;
      }

      /* Fill array with for this VO resulting auxiliar values: */
      r_qp[vo_i][vol_i] = r_aux;
      d_qp[vo_i][vol_i] = d_aux;
      qp[vo_i][vol_i]   = qp_aux;

      r_sum += r_qp[vo_i][vol_i];

#ifdef _RC_DEBUG_
      fprintf(stdout,"d[%d][%d] = %f, r[%d][%d] = %f with qp[%d][%d] = %d \n", 
              vo_i, vol_i, d_qp[vo_i][vol_i], 
              vo_i, vol_i, r_qp[vo_i][vol_i], 
              vo_i, vol_i, qp[vo_i][vol_i]);
      fflush(stdout);
#endif
   } /* rof */


  /***
   Second step:
   We assign the remaining bits until as many VOPs as possible 
   have QP = 1. 
   ***/
   

  break_flag = 0;

  for (p = 0; p < num_vols_tot; p++) {      
    vo_i = pos[p].vo_id;
    vol_i = pos[p].vol_id;
    
      /* Set auxiliar variables for VO dependent values: */
      rch_aux = &rc_hist_upm[vo_i][vol_i][frame_type];
      qp_aux  = qp[vo_i][vol_i];
      r_aux   = r_qp[vo_i][vol_i];
      d_aux   = (*dist)(rch_aux, vo_i, vol_i, qp_aux, mode_mod, frame_type);
      qp_min  = get_quant(p, qp[pos[p-1].vo_id][pos[p-1].vol_id], mode_alg);

    if (rch_proc_stat(rch_aux) != 0){ 
      
      /* decrease d[pos_p] */
#ifdef _RC_DEBUG_
      fprintf(stdout,"Increase rate of VOP[%d,%d]\n", vo_i, vol_i); 
      fflush(stdout);
#endif
      r_sum -= r_aux;

      while (qp_aux > qp_min)
      {         
        double r_new = 
          (Double) (*rate)(rch_aux, vo_i, vol_i, qp_aux-1, mad, frame_type);
        
        if (r_sum + r_new < R) {
          qp_aux--;
          r_aux = r_new;
        }
        else{
          break_flag = 1;
          break;
        }
      }

      /* Fill array with for this VO resulting auxiliar values: */
      r_qp[vo_i][vol_i] = r_aux; 
      d_qp[vo_i][vol_i] = d_aux;
      qp[vo_i][vol_i]   = qp_aux;

      r_sum += r_qp[vo_i][vol_i];
      
      printf(">>>>>> R_sum = %f \n", r_sum);
      if (break_flag) break;
      
    } /* fi */
  }/* rof */

   
  /* Finally obtained distortion / PSNR value for most important VO: */

  if (vo_id == pos[0].vo_id && vol_id == pos[0].vol_id) {
    d_fin = dist(&rc_hist_upm[pos[0].vo_id][pos[0].vol_id][frame_type],
                 pos[0].vo_id, 
                 pos[0].vol_id, 
                 qp[pos[0].vo_id][pos[0].vol_id], 
                 mode_mod,
                 frame_type);
    psnr_fin = 20 * log10 (255.0 / d_fin);
    
#ifdef _RC_DEBUG_
    WRITE_FLOAT("distor_script", psnr_fin);
#endif

    if (d_fin > D_max) { 
      fprintf(stdout,"Global RC - WARNING! PSNR = %f dB of VOP [%d,%d] can \
not be fulfilled! (estimation: %f dB)\n",
              psnr_min, pos[0].vo_id, pos[0].vol_id, psnr_fin);
      fflush(stdout);
    }
  }
  

  r_sum = 0.;
     
  for (p = 0; p < num_vols_tot; p++){
    vo_i = pos[p].vo_id;
    vol_i = pos[p].vol_id;
    r_sum +=  r_qp[vo_i][vol_i];
    
#ifdef _RC_DEBUG_
    fprintf(stdout,"d[%d][%d] = %f, r[%d][%d] = %f with qp[%d][%d] = %d \n",
            vo_i, vol_i, 
            dist(&rc_hist_upm[vo_i][vol_i][frame_type], vo_i, vol_i, 
                 qp[vo_i][vol_i], mode_mod, frame_type), 
            vo_i, vol_i, r_qp[vo_i][vol_i], 
            vo_i, vol_i, qp[vo_i][vol_i]);
    fflush(stdout);
#endif
  }


#ifdef _RC_DEBUG_
  fprintf(stdout,"r_sum = %f \n",r_sum); fflush(stdout);
#endif


  return r_sum;
  
}

/***********************************************************CommentBegin******
 *
 * -- RC_UPM2a_QuantAdjust --
 *
 * Author : 
 *      Martina Eckert, UPM-GTI     
 *
 * Created :            
 *      22-07-98
 *
 * Purpose :
 *      Gets alpha values and min_psnr and starts the UPM2a control algorithm
 *      for global rate distribution with the selected prediction model and
 *      restrictive mode mode_alg.
 *      Returns QP for current VO.
 * 
 * Arguments in :       
 *      Int    vo_id          -  VO Id
 *      Int    vol_id         -  VOL Id
 *      VOConfig *vo_list     -  To read rc-parameters
 *      Double R              -  Global target bitrate 
 *      Double mad            -  Current MAD
 *      UInt   num_pels       -  Current number of pixels in VO
 *      Int    rc_rate_model  -  Used rate prediction model
 *      Int    rc_dist_model  -  Used distortion prediction model
 *      Int    mode_mod       -  0 = calculation of distortion per pixel,
 *                               1 = calculation of total distortion 
 *      Int    mode_alg       -  0 = qp has to be higher than qp_min
 *                               1 = qp has to be higher than qp_min AND
 *                                   than qp of Vop of lower priority
 *      Int    frame_type     -  I=0, P=1, B=2
 *
 * Arguments in/out :   
 *      -
 *
 * Arguments out :      
 *      -
 *
 * Return values :      
 *      Int    qp       - Quantization parameter for current VO
 *
 * Side effects :       
 *      -
 *
 * Description :        
 *
 *
 * See also :
 *      File "README"
 *
 * Modified :       
 *       27.01.99 M.Eckert: Pass parameter mode_alg to select between two 
 *                          modes
 *       24.09.99 M.Eckert: Introduce variables "next_coding_time" and
 *                          "end_time" to facilitate coding of VOLs with
 *                          different length
 *
 ***********************************************************CommentEnd********/

Int RC_UPM2a_QuantAdjust(
                         Int    vo_id, 
                         Int    vol_id,
                         VOConfig *vo_list, 
                         Double R,          
                         Double mad,        
                         UInt   num_pels,   
                         Int    rc_rate_model,
                         Int    rc_dist_model,
                         Int    mode_mod,
                         Int    mode_alg,
                         Int    frame_type
                         )
                        
{

   VOConfig  *curr_vo;
   VolConfig *curr_vol;
   Int   (* rate_model) (RC_HIST*, Int, Int, Int, Double, Int);
   Double (* dist_model) (RC_HIST*, Int, Int, Int, Int, Int);
   RC_CFG_PARAM *rc_cfg = &rc_cfg_upm[vo_id][vol_id];

   /* Arrays including data of all VOPs: */

   Int    qp[MAX_NUM_VOS][MAX_NUM_VOLS];      /* QPs */
   Double alpha[MAX_NUM_VOS*MAX_NUM_VOLS];    /* Table of ordinal numbers 
                                                 for priority list */
   Float  psnr_min[MAX_NUM_VOS*MAX_NUM_VOLS]; /* Minimal assured qualities of 
                                                 VOLs */
   LIST   pos[MAX_NUM_VOS*MAX_NUM_VOLS];      /* Priority list */

   Short   vo_i, vol_i, id_vol[MAX_NUM_VOS*MAX_NUM_VOLS],
           id_vo[MAX_NUM_VOS*MAX_NUM_VOLS];   /* Counter */
   Int     i, j, i_min=0, num_vols_tot;
   Int     first_vo, first_vol;
   Double  alpha_min, psnr_thresh;   /* PSNR value to reach for the most 
                                        important VOL */

   Float   next_coding_time, end_time;

#ifdef _RC_DEBUG_
   Short   last_vo, last_vol;
   Double  r_tot;                    /* ---> Totally achieved bit-rate */
#endif

   /* Set process flags for global control: */



   curr_vo   = vo_list;
   first_vo  = GetVOConfigId(curr_vo);
   curr_vol  = GetVOConfigLayers(curr_vo);
   first_vol = GetVolConfigId(curr_vol);
   
   if (vo_id == first_vo && vol_id == first_vol)
      while(curr_vo != NULL)
      {
         vo_i = GetVOConfigId(curr_vo);
         curr_vol = GetVOConfigLayers(curr_vo);
         while(curr_vol != NULL)
         {
            vol_i = GetVolConfigId(curr_vol);
            rch_set_process_flag(&rc_hist_upm[vo_i][vol_i][frame_type], 2);    /* Status of all VOs: 
                                                                                  unprocessed */
            curr_vol = GetVolConfigNext(curr_vol);
         }
         curr_vo = GetVOConfigNext(curr_vo);
      }

   /* Set process flag for currently processed VO */

   rch_set_process_flag(&rc_hist_upm[vo_id][vol_id][frame_type], 1);   /*  VOL = [vo_id,vol_id] is 
                                                                           currently processed */

   for(vo_i = 0; vo_i < MAX_NUM_VOS; vo_i++)
      for(vol_i = 0; vol_i < MAX_NUM_VOLS; vol_i++)
      {
         qp[vo_i][vol_i]                     = 0;
         alpha[vo_i*MAX_NUM_VOLS+vol_i]      = -1.;
         psnr_min[vo_i*MAX_NUM_VOLS+vol_i]   = 0.;
         pos[vo_i*MAX_NUM_VOLS+vol_i].vo_id  = 999;
         pos[vo_i*MAX_NUM_VOLS+vol_i].vol_id = 999;
      }

#ifdef _RC_DEBUG_
   fprintf(stdout,"Global target rate: %f \n\n",R); fflush(stdout); 
#endif

   rch_store(&rc_hist_upm[vo_id][vol_id][frame_type], (Int)num_pels, mad);
 
   if (first_frame(&rc_hist_upm[vo_id][vol_id][frame_type])) 
   {
     /* Code  all VOs of first frame with initial qp: */
     qp[vo_id][vol_id] = rc_upm_get_QPfirst(rc_cfg, frame_type);
   }
   else 
   {
      /* Calculate QPs of optimal global distribution of target bit rate: */
      i = 0;
      curr_vo = vo_list;
      while(curr_vo != NULL)
      {
         curr_vol =  GetVOConfigLayers(curr_vo);
         while(curr_vol != NULL)
         {
            next_coding_time = GetVolConfigNextCodingTime(curr_vol);
            end_time = GetVolConfigEndTime(curr_vol);
	    if(next_coding_time <= end_time) /* VO in process */
	    {
	      id_vo[i] = GetVOConfigId(curr_vo);
	      id_vol[i] = GetVolConfigId(curr_vol);
	      alpha[i] = rc_get_RCParam(&rc_cfg_upm[id_vo[i]][id_vol[i]], 0);
	      psnr_min[i] = rc_get_RCParam(&rc_cfg_upm[id_vo[i]][id_vol[i]], 1);
	      i++; 
	    }
	    curr_vol = GetVolConfigNext(curr_vol);
         }
         curr_vo = GetVOConfigNext(curr_vo);
      }

      num_vols_tot = i;

#ifdef _RC_DEBUG_
      last_vo = id_vo[i];
      last_vol = id_vol[i];
#endif

      for(j = 0; j < num_vols_tot; j++)
      {
         alpha_min = 1E99;
         for(i = 0; i < num_vols_tot; i++)
            if (alpha[i] < alpha_min)
            {
               alpha_min = alpha[i];
               i_min = i;
            }
         pos[j].vo_id = id_vo[i_min];
         pos[j].vol_id = id_vol[i_min];
#ifdef _RC_DEBUG_
         fprintf(stdout,"%d, %d on position %d \n",pos[j].vo_id,pos[j].vol_id,j);
#endif
         if (j==0) psnr_thresh = psnr_min[i_min];
         alpha[i_min] = 1E99;
         i_min = 0;
      }


      switch (rc_rate_model) 
         {
         case RC_MODEL_UPM1:
         case RC_MODEL_UPM2:
            rate_model = RC_UPM_pred_bits;
            break; 
         default:
            error_exit("RC_UPM1_QuantAdjust: Error: Mode not supported\n");
         }     

      switch (rc_dist_model) 
         {
         case RC_MODEL_UPM1:
         case RC_MODEL_UPM2:
            dist_model = RC_UPM_pred_dist;
            break; 
         default:
            error_exit("RC_UPM1_QuantAdjust: Error: Mode not supported\n");
         }     

#ifdef _RC_DEBUG_
      r_tot = PriorityOptimization_a(vo_id, vol_id, 
                                     num_vols_tot, R,
                                     rate_model, dist_model, 
                                     mad,  
                                     pos, psnr_thresh, 
                                     qp, mode_mod, mode_alg,
                                     frame_type); 
#else
      PriorityOptimization_a(vo_id, vol_id, 
                             num_vols_tot, R,
                             rate_model, dist_model, 
                             mad,
                             pos, psnr_thresh, 
                             qp, mode_mod, mode_alg,
                             frame_type); 
#endif
   }
   
#ifdef _RC_DEBUG_

   if (vo_id == last_vo && vol_id == last_vol)
      WRITE_INT("rate_script",r_tot);


#endif
   
   rch_store_qp_before(&rc_hist_upm[vo_id][vol_id][frame_type], qp[vo_id][vol_id]);


#ifdef _RC_DEBUG_
   curr_vo = vo_list;
   while(curr_vo != NULL)
   {
      vo_i = GetVOConfigId(curr_vo);
      curr_vol = GetVOConfigLayers(curr_vo);
      while(curr_vol != NULL)
      {
         vol_i = GetVolConfigId(curr_vol);
         fprintf(stdout,"qp[%d][%d] = %d \n",vo_i, vol_i, qp[vo_i][vol_i]); 
         fflush(stdout);
         curr_vol = GetVolConfigNext(curr_vol);
      }
      curr_vo = GetVOConfigNext(curr_vo);
   }
#endif

   rch_set_process_flag(&rc_hist_upm[vo_id][vol_id][frame_type], 0);   
   /*  VOL = [vo_id,vol_id] has been finished */

   return qp[vo_id][vol_id];
}

/***********************************************************CommentBegin******
 *
 * -- ret_rch_upm2_a --
 *
 * Author : 
 *      Martina Eckert, UPM-GTI     
 *
 * Created :            
 *      22.07.98
 *
 * Purpose :
 *      Access to history data from other files
 *
 * Arguments in :       
 *      Int    vo_id      -  VO Id
 *      Int    vol_id     -  VOL Id
 *      Int    frame_type -  I, P or B frame
 *
 * Arguments in/out :   
 *      -
 *
 * Arguments out :      
 *      -
 *
 * Return values :      
 *      RC_HIST  *rch  - History data array  
 *
 *
 * Modified : 
 *
 *
 ***********************************************************CommentEnd********/

RC_HIST *get_rch_upm2_a(
                        Int  vo_id,
                        Int  vol_id,
                        Int  frame_type
                        )
{

  return &rc_hist_upm[vo_id][vol_id][frame_type];

}

/***********************************************************CommentBegin******
 *
 * -- get_rc_cfg_upm2_a --
 *
 * Author : 
 *      Martina Eckert, UPM-GTI     
 *
 * Created :            
 *      22.07.98
 *
 * Purpose :
 *      Access to history data from other files
 *
 * Arguments in :       
 *      Int    vo_id   -  VO Id
 *      Int    vol_id  -  VOL Id
 *
 * Arguments in/out :   
 *      -
 *
 * Arguments out :      
 *      -
 *
 * Return values :      
 *      RC_CFG_PARAM  *rc_cfg  - Configuration data (model parameters)
 *
 *
 * Modified : 
 *
 *
 ***********************************************************CommentEnd********/

RC_CFG_PARAM *get_rc_cfg_upm2_a(
                                Int  vo_id,
                                Int  vol_id
                                )
{

   return &rc_cfg_upm[vo_id][vol_id];

}

/***********************************************************CommentBegin******
 *
 * -- get_quant --
 *
 * Author : 
 *      Martina Eckert, UPM-GTI     
 *
 * Created :            
 *      27.01.99
 *
 * Purpose :
 *      Decide, which quantization parameter has to be valid as minimum value 
 *      according to current restriction mode of algorithm.
 *
 * Arguments in :       
 *      Int   p        -  Position of VO in hierarchy list
 *      Int   qp_ant   -  Quantization parameter of previous VO in priority list 
 *      Int   mode_alg -  mode for decision
 *
 * Arguments in/out :   
 *      -
 *
 * Arguments out :      
 *      -
 *
 * Return values :      
 *      Int   qp  -  Quantization parameter 
 *
 * Modified : 
 *
 *
 ***********************************************************CommentEnd********/

Int get_quant(
              Int   p, 
              Int   qp_ant, 
              Int   mode_alg
              )
{
  Int  qp;

  if (mode_alg == 0)
    qp = MIN_QUANT;
  else
    qp = (p==0 ? MIN_QUANT : qp_ant);

  return qp;
}

/*********************************************************** End of file ***/

