/******************************************************************************
 *
 * This software module was originally developed by
 *
 *   J.Ignacio Ronda (UPM-GTI / ACTS-MoMuSys)
 *   Angel Pacheco (UPM-GTI / ACTS-MoMuSys)
 *   Fernando Jaureguizar (UPM-GTI / ACTS-MoMuSys)
 *
 * and edited by
 *
 *   Angel Pacheco (UPM-GTI / ACTS-MoMuSys)
 *   Fernando Jaureguizar (UPM-GTI / ACTS-MoMuSys)
 *   Robert Danielsen (Telenor / ACTS-MoMuSys)
 *   Ferran Marques (UPC / ACTS-MoMuSys)
 *   Minhua Zhou (HHI / ACTS-MoMuSys)
 *   C.S. Boon (Matsushita Corporation)
 *   J. Takahashi
 *   Ji Heon Kweon (HYUNDAI)
 *   Jong Deuk Kim (HYUNDAI)
 *   Fujitsu Laboratories Ltd. (contact: Eishi Morimatsu)
 *   Dae-Sung Cho (Samsung AIT)
 *   Osamu Sunohara (SONY)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original
 * developer of this software module and his/her company, the subsequent
 * editors and their companies, and ISO/IEC have no liability for use of this
 * software module or modifications thereof in an implementation. Copyright is
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) standard conforming
 * products.
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own
 * purpose, assign or donate the code to a third party and to inhibit third
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works.
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/

/***********************************************************HeaderBegin*******
 *
 * File:        mot_padding.c
 *
 * Authors:     UPM - J.Ignacio Ronda / Angel Pacheco
 * Created:     02.02.96
 *
 * Description: Repetitive padding for vops (Y/YUV images) and separate images
 *
 * Flags:       -D_DEBUG_        :  debugging messages are printed
 *              -D_NO_PADDING_   :  no padding is performed
 *              -D_NO_TEXTURE_PADDING_ : TexturePadding() is not carried out
 *
 * Modified:
 *      version 1.1: Angel Pacheco: changes in the interface to allow
 *               one Vop as input, and to develop the padding over YUV values
 *               or only over Y, depending on the lumi field
 *                 lumi==1 only pad Y field
 *                 lumi!=1 pad YUV fields
 *      version 1.2: Angel Pacheco: fixes a bug that produces
 *               incorrect accesses in *_FinalVerticalPadding
 *      version 1.2.1: Angel Pacheco (UPM-GTI): bypassed a bug that produces
 *               incorrect padding if the shape is one pixel width and is
 *               touching the right side.
 *               THIS WILL BE FIXED IN THE NEXT VERSION (coming soon)
 *      21.04.96 Robert Danielsen: Reformatted. New headers.
 *      version 2.0: Angel Pacheco: fixed the bug bypassed of
 *               version 1.2.1 and modified the alpha plane subsampling for
 *               the chrominaces, acording to the VM2.x
 *      30.08.96 Angel Pacheco & Fernando Jaureguizar:
 *               Padding modification to perform TexturePadding as required
 *               for VM3.0.
 *      22.10.96 Angel Pacheco: New MB-based padding scheme as
 *               VM 4.0 describes.
 *      28.11.96 Ferran Marques: introduces changes of (22.10.96)
 *      05.02.97 Angel Pacheco: modified to include the new low
 *               pass extrapolation padding technique of the VM5.1
 *      05.02.97 Angel Pacheco: modified to fill the pixels
 *               acording the original (=!reconstructed) alpha values (VM5.1)
 *      17.04.97 Minhua Zhou: added vm_compos.h, modified
 *               MacroblockBasedPadding.
 *      09.04.97 MInhua Zhou: rewrote MacroblockBasedImagePadding
 *      09.04.97 Minhua Zhou: modified BlockRepetitivePadding
 *      26.05.97 Minhua Zhou: removed DoAlphaAndStatusPlanes(),
 *               Removed DoAlphaUVPlane(), added VopPadding().
 *      16.06.97 Angel Pacheco: renaming the B_* modes by MBM_ modes.
 *               default initialization of some variables to avoid warnings.
 *      13.11.97 Angel Pacheco: formating and new header.
 *      09.03.98 Fernando Jaureguizar: New formating. Deleted some not used
 *               variables and parameters.
 *      10.03.98 Fernando Jaureguizar: Deleted not used routines
 *               MacroblockBasedImagePadding(), BlockDefaultImagePadding()
 *               BlockRepetitiveImagePadding(), BlockImagePadding()
 *	07.05.98 Jong Deuk Kim (jdkim97@hei.co.kr, HYUNDAI): interlaced tools
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *      09.12.98 Ulrich Benzler (University of Hannover) : added further greyscale support
 *               removed RepetitiveImagePadding()
 *	06.09.99 Eishi Morimatsu (Fujitsu Labs.): added DRC support
 *      01.04.01 Osamu Sunohara (SONY): Functions for Studio Profile
 *
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/

#include "mot_padding.h"
#include "mot_util.h"
#include "mot_est.h"
#include "vm_compos.h"
#include "mom_image.h"

#ifdef _DEBUG_
#define INDEX(h,v,hdim,vdim)   ( (((h)<(hdim))&&((v)<(vdim)))? \
   ((h)+(v)*(hdim)) : fprintf(stderr,"out of bounds: (i,j,k) (%d,%d,%d); \
   (%d,%d)>(%d,%d)\n",(int)i,(int)j,(int)k,(int)h,(int)v,(int)hdim,(int)vdim) )
#else
#define INDEX(h,v,hdim,vdim)   ( (h)+(v)*(hdim) )
#endif

/***********************************************************CommentBegin******
 *
 * -- Y_HorizontalPadding -- Horizontal padding of the Y data
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Horizontal padding of the Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    alpha plane data
 *      Int    w,               alpha width
 *      Int    h,               alpha height
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane data
 *      SInt   *y_plane         Y plane data
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
Y_HorizontalPadding (
   SInt   *alpha_plane,  /* <-- alpha plane data  */
   Int    w,             /* <-- alpha width       */
   Int    h,             /* <-- alpha height      */
   SInt   *status_plane, /* --> status plane data */
   SInt   *y_plane       /* --> Y plane data      */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value;

  i=0;
  while(i<h)
    {
    j=previous=0;

    status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE : FIRST_PAD;
    if(status==INSIDE)
      status_plane[INDEX(j,i,w,h)]=INSIDE;

    j++;
    while(j<w)
      {
      if(alpha_plane[INDEX(j,i,w,h)]!=0)
        {
        if(status==FIRST_PAD)
          {
          value=y_plane[INDEX(j,i,w,h)];
          for(k=0;k<j;k++)
            {
            y_plane[INDEX(k,i,w,h)]=value;
            status_plane[INDEX(k,i,w,h)]=HOR_PAD;
            }
          status=INSIDE;
          status_plane[INDEX(j,i,w,h)]=INSIDE;
          }
        else if (status==MIDDLE_PAD)
          {
          value=((Int)y_plane[INDEX(previous,i,w,h)]+
                      (Int)y_plane[INDEX(j,i,w,h)]) >> 1;
          for(k=previous+1;k<j;k++)
            {
            y_plane[INDEX(k,i,w,h)] = value;
            status_plane[INDEX(k,i,w,h)]=HOR_PAD;
            }
          status_plane[INDEX(j,i,w,h)]=INSIDE;
          status=INSIDE;
          }
        else /* INSIDE */
          {
          status_plane[INDEX(j,i,w,h)]=INSIDE;
          }
        }
      else /* alpha_plane[INDEX(j,i,w,h)]==0 */
        {
        if(status==INSIDE)
          {
          status=MIDDLE_PAD;
          previous=j-1;
          }/* MIDDLE_PAD or FIRST_PAD do nothig */
        }
        j++;
      } /* while j */

    /* last part of the line */
    if(status==FIRST_PAD)
      {
      /* all line long */
      for(k=0;k<w;k++)
        {
        y_plane[INDEX(k,i,w,h)]=0;
        /* NO_PAD is the default for status_plane */
        }
      }
    else if(status==MIDDLE_PAD)
      {
      value=y_plane[INDEX(previous,i,w,h)];
      for(k=previous+1;k<w;k++)
        {
        y_plane[INDEX(k,i,w,h)]=value;
        status_plane[INDEX(k,i,w,h)]=HOR_PAD;
        }
      } /* if INSIDE, do nothing */
    i++;
  } /* while i */
}

/***********************************************************CommentBegin******
 *
 * -- Y_VerticalOverlapedPadding -- Vertical and overlapped padding
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Vertical and overlapped (horizontal and vertical zones) padding of the
 *      Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    original alpha plane
 *      Int    w,               original width  (Y or Alpha)
 *      Int    h,               original height (Y or Alpha)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane
 *      SInt   *y_plane         Y plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be possible to compute here the padding of the null (zero)
 *      values in the vertical direction
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
Y_VerticalOverlapedPadding (
   SInt   *alpha_plane,  /* <-- original alpha plane         */
   Int    w,             /* <-- original width  (Y or Alpha) */
   Int    h,             /* <-- original height (Y or Alpha) */
   SInt   *status_plane, /* --> status plane                 */
   SInt   *y_plane       /* --> y plane                      */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value;

  j=0;
  while(j<w)
    {
    i=previous=0;

    status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE:FIRST_PAD;
    /* status_plane[j,i] is updated if INSIDE */

    i++;
    while(i<h)
      {
      if(alpha_plane[INDEX(j,i,w,h)]!=0)
        {
        if(status==FIRST_PAD)
          {
          value=y_plane[INDEX(j,i,w,h)];
          for(k=0;k<i;k++)
            {
            if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
              {
              y_plane[INDEX(j,k,w,h)]=((Int)y_plane[INDEX(j,k,w,h)]+
                                         (Int)value)>>1;
              }
            else
              {
              y_plane[INDEX(j,k,w,h)]=value;
              status_plane[INDEX(j,k,w,h)]=VER_PAD;
              }
            }
          status=INSIDE;
          }
        else if (status==MIDDLE_PAD)
          {
          value=((Int)y_plane[INDEX(j,previous,w,h)]+
                      (Int)y_plane[INDEX(j,i,w,h)]) >> 1;
          for(k=previous+1;k<i;k++)
            {
            if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
              {
              y_plane[INDEX(j,k,w,h)]=((Int)y_plane[INDEX(j,k,w,h)]+
                                              (Int)value)>>1;
              }
            else if (status_plane[INDEX(j,k,w,h)]==NO_PAD)
              {
              y_plane[INDEX(j,k,w,h)]=value;
              status_plane[INDEX(j,k,w,h)]=VER_PAD;
              } /*else is INSIDE, do nothing */
            }
          status=INSIDE;
          } /* INSIDE do nothing */
        }
      else /* alpha_plane[INDEX(j,i,w,h)]==0 */
        {
        if(status==INSIDE)
          {
          status=MIDDLE_PAD;
          previous=i-1;
          }/* MIDDLE_PAD or FIRST_PAD do nothig */
        }
      i++;
      } /* while i */

    /* last part of the column */
    if(status==MIDDLE_PAD)
      {
      value=y_plane[INDEX(j,previous,w,h)];
      for(k=previous+1;k<h;k++)
        {
        if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
          {
          y_plane[INDEX(j,k,w,h)]=((Int)y_plane[INDEX(j,k,w,h)]+
                                     (Int)value)>>1;
          }
        else if (status_plane[INDEX(j,k,w,h)]==NO_PAD)
          {
          y_plane[INDEX(j,k,w,h)]=value;
          status_plane[INDEX(j,k,w,h)]=VER_PAD;
          }/* else INSIDE no change in the y_value */
        }/*for*/
      } /* else if(status==FIRST_PAD) */
    /* here we could perform the vertical padding in this column, i.e.,
       the same process made inside the first while of the
       Y_FinalVerticalPadding() fuction. The process would be the filling
       of the empty regions corresponding to the zones padded in the
       last part of the padding process (only the vertical value, which
       will be used to padd the region with the mean between the nearest
       horizontal&vertical values */

    j++;
    } /* while j */
}

/***********************************************************CommentBegin******
 *
 * -- Y_FinalHorizontalPadding -- Final horizontal zero padding
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
        02.02.96
 *
 * Purpose :
 *      Final zero padding in the horizontal direction of the Y data
 *
 * Arguments in :
 *      Int    w,               original width  (Y or Alpha)
 *      Int    h,               original height (Y or Alpha)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane
 *      SInt   *y_plane         Y plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be modified if VerticalOverlapedPadding() performs the
 *      vertical padding in the null (zero) values.
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
Y_FinalHorizontalPadding (
   Int    w,             /* <-- original width  (Y or Alpha) */
   Int    h,             /* <-- original height (Y or Alpha) */
   SInt   *status_plane, /* --> status plane                 */
   SInt   *y_plane       /* --> y plane                      */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value;
  Int    middle;

  i=0;
  while(i<h)
    {
    j=previous=0;

    /* perhaps, the first part of the line */
    if((status_plane[INDEX(j,i,w,h)]==HOR_PAD)||
          (status_plane[INDEX(j,i,w,h)]==INSIDE))
      {
      i++;
      }
    else
      {
      if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
        {
        /* first part of the line */

        j++;
        while(j<(w-1))
          {
          if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
            {
            value=y_plane[INDEX(j,i,w,h)];
            for(k=0;k<j;k++)
              y_plane[INDEX(k,i,w,h)]=value;
            j++;
            break;
            }
          else
            j++;
          }
        }

      /* the inner parts */

      status=VER_PAD;
      while(j<(w-1))
        {
        if (status==VER_PAD)
          {
          j++;
          if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
            {
            previous=j-1;
            }
          }
        else /* NO_PAD */
          {
          j++;
          if ((status=status_plane[INDEX(j,i,w,h)])==VER_PAD)
            {
            /*from the begin to the middle */
            middle=(previous+j) >> 1;
            value=y_plane[INDEX(previous,i,w,h)];
            for(k=previous+1;k<middle;k++)
              y_plane[INDEX(k,i,w,h)]=value;
            /*from the middle to the end */
            value=y_plane[INDEX(j,i,w,h)];
            for(k=middle;k<j;k++)
              y_plane[INDEX(k,i,w,h)]=value;
            }
          }
        }   /* while j */

      /* the last part of the line */
      if(status==VER_PAD)
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
          {
          y_plane[INDEX(j,i,w,h)]=y_plane[INDEX(j-1,i,w,h)];
          }
        else if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
          {
          /* the case of an empty line with a pixel object at the
             last column */
          value=y_plane[INDEX(j,i,w,h)];
          for(k=0;k<j;k++)
            y_plane[INDEX(k,i,w,h)]=value;
          }
       }
     else
       {
       if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
         {
         value=y_plane[INDEX(previous,i,w,h)];
         for(k=previous+1;k<w;k++)
           y_plane[INDEX(k,i,w,h)]=value;
         }
       else
         {
         /*from the begin to the middle */
         middle=(previous+j) >> 1;
         value=y_plane[INDEX(previous,i,w,h)];
         for(k=previous+1;k<middle;k++)
           y_plane[INDEX(k,i,w,h)]=value;
         /*from the middle to the end */
         value=y_plane[INDEX(j,i,w,h)];
         for(k=middle;k<w;k++)
           y_plane[INDEX(k,i,w,h)]=value;
         }
       } /* else VER_PAD */

     i++;
     }
   } /* while i */
}

/***********************************************************CommentBegin******
 *
 * -- Y_FinalVerticalPadding -- Final zero padding in the vertical direction
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Final zero padding in the vertical direction of the Y data
 *
 * Arguments in :
 *      Int    w,               original width  (Y or Alpha)
 *      Int    h,               original height (Y or Alpha)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane
 *      SInt   *y_plane         Y plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would disapear if VerticalOverlapedPadding() performs the vertical
 *      zero padding (see the comment in the last part of the that function
 *      code)
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
Y_FinalVerticalPadding (
   Int    w,             /* <-- original width  (Y or Alpha) */
   Int    h,             /* <-- original height (Y or Alpha) */
   SInt   *status_plane, /* --> status plane                 */
   SInt   *y_plane       /* --> y plane                      */

   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value;
  Int    middle;

  j=0;
  while(j<w)
    {
    i=previous=0;

    /* perhaps, the first part of the column */
    if((status_plane[INDEX(j,i,w,h)]==VER_PAD)||
          (status_plane[INDEX(j,i,w,h)]==INSIDE))
      {
      j++;
      }
    else
      {
      if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
        {
        /* first part of the column */

        i++;
        while(i<(h-1))
          {
          if(status_plane[INDEX(j,i,w,h)]==HOR_PAD)
            {
            value=y_plane[INDEX(j,i,w,h)];
            for(k=0;k<i;k++)
              y_plane[INDEX(j,k,w,h)]=(y_plane[INDEX(j,k,w,h)]+value)>>1;
            i++;
            break;
            }
          else
            i++;
          }
        }

      /* the inner parts */
      status=HOR_PAD;
      while(i<(h-1))
        {
        i++;
        if (status==HOR_PAD)
          {
          if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
            {
            previous=i-1;
            }
          }
        else if(status==NO_PAD) /* NO_PAD */
          {
          if ((status=status_plane[INDEX(j,i,w,h)])==HOR_PAD)
            {
            /*from the begin to the middle */
            middle=(previous+i) >> 1;
            value=y_plane[INDEX(j,previous,w,h)];
            for(k=previous+1;k<middle;k++)
              y_plane[INDEX(j,k,w,h)]=(value+y_plane[INDEX(j,k,w,h)])>>1;
            /*from the middle to the end */
            value=y_plane[INDEX(j,i,w,h)];
            for(k=middle;k<i;k++)
              y_plane[INDEX(j,k,w,h)]=(value+y_plane[INDEX(j,k,w,h)])>>1;
            }
          }/* else VER_PAD */
        } /* while i */

      /* the last part of the column */
      if(status==HOR_PAD)
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
        y_plane[INDEX(j,i,w,h)]=y_plane[INDEX(j,i-1,w,h)];
        }
      else
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
          {
          value=y_plane[INDEX(j,previous,w,h)];
          for(k=previous+1;k<h;k++)
            y_plane[INDEX(j,k,w,h)]=(value+y_plane[INDEX(j,k,w,h)])>>1;
          }
        } /* else HOR_PAD */

      j++;
    } /* else skip the column */

  } /* while j */
}

/***********************************************************CommentBegin******
 *
 * -- UV_HorizontalPadding -- Horizontal padding of the Y data
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Horizontal padding of the Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    subsampled alpha plane
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
UV_HorizontalPadding (
   SInt   *alpha_plane,  /* <-- subsampled alpha plane     */
   Int    w,             /* <-- subsampled width  (U or V) */
   Int    h,             /* <-- subsampled height (U or V) */
   SInt   *status_plane, /* --> subsampled status plane    */
   SInt   *u_plane,      /* --> u plane                    */
   SInt   *v_plane       /* --> v plane                    */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value1,value2;

  i=0;
  while(i<h)
    {
    j=previous=0;

    status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE:FIRST_PAD;
    if(status==INSIDE)
      status_plane[INDEX(j,i,w,h)]=INSIDE;

    j++;
    while(j<w)
      {
      if(alpha_plane[INDEX(j,i,w,h)]!=0)
        {
        if(status==FIRST_PAD)
          {
          value1=u_plane[INDEX(j,i,w,h)];
          value2=v_plane[INDEX(j,i,w,h)];
          for(k=0;k<j;k++)
            {
            u_plane[INDEX(k,i,w,h)]=value1;
            v_plane[INDEX(k,i,w,h)]=value2;
            status_plane[INDEX(k,i,w,h)]=HOR_PAD;
            }
          status=INSIDE;
          status_plane[INDEX(j,i,w,h)]=INSIDE;
          }
        else if (status==MIDDLE_PAD)
          {
          value1=((Int)u_plane[INDEX(previous,i,w,h)]+
                    (Int)u_plane[INDEX(j,i,w,h)]) >> 1;
          value2=((Int)v_plane[INDEX(previous,i,w,h)]+
                    (Int)v_plane[INDEX(j,i,w,h)]) >> 1;
          for(k=previous+1;k<j;k++)
            {
            u_plane[INDEX(k,i,w,h)] = value1;
            v_plane[INDEX(k,i,w,h)] = value2;
            status_plane[INDEX(k,i,w,h)]=HOR_PAD;
            }
          status_plane[INDEX(j,i,w,h)]=INSIDE;
          status=INSIDE;
          }
        else /* INSIDE */
          {
          status_plane[INDEX(j,i,w,h)]=INSIDE;
          }
        }
      else /* alpha_plane[INDEX(j,i,w,h)]==0 */
        {
        if(status==INSIDE)
          {
          status=MIDDLE_PAD;
          previous=j-1;
          }/* MIDDLE_PAD or FIRST_PAD do nothig */
        }
      j++;
      } /* while j */

    /* last part of the line */
    if(status==FIRST_PAD)
      {
      /* all line long */
      for(k=0;k<w;k++)
        {
        u_plane[INDEX(k,i,w,h)]=0;
        v_plane[INDEX(k,i,w,h)]=0;
        /* NO_PAD is the default for status_plane */
        }
      }
    else if(status==MIDDLE_PAD)
      {
      value1=u_plane[INDEX(previous,i,w,h)];
      value2=v_plane[INDEX(previous,i,w,h)];
      for(k=previous+1;k<w;k++)
        {
        u_plane[INDEX(k,i,w,h)]=value1;
        v_plane[INDEX(k,i,w,h)]=value2;
        status_plane[INDEX(k,i,w,h)]=HOR_PAD;
        }
      } /* if INSIDE, do nothing */
    i++;
    } /* while i */
}

/***********************************************************CommentBegin******
 *
 * -- UV_VerticalOverlapedPadding -- Vertical and overlapped padding
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Vertical and overlapped (horizontal and vertical zones) padding of the
 *      Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    subsampled alpha plane
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be possible to compute here the zero padding in the vertical
 *      direction
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
UV_VerticalOverlapedPadding (
   SInt   *alpha_plane,  /* <-- subsampled alpha plane     */
   Int    w,             /* <-- subsampled width  (U or V) */
   Int    h,             /* <-- subsampled height (U or V) */
   SInt   *status_plane, /* --> subsampled status plane    */
   SInt   *u_plane,      /* --> u plane                    */
   SInt   *v_plane       /* --> v plane                    */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value1,value2;

  j=0;
  while(j<w)
    {
    i=previous=0;

    status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE:FIRST_PAD;
    /* status_plane[j,i] is updated if INSIDE */

    i++;
    while(i<h)
      {
      if(alpha_plane[INDEX(j,i,w,h)]!=0)
        {
        if(status==FIRST_PAD)
          {
          value1=u_plane[INDEX(j,i,w,h)];
          value2=v_plane[INDEX(j,i,w,h)];

          for(k=0;k<i;k++)
            {
            if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
              {
              u_plane[INDEX(j,k,w,h)]=((Int)u_plane[INDEX(j,k,w,h)]+
                                         (Int)value1)>>1;
              v_plane[INDEX(j,k,w,h)]=((Int)v_plane[INDEX(j,k,w,h)]+
                                         (Int)value2)>>1;
              }
            else
              {
              u_plane[INDEX(j,k,w,h)]=value1;
              v_plane[INDEX(j,k,w,h)]=value2;
              status_plane[INDEX(j,k,w,h)]=VER_PAD;
              }
            }
          status=INSIDE;
          }
        else if (status==MIDDLE_PAD)
          {
          value1=((Int)u_plane[INDEX(j,previous,w,h)]+
                    (Int)u_plane[INDEX(j,i,w,h)]) >> 1;
          value2=((Int)v_plane[INDEX(j,previous,w,h)]+
                    (Int)v_plane[INDEX(j,i,w,h)]) >> 1;
          for(k=previous+1;k<i;k++)
            {
            if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
              {
              u_plane[INDEX(j,k,w,h)]=((Int)u_plane[INDEX(j,k,w,h)]+
                                         (Int)value1)>>1;
              v_plane[INDEX(j,k,w,h)]=((Int)v_plane[INDEX(j,k,w,h)]+
                                         (Int)value2)>>1;
              }
            else
              {
              u_plane[INDEX(j,k,w,h)]=value1;
              v_plane[INDEX(j,k,w,h)]=value2;
              status_plane[INDEX(j,k,w,h)]=VER_PAD;
              }
            }
          status=INSIDE;
          } /* INSIDE do nothing */

        }
      else /* alpha_plane[INDEX(j,i,w,h)]==0 */
        {
        if(status==INSIDE)
          {
          status=MIDDLE_PAD;
          previous=i-1;
          }/* MIDDLE_PAD or FIRST_PAD do nothig */
        }
      i++;
      } /* while i */

    /* last part of the column */
    if(status==MIDDLE_PAD)
      {
      value1=u_plane[INDEX(j,previous,w,h)];
      value2=v_plane[INDEX(j,previous,w,h)];
      for(k=previous+1;k<h;k++)
        {
        if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
          {
          u_plane[INDEX(j,k,w,h)]=((Int)u_plane[INDEX(j,k,w,h)]
                                    +(Int)value1)>>1;
          v_plane[INDEX(j,k,w,h)]=((Int)v_plane[INDEX(j,k,w,h)]
                                    +(Int)value2)>>1;
          }
        else if (status_plane[INDEX(j,k,w,h)]==NO_PAD)
          {
          u_plane[INDEX(j,k,w,h)]=value1;
          v_plane[INDEX(j,k,w,h)]=value2;
          status_plane[INDEX(j,k,w,h)]=VER_PAD;
          }/* else INSIDE no change in the y_value */
        }/*for*/
      } /*else if(status==FIRST_PAD) */
    /* here we could perform the vertical padding in this column, i.e.,
       the same process made inside the first while of the
       UV_FinalVerticalPadding() fuction. The process would be the filling
       of the empty regions corresponding to the zones padded in the
       last part of the padding process (only the vertical value, which
       will be used to padd the region with the mean between the nearest
       horizontal&vertical values */

    j++;
    } /* while j */
}

/***********************************************************CommentBegin******
 *
 * -- UV_FinalHorizontalPadding -- Final zero padding in horizontal direction
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Final zero padding in the horizontal direction of the Y data
 *
 * Arguments in :
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be modified if VerticalOverlapedPadding() performs the
 *      vertical zero padding
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
UV_FinalHorizontalPadding (
   Int    w,             /* <-- subsampled width  (U or V) */
   Int    h,             /* <-- subsampled height (U or V) */
   SInt   *status_plane, /* --> subsampled status plane    */
   SInt   *u_plane,      /* --> u plane                    */
   SInt   *v_plane       /* --> v plane                    */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value1,value2;
  Int    middle;

  i=0;
  while(i<h)
    {
    j=previous=0;

    /* perhaps, the first part of the line */
    if((status_plane[INDEX(j,i,w,h)]==HOR_PAD)||
         (status_plane[INDEX(j,i,w,h)]==INSIDE))
      {
      i++;
      }
    else
      {
      if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
        {
        /* first part of the line */

        j++;
        while(j<(w-1))
          {
          if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
            {
            value1=u_plane[INDEX(j,i,w,h)];
            value2=v_plane[INDEX(j,i,w,h)];
            for(k=0;k<j;k++)
              {
              u_plane[INDEX(k,i,w,h)]=value1;
              v_plane[INDEX(k,i,w,h)]=value2;
              }
            j++;
            break;
            }
          else
            j++;
          }
        }

      /* the inner parts */

      status=VER_PAD;
      while(j<(w-1))
        {
        if (status==VER_PAD)
          {
          j++;
          if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
            {
            previous=j-1;
            }
          }
        else /* NO_PAD */
          {
          j++;
          if ((status=status_plane[INDEX(j,i,w,h)])==VER_PAD)
            {
            /*from the begin to the middle */
            middle=(previous+j) >> 1;
            value1=u_plane[INDEX(previous,i,w,h)];
            value2=v_plane[INDEX(previous,i,w,h)];
            for(k=previous+1;k<middle;k++)
              {
              u_plane[INDEX(k,i,w,h)]=value1;
              v_plane[INDEX(k,i,w,h)]=value2;
              }
            /*from the middle to the end */
            value1=u_plane[INDEX(j,i,w,h)];
            value2=v_plane[INDEX(j,i,w,h)];
            for(k=middle;k<j;k++)
              {
              u_plane[INDEX(k,i,w,h)]=value1;
              v_plane[INDEX(k,i,w,h)]=value2;
              }
            }
          }
        }  /* while j */

      /* the last part of the line */
      if(status==VER_PAD)
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
          {
          u_plane[INDEX(j,i,w,h)]=u_plane[INDEX(j-1,i,w,h)];
          v_plane[INDEX(j,i,w,h)]=v_plane[INDEX(j-1,i,w,h)];
          }
        else if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
          {
          /* the case of an empty line with a pixel object at the
             last column */
          value1=u_plane[INDEX(j,i,w,h)];
          value2=v_plane[INDEX(j,i,w,h)];
          for(k=0;k<j;k++)
            {
            u_plane[INDEX(k,i,w,h)]=value1;
            v_plane[INDEX(k,i,w,h)]=value2;
            }
          }

        }
      else
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
          {
          value1=u_plane[INDEX(previous,i,w,h)];
          value2=v_plane[INDEX(previous,i,w,h)];
          for(k=previous+1;k<w;k++)
            {
            u_plane[INDEX(k,i,w,h)]=value1;
            v_plane[INDEX(k,i,w,h)]=value2;
            }
          }
        else
          {
          /*from the begin to the middle */
          middle=(previous+j) >> 1;
          value1=u_plane[INDEX(previous,i,w,h)];
          value2=v_plane[INDEX(previous,i,w,h)];
          for(k=previous+1;k<middle;k++)
            {
            u_plane[INDEX(k,i,w,h)]=value1;
            v_plane[INDEX(k,i,w,h)]=value2;
            }
          /*from the middle to the end */
          value1=u_plane[INDEX(j,i,w,h)];
          value2=v_plane[INDEX(j,i,w,h)];
          for(k=middle;k<w;k++)
            {
            u_plane[INDEX(k,i,w,h)]=value1;
            v_plane[INDEX(k,i,w,h)]=value2;
            }
          }
        } /* else VER_PAD */

      i++;
      }
    } /* while i */
}

/***********************************************************CommentBegin******
 *
 * -- UV_FinalVerticalPadding -- Final zero padding in the vertical direction
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Final zero padding in the vertical direction of the Y data
 *
 * Arguments in :
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would disapear if VerticalOverlapedPadding() performs the vertical
 *      zero padding
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
UV_FinalVerticalPadding (
   Int    w,             /* <-- subsampled width  (U or V) */
   Int    h,             /* <-- subsampled height (U or V) */
   SInt   *status_plane, /* --> subsampled status plane    */
   SInt   *u_plane,      /* --> u plane                    */
   SInt   *v_plane       /* --> v plane                    */
   )
{
  Int    i, j;
  Int    k=0;
  Int    status;
  Int    previous;
  SInt   value1,value2;
  Int    middle;

  j=0;
  while(j<w)
    {
    i=previous=0;

    /* perhaps, the first part of the column */
    if((status_plane[INDEX(j,i,w,h)]==VER_PAD)||
         (status_plane[INDEX(j,i,w,h)]==INSIDE))
      {
      j++;
      }
    else
      {
      if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
        {
        /* first part of the column */

        i++;
        while(i<(h-1))
          {
          if(status_plane[INDEX(j,i,w,h)]==HOR_PAD)
            {
            value1=u_plane[INDEX(j,i,w,h)];
            value2=v_plane[INDEX(j,i,w,h)];
            for(k=0;k<i;k++)
              {
              u_plane[INDEX(j,k,w,h)]=(u_plane[INDEX(j,k,w,h)]+value1)>>1;
              v_plane[INDEX(j,k,w,h)]=(v_plane[INDEX(j,k,w,h)]+value2)>>1;
              }
            i++;
            break;
            }
          else
            i++;
          }
        }

      /* the inner parts */
      status=HOR_PAD;
      while(i<(h-1))
        {
        i++;
        if (status==HOR_PAD)
          {
          if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
            {
            previous=i-1;
            }
          }
        else if(status==NO_PAD) /* NO_PAD */
          {
          if ((status=status_plane[INDEX(j,i,w,h)])==HOR_PAD)
            {
            /*from the begin to the middle */
            middle=(previous+i) >> 1;
            value1=u_plane[INDEX(j,previous,w,h)];
            value2=v_plane[INDEX(j,previous,w,h)];
            for(k=previous+1;k<middle;k++)
              {
              u_plane[INDEX(j,k,w,h)]=(value1+u_plane[INDEX(j,k,w,h)])>>1;
              v_plane[INDEX(j,k,w,h)]=(value2+v_plane[INDEX(j,k,w,h)])>>1;
              }
            /*from the middle to the end */
            value1=u_plane[INDEX(j,i,w,h)];
            value2=v_plane[INDEX(j,i,w,h)];
            for(k=middle;k<i;k++)
              {
              u_plane[INDEX(j,k,w,h)]=(value1+u_plane[INDEX(j,k,w,h)])>>1;
              v_plane[INDEX(j,k,w,h)]=(value2+v_plane[INDEX(j,k,w,h)])>>1;
              }
            }
          }
        } /* while i */

      /* the last part of the column */
      if(status==HOR_PAD)
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
          {
          u_plane[INDEX(j,i,w,h)]=u_plane[INDEX(j,i-1,w,h)];
          v_plane[INDEX(j,i,w,h)]=v_plane[INDEX(j,i-1,w,h)];
          }
        }
      else
        {
        if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
          {
          value1=u_plane[INDEX(j,previous,w,h)];
          value2=v_plane[INDEX(j,previous,w,h)];
          for(k=previous+1;k<h;k++)
            {
            u_plane[INDEX(j,k,w,h)]=(value1+u_plane[INDEX(j,k,w,h)])>>1;
            v_plane[INDEX(j,k,w,h)]=(value2+v_plane[INDEX(j,k,w,h)])>>1;
            }

          }
        } /* else HOR_PAD */

      j++;
      } /* else skip the column */

    }/* while j */
}

/********************************************************** *CommentBegin******
 *
 * -- RepetitivePadding -- Repetitive padding of the YUV images
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Repetitive padding of the YUV images of vop_ima depending on lumi field
 *
 * Arguments in :
 *      Int   lumi,      1=only padd Y image, !=1 padd YUV images
 *
 * Arguments in/out :
 *      Vop   *vop_ima   Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      The downsampling of the alpha plane is done, if necesary, acording
 *      VM 2.x (valid alpha values are only the odd columns/files ones)
 *
 * See also :
 *
 *
 * Modified : 981209 U. Benzler (University of Hannover) : added greyscal support
 *
 *
 ***********************************************************CommentEnd********/

Int
RepetitivePadding (
   Int   lumi,    /* <-- 1=only padd Y image, !=1 padd YUV images  */
   Vop   *vop_ima /* --> Vop to pad                                */
   )
{
  SInt    *alpha_plane;
  SInt    *y_plane;
  SInt    *g_plane[MAX_MAC];/* U. Benzler, 981209 : added gray scale support */ /* MAC (SB) 12-Nov-99 */
  SInt    *u_plane;
  SInt    *v_plane;
  SInt    *status_plane;
  SInt    *alp_plane;
  Int     i, j, base, n;
  Image   *alpha_ima;
  Image   *y_ima;
  Image   *g_ima[MAX_MAC];/* U. Benzler, 981209 : added gray scale support */ /* MAC (SB) 12-Nov-99 */
  Image   *u_ima;
  Image   *v_ima;

  Int     y_x, y_y;
  Int     u_x, u_y;
  Int     v_x, v_y;

  Int     h;
  Int     w;
  Int     aux_comp_count = GetVopAuxCompCount(vop_ima);

  /* checkings */

#ifdef _NO_PADDING_
  return(2);
#endif

  if( (vop_ima==NULL) || ((alpha_ima=vop_ima->a_chan)==NULL) ||
      ((y_ima=vop_ima->y_chan)==NULL) || ((u_ima=vop_ima->u_chan)==NULL) ||
      ((v_ima=vop_ima->v_chan)==NULL) ||
      ((alpha_plane=(SInt *)GetImageData(alpha_ima))==NULL) ||
      ((y_plane=(SInt *)GetImageData(y_ima))==NULL) ||
      ((u_plane=(SInt *)GetImageData(u_ima))==NULL)||
      ((v_plane=(SInt *)GetImageData(v_ima))==NULL) )
  {
    fprintf(stderr,"RepetitivePadding: NULL input pointers\n");
    return(0);
  }

  if ( (GetVopShape(vop_ima)==GREY_SCALE) ) {
     for(n=0;n<aux_comp_count;n++) { /* MAC (SB) 12-Nov-99 */
       if (((g_ima[n]=vop_ima->g_chan[n])==NULL) ||
           ((g_plane[n]=(SInt *)GetImageData(g_ima[n]))==NULL))
       {
         fprintf(stderr,"RepetitivePadding: NULL input pointers\n");
         return(0);
       }
     }
  }


  if ( (alpha_ima->grid=='h') || (y_ima->grid=='h') ||
       (u_ima->grid=='h')||(v_ima->grid=='h') )
    {
    fprintf(stderr,"RepetitivePadding: hexadecimal grid not supported\n");
    return(-1);
    }

  w=y_x=(Int)GetImageSizeX(y_ima);
  h=y_y=(Int)GetImageSizeY(y_ima);

  if ( (y_x!=GetImageSizeX(alpha_ima)) || (y_y!=GetImageSizeY(alpha_ima)) )
    {
    fprintf(stderr,"RepetitivePadding: different image dimensions \n");
    return(0);
    }


  if(lumi!=1)
    { /* do YUV padding */
    u_x=(Int)GetImageSizeX(u_ima);
    u_y=(Int)GetImageSizeY(u_ima);
    v_x=(Int)GetImageSizeX(v_ima);
    v_y=(Int)GetImageSizeY(v_ima);

    if ( (u_x!=v_x) || (u_y!=v_y) || ((y_y/2)!=u_y) ||
       (!((y_x/2==u_x)||(y_x==u_x)))  )
      {
      fprintf(stderr,"RepetitivePadding: different image dimensions \n");
      return(0);
      }
    }

  if ((status_plane= (SInt*)malloc(h*w*sizeof(SInt)))==NULL)
    {
    fprintf(stderr,"RepetitivePadding: malloc failed\n");
    return(-3);
    }

  for(j=0;j<h;j++)
    {
    base=j*w;
    for(i=0;i<w;i++)
      status_plane[base+i]= NO_PAD;
    }

  /* do Y padding */

  Y_HorizontalPadding (alpha_plane, w, h, status_plane, y_plane);

  Y_VerticalOverlapedPadding (alpha_plane, w, h, status_plane, y_plane);

  Y_FinalHorizontalPadding (w, h, status_plane, y_plane);

  Y_FinalVerticalPadding (w, h, status_plane, y_plane);

/* U. Benzler, 981208 : added gray scale support >>*/
  if (GetVopShape(vop_ima)== GREY_SCALE)
    {
      for(n=0;n<aux_comp_count;n++) {	/* MAC (SB) 12-Nov-99 */

        for(j=0;j<h;j++)
	{
	  base=j*w;
	  for(i=0;i<w;i++)
	    status_plane[base+i]= NO_PAD;
	}
      
        Y_HorizontalPadding (alpha_plane, w, h, status_plane, g_plane[n]);
        Y_VerticalOverlapedPadding (alpha_plane, w, h, status_plane, g_plane[n]);
        Y_FinalHorizontalPadding (w, h, status_plane, g_plane[n]);
        Y_FinalVerticalPadding (w, h, status_plane, g_plane[n]);
      }
    }
/* U. Benzler, 981208 : added gray scale support <<*/
  if (lumi!=1)
    {
    free ((Char*)status_plane);

    /* do UV padding */

    h=h/2;
    w=w/2;
    if ((status_plane= (SInt*)malloc(h*w*sizeof(SInt)))==NULL)
      {
      fprintf(stderr,"RepetitivePadding: malloc failed\n");
      return(-3);
      }

    for(j=0;j<h;j++)
      {
      base=j*w;
      for(i=0;i<w;i++)
        status_plane[base+i]= NO_PAD;
      }
    alp_plane = (SInt *)GetImageData(vop_ima->a_uv_chan);

    UV_HorizontalPadding (alp_plane, w, h, status_plane, u_plane,v_plane);

    UV_VerticalOverlapedPadding (alp_plane, w, h, status_plane,u_plane,v_plane);

    UV_FinalHorizontalPadding (w, h, status_plane, u_plane,v_plane);

    UV_FinalVerticalPadding (w, h, status_plane, u_plane,v_plane);

    free ((Char*)status_plane);

    return(1);
    }
  else
    {
    free ((Char*)status_plane);

    return(-4);
    }
}

/***********************************************************CommentBegin******
 *
 * -- CheckParameters --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Checks the Input Parameters the TexturePadding fuction.
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   previous reconstructed alpha plane data
 *      Image   *alpha_decisions,  the previous reconstructed alpha plane
 *                                 subsampled (in block units) which values
 *                                 are:
 *                                  MBM_TRANSPARENT = all pixels are transparent
 *                                        (outside the shape)
 *                                  MBM_OPAQUE      = all pixels are not transp.
 *                                        (inside the shape)
 *                                  MBM_BOUNDARY    = transparent and not
 *                                        transparent pixels at the same time
 *                                        (in the boundary of the shape)
 *      Image   *MB_decisions,     encoding mode for each MB. If this Image is
 *                                  NULL      => Frame based repetitive padd.
 *                                  otherwise => Block based repetitive and
 *                                               zero padding.
 *
 * Arguments in/out :
 *      Vop     *cur_vop           Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified : 981209 U. Benzler (University of Hannover) : added greyscal support
 *
 *
 ***********************************************************CommentEnd********/

Int
CheckParameters (
   Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
   Image   *alpha_decisions, /* <-- subsampled (per block) prev_rec_alpha   */
   Image   *MB_decisions,    /* <-- encoding mode for each MB               */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )
{

  Image   *y_ima,*u_ima,*v_ima; /*,*g_ima; */ /* U. Benzler, 981209 : added gray scale support */
  Int     y_x,y_y,u_x,v_x,u_y,v_y, n;

  if( (cur_vop==NULL) ||
      ((y_ima=cur_vop->y_chan)==NULL) ||((u_ima=cur_vop->u_chan)==NULL) ||
      ((v_ima=cur_vop->v_chan)==NULL) ||
      (GetImageData(prev_rec_alpha)==NULL) ||
      (GetImageData(y_ima)==NULL) ||
      (GetImageData(u_ima)==NULL) ||
      (GetImageData(v_ima)==NULL) ||
      (alpha_decisions==NULL) || (MB_decisions==NULL)  ||
      (GetImageData(alpha_decisions)==NULL) ||
      (GetImageData(MB_decisions)==NULL))
  {
    fprintf(stderr,"CheckParameters: NULL input pointers\n");
    return(0);
  }

  if ( (GetVopShape(cur_vop)==GREY_SCALE) ) {
    for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { /* MAC (SB) 12-Nov-99 */
      if (cur_vop->g_chan[n]==NULL)
      {
        fprintf(stderr,"CheckParameters: NULL input pointers\n");
        return(0);
      }
      if (cur_vop->g_chan[n]->grid=='h')
      {
        fprintf(stderr,"CheckParameters: hexadecimal grid not supported\n");
        return(-1);
      }
    }
  }



  if ( (prev_rec_alpha->grid=='h') || (y_ima->grid=='h') ||
       (u_ima->grid=='h')||(v_ima->grid=='h')) /* ||
       ((GetVopShape(cur_vop) == GREY_SCALE) && (g_ima->grid=='h')))*/ 
    /* U. Benzler, 981209 : added gray scale support */
    {
    fprintf(stderr,"CheckParameters: hexadecimal grid not supported\n");
    return(-1);
    }

  y_x=(Int)GetImageSizeX(y_ima);
  y_y=(Int)GetImageSizeY(y_ima);

  if ( (y_x!=GetImageSizeX(y_ima)) || (y_y!=GetImageSizeY(y_ima)) )
    {
    fprintf(stderr,"CheckParameters: different image dimensions \n");
    return(-1);
    }

  u_x=(Int)GetImageSizeX(u_ima);
  u_y=(Int)GetImageSizeY(u_ima);
  v_x=(Int)GetImageSizeX(v_ima);
  v_y=(Int)GetImageSizeY(v_ima);

  if ( (u_x!=v_x) || (u_y!=v_y) || ((y_y/2)!=u_y) ||
     (!((y_x/2==u_x)||(y_x==u_x)))  )
     {
     fprintf(stderr,"CheckParameters: different image dimensions \n");
     return(-1);
     }

  return 1;
}

/***********************************************************CommentBegin******
 *
 * -- MacroblockBasedPadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      21.10.96
 *
 * Purpose :
 *      Macroblock-based padding of the Vop images, according to the
 *      VM 5.1
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   Previous reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Vop     *cur_vop,          Vop to pad
 *
 * Arguments out :
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the YUV images of a given VOP
 *      needed in the motion estimation and compensation (see section 3.3.1).
 *
 *      1) The padded Vop is NOT extended 16 pixels in all directions (it is
 *         supposed that this extension has been previously done).
 *      2) Padding is performed SEPARATELY FOR EACH of the Y, U/V bands, but
 *         grouped in MB/B units.
 *      3) Padding of the MB/B uses the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block (input parameter).
 *      4) Padding of chroma 8x8 block is performed by decimating the 16x16
 *         alpha block (by descarding every other row and column of pixels,
 *         starting from the second row and column, with an OR operation
 *         over the alpha pixel values).
 *
 * See also :
 *
 *
 * Modified :
 *      28-01-97 Robert Danielsen: Fix for freeing alpha_modes. Fix by
 *                Noel O'Connor.
 *      06.02.97 Angel Pacheco: the padding of the transparent blocks
 *                 is made looking at the most recently padded boundary block.
 *      17.04.97 Minhua Zhou: rewrote this program
 *      27.07.97 C.S. Boon: Changed to VM8.0 padding -- simplified padding
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *
 ***********************************************************CommentEnd********/

Int
MacroblockBasedPadding (
   Image   *prev_rec_alpha,/* <-- previous reconstructed alpha plane data  */
   Vop     *cur_vop       /* <-> Vop to pad                                */
   )
{
  Image   *alpha_sub_ima,
          *y_ima, *u_ima, *v_ima;
  Int 	  image_type; /* HYUNDAI (Grayscale) */ 
  SInt    *alpha_modes;
  SInt    *alpha,
          *alpha_uv,
          *alpha_sub,
          *buf;
  Int     a_w,a_h,i,j,i_ref,j_ref, padding_type, n;
  Int     h_size,w_size;
  Int     default_padding_value =1<<(GetVopBitsPerPixel(cur_vop) -1);
 
#ifdef _NO_PADDING_
  return(2);
#endif

  /*
   * MACROBLOCK-BASED REPETITIVE PADDING
   */

  /* checkings */
  if ( (cur_vop==NULL) || (prev_rec_alpha==NULL) ||
       ((y_ima=GetVopY(cur_vop))==NULL) || ((u_ima=GetVopU(cur_vop))==NULL) ||
       ((v_ima=GetVopV(cur_vop))==NULL) ||
       ((a_w=GetImageSizeX(GetVopY(cur_vop)))<1)  ||
       ((a_h=GetImageSizeY(GetVopY(cur_vop)))<1)  ||
       ((alpha=(SInt*)GetImageData(prev_rec_alpha))==NULL) ||
       (a_w!=GetImageSizeX(prev_rec_alpha)) ||
       (a_h!=GetImageSizeY(prev_rec_alpha)))
    {
      fprintf(stderr,"MacroblockBasedPadding: incorrect input parameters\n\n");
      return(-1);
    }
  
  if (GetVopShape(cur_vop)==GREY_SCALE) {
    for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { /* MAC (SB) 12-Nov-99 */
      if (GetVopG(n,cur_vop)==NULL) {
        fprintf(stderr,"MacroblockBasedPadding: incorrect input parameters\n\n");
        return(-1);
      }
    }
  }

  w_size=a_w/MB_SIZE;
  h_size=a_h/MB_SIZE;
  alpha_sub_ima=GetVopAuv(cur_vop);

  /* array with MBM_TRANSPARENT/MBM_OPAQUE/MBM_BOUNDARY modes per MB */
  alpha_modes=(SInt*)malloc(sizeof(SInt)*w_size*h_size);
  subsamp_alpha(alpha,a_w,a_h,0/*MB-based*/,alpha_modes);

  alpha_uv= (SInt*)malloc(sizeof(SInt)*w_size*h_size);
  alpha_sub=(SInt*)GetImageData(alpha_sub_ima);
  subsamp_alpha(alpha_sub,a_w/2,a_h/2,1/*B-based*/,alpha_uv);

  buf = (SInt *)GetImageData(GetVopY(cur_vop));
  for (i=0;i<a_w*a_h;i++)
    if (alpha[i]==0)  buf[i]=default_padding_value;
  buf = (SInt *)GetImageData(GetVopU(cur_vop));
  for (i=0;i<a_w*a_h/4;i++)
    if (alpha_sub[i]==0)  buf[i]=default_padding_value;
  buf = (SInt *)GetImageData(GetVopV(cur_vop));
  for (i=0;i<a_w*a_h/4;i++)
    if (alpha_sub[i]==0)  buf[i]=default_padding_value;


  /* HYUNDAI (Grayscale) */ 
  if(GetVopShape(cur_vop) == GREY_SCALE) {
    for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { /* MAC (SB) 12-Nov-99 */
      buf = (SInt *)GetImageData(GetVopG(n,cur_vop));
      for (i=0;i<a_w*a_h;i++)
        if (alpha[i]==0)  buf[i]=128;
    }
    image_type = 2;
  }
  else
    image_type = 0;


  padding_type=1; /*==repetitive padding */

  /* Padding of boundary blocks */
  for(j=1;j<(h_size-1);j++)
    for(i=1;i<(w_size-1);i++)
      {
      if(alpha_modes[w_size*j+i]==MBM_BOUNDARY)
        BlockPadding(j,i,padding_type, MB_SIZE, prev_rec_alpha,image_type,cur_vop); /* HYUNDAI (Grayscale) */
      if(alpha_uv[w_size*j+i]==MBM_BOUNDARY)
        BlockPadding(j,i,padding_type, B_SIZE, alpha_sub_ima,1,cur_vop);
      }

  /* Blocks surrounded by transparent blocks should be also
     treated as boundary blocks in the next step of block repetiting */
  for(j=1;j<(h_size-1);j++)
    for(i=1;i<(w_size-1);i++)
      {
      if ((alpha_modes[w_size*j+i]==MBM_OPAQUE)&&
          ((alpha_modes[w_size*j+i-1]==MBM_TRANSPARENT)||
          (alpha_modes[w_size*j+i+1]==MBM_TRANSPARENT)||
          (alpha_modes[w_size*(j+1)+i]==MBM_TRANSPARENT)||
          (alpha_modes[w_size*(j-1)+i]==MBM_TRANSPARENT)))
        alpha_modes[w_size*j+i]=MBM_BOUNDARY;
      if ((alpha_uv[w_size*j+i]==MBM_OPAQUE)&&
          ((alpha_uv[w_size*j+i-1]==MBM_TRANSPARENT)||
          (alpha_uv[w_size*j+i+1]==MBM_TRANSPARENT)||
          (alpha_uv[w_size*(j+1)+i]==MBM_TRANSPARENT)||
          (alpha_uv[w_size*(j-1)+i]==MBM_TRANSPARENT)))
        alpha_uv[w_size*j+i]=MBM_BOUNDARY;
      }

  /* priority order for extended padding according to VM8.0. boon */
  for (j=0;j<h_size;j++)
    for (i=0;i<w_size;i++)
      {
      if(alpha_modes[w_size*j+i]==MBM_TRANSPARENT)
        {
        i_ref=j_ref=-1;
        if ((i-1>=0)&&(alpha_modes[w_size*j+i-1]==MBM_BOUNDARY))
          {     /*LEFT*/
          j_ref =j;i_ref =i-1;
          }
        else if ((j-1>=0)&&(alpha_modes[w_size*(j-1)+i]==MBM_BOUNDARY))
          {   /*TOP*/
          j_ref =j-1;i_ref =i;
          }
        else if ((i+1<=w_size-1)&&(alpha_modes[w_size*j+i+1]==MBM_BOUNDARY))
          { /*RIGHT*/
          j_ref =j;i_ref =i+1;
          }
        else if ((j+1<=h_size-1)&&(alpha_modes[w_size*(j+1)+i]==MBM_BOUNDARY))
          { /*BOTTOM*/
          j_ref =j+1;i_ref =i;
          }
        if (i_ref!=-1)
          BlockRepetitivePadding(j,i,j_ref,i_ref,MB_SIZE,
                                 prev_rec_alpha,image_type,cur_vop); /* HYUNDAI (Grayscale) */
        }

      if (alpha_uv[w_size*j+i]==MBM_TRANSPARENT)
        {
        i_ref=j_ref=-1;
        if ((i-1>=0)&&(alpha_uv[w_size*j+i-1]==MBM_BOUNDARY))
          {  /*LEFT*/
          j_ref =j;i_ref =i-1;
          }
        else if ((j-1>=0)&&(alpha_uv[w_size*(j-1)+i]==MBM_BOUNDARY))
          {   /*TOP*/
          j_ref =j-1;i_ref =i;
          }
        else if ((i+1<=w_size-1)&&(alpha_uv[w_size*j+i+1]==MBM_BOUNDARY))
          { /*RIGHT*/
          j_ref =j;i_ref =i+1;
          }
        else if ((j+1<=h_size-1)&&(alpha_uv[w_size*(j+1)+i]==MBM_BOUNDARY))
          { /*BOTTOM*/
          j_ref =j+1;i_ref =i;
          }
        if (i_ref!=-1)
          BlockRepetitivePadding(j,i,j_ref,i_ref,B_SIZE,
                                 alpha_sub_ima,1,cur_vop);
        }

      }

  /* Free alpha_sub_ima  & alpha_uv before return - NO'C, 13/Sep/96 */
  free((Char *)alpha_uv);
  free((Char *)alpha_modes);
  return 1;
}

/***********************************************************CommentBegin******
 *
 * -- TexturePadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      LPE/zero block-based padding of the YUV images
 *
 * Arguments in :
 *      Image *prev_ori_alpha,   previous original alpha plane data
 *      Image *alpha_decisions,  the previous original alpha plane
 *                               subsampled (in block units) which values
 *                               are:
 *                                MBM_TRANSPARENT = all pixels are transparent
 *                                      (outside the shape)
 *                                MBM_OPAQUE      = all pixels are not transp.
 *                                      (inside the shape)
 *                                MBM_BOUNDARY    = transparent and not
 *                                      transparent pixels at the same time
 *                                      (in the boundary of the shape)
 *      Image *MB_decisions,   encoding mode for each MB. If this Image is
 *                                NULL      => Frame based repetitive padd.
 *                                otherwise => Block based LPE and zero padding.
 *
 * Arguments in/out :
 *      Vop   *cur_vop         Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the YUV images of a given VOP
 *      needed in the texture coding.
 *
 *      0) Padding is performed in those MB that lie on the boundary of
 *         the shape, and in some neihbourgs.
 *      1) Padding is performed SEPARATELY FOR EACH of the Y, U and V 8x8
 *         blocks.
 *      2) Padding of the block uses the ORIGINAL ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) Padding of chroma 8x8 block is performed by decimating the 16x16
 *         alpha block (by descarding every other row and column of pixels,
 *         starting from the second row and column).
 *      4) If all the pixels in an 8x8 block are transparent, their values
 *         are replaced by zero (not strictly necessary).
 *      5) There are two cases depending on the MB type:
 *           o intra 8x8 blocks
 *              ==> LPE padding
 *           o residue 8x8 blocks (prediction error after the MC)
 *              ==> padding with zeroes
 *
 * See also :
 *
 *
 * Modified :
 *      05.02.97 A. Pacheco: modified to include the new low pass extrapolation
 *                 padding technique of the VM5.1
 *      05.02.97 A. Pacheco: modified to fill the pixels acording the original
 *                 alpha values (VM5.1)
 *      26.05.97 Minhua Zhou: Removed DoAlphaUVPlane()
 *	981208   U. Benzler (Univ. Hannover): adapted support for grayscale coding 
 *
 ***********************************************************CommentEnd********/

Int
TexturePadding (
   Vop      *rec_vop,  /* <-- previous original alpha plane data */
   Image   *alpha_decisions, /* <-- subsampled (per block) prev_ori_alpha   */
   Image   *MB_decisions,    /* <-- encoding mode for each MB               */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )
{
  Image   *aux_a, *prev_ori_alpha=GetVopA(rec_vop);
  Image   *alpha_sub_ima;
  SInt    *mb_modes;
  SInt    *a_sub,
          *alpha_uv,
          *alpha_sub;
  Int     error,a_w,a_h,i,j,a_sub_pos, mb_pos,padding_type;
  Int     h_size,w_size,uv_w_size;
  Int image_type; /* HYUNDAI (Grayscale) */ 

  /*  HYUNDAI (Grayscale) */
  if(GetVopShape(cur_vop) == GREY_SCALE)
     image_type = 2;
   else
     image_type = 0;




#ifdef _NO_TEXTURE_PADDING_
  return(2);
#endif


  /* choose between frame-based or block based repetitive padding */

  if(MB_decisions==NULL)
    {

    /*
     * FRAME-BASED REPETITIVE PADDING
     */

    /* update alpha plane to pad with */
    aux_a=GetVopA(cur_vop);
    cur_vop->a_chan=GetVopA(rec_vop);

    /* do the repetitive frame-based padding */
    error=RepetitivePadding(0,cur_vop);

    /* update the vop */
    cur_vop->a_chan=aux_a;
    if(error!=1)
      fprintf(stderr,"TexturePadding: error in RepetitivePadding\n");

    return (error);
    }

  /*
   * BLOCK-BASED REPETITIVE PADDING
   */

  /* checkings */
  if (CheckParameters(GetVopA(rec_vop),alpha_decisions,MB_decisions,cur_vop)!=1)
    {
    fprintf(stderr,"TexturePadding: incorrect input parameters\n\n");
    return(-1);
    }

  a_sub=   (SInt*)GetImageData(alpha_decisions);
  mb_modes=(SInt*)GetImageData(MB_decisions);
  a_w=GetImageSizeX(GetVopY(cur_vop));
  a_h=GetImageSizeY(GetVopY(cur_vop));

  alpha_sub=(SInt*)GetImageData(GetVopAuv(rec_vop));

  w_size=GetImageSizeX(GetVopY(cur_vop))/MB_SIZE;
  h_size=GetImageSizeY(GetVopY(cur_vop))/MB_SIZE;
  uv_w_size=a_w/B_SIZE/2;
  alpha_sub_ima= GetVopAuv(rec_vop);
  alpha_uv= (SInt*)malloc(sizeof(SInt)*w_size*h_size);
  alpha_sub=(SInt*)GetImageData(alpha_sub_ima);
  subsamp_alpha(alpha_sub,a_w/2,a_h/2,1/*B-based*/,alpha_uv);

  for(j=0;j<h_size;j++)
    for(i=0;i<w_size;i++)
      {
      mb_pos=w_size*j+i;
      if (mb_modes[mb_pos]==MBM_INTRA)
        padding_type=1;
      else /* MBM_INTER16) || MBM_INTER8 or (MBM_TRANSPARENT||MBM_OUT??*/
        padding_type=0;

      /* Y blocks */
      a_sub_pos=w_size*2*j*2+i*2;
      if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2  ,i*2  ,image_type,cur_vop); /* HYUNDAI (Grayscale) */
      else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
        TextureBlockPadding(j*2,i*2,padding_type,B_SIZE,prev_ori_alpha,
                            image_type,cur_vop); /* HYUNDAI (Grayscale) */
      /* else MBM_OPAQUE, do nothing */

      a_sub_pos++;
      if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2  ,i*2+1,image_type,cur_vop); /* HYUNDAI (Grayscale) */
      else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
        TextureBlockPadding(j*2,i*2+1,padding_type,B_SIZE,prev_ori_alpha,
                            image_type,cur_vop); /* HYUNDAI (Grayscale) */
      /* else MBM_OPAQUE, do nothing */

      a_sub_pos=w_size*2*(2*j+1)+i*2;
      if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2+1,i*2  ,image_type,cur_vop); /* HYUNDAI (Grayscale) */
      else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
        TextureBlockPadding(j*2+1,i*2,padding_type,B_SIZE,prev_ori_alpha,
                            image_type,cur_vop); /* HYUNDAI (Grayscale) */
      /* else MBM_OPAQUE, do nothing */

      a_sub_pos++;
      if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2+1,i*2+1,image_type,cur_vop); /* HYUNDAI (Grayscale) */
      else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
        TextureBlockPadding(j*2+1,i*2+1,padding_type,B_SIZE, prev_ori_alpha,
                            image_type,cur_vop); /* HYUNDAI (Grayscale) */
      /* else MBM_OPAQUE, do nothing */

      /* U-V blocks */
      a_sub_pos=uv_w_size*j+i;
      if(alpha_uv[a_sub_pos]==MBM_TRANSPARENT)
        AllZeroFillBlock(j,i,1,cur_vop);
      else if(alpha_uv[a_sub_pos]==MBM_BOUNDARY)
        /* has alpha_sub*/
        TextureBlockPadding(j,i ,padding_type,B_SIZE, alpha_sub_ima,1,cur_vop);
      /* else MBM_OPAQUE, do nothing */
    }

  /* Free alpha_sub_ima  & alpha_uv before return - NO'C, 13/Sep/96 */
  free((Char *)alpha_uv);

  return 1;
}

/***********************************************************CommentBegin******
 *
 * -- BlockRepetitivePadding --
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      22.10.96
 *
 * Purpose :
 *      Repetitive padding of an image blocks (section 3.3.1 and figs. 2,3
 *      of VM4.0)
 *
 * Arguments in :
 *      Int   v,               vertical coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   h,               horizontal coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   vref,            vertical reference coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   href,            horizontal reference coord. (in block units) of
 *                             the block within the lumi image or chroma images
 *      Int   block_size,      B_SIZE or MB_SIZE
 *      Image *prev_rec_alpha, previous reconstructed alpha plane (Y or
 *                             subsampled UV)
 *      Int   image_type,      0=Y; 1=U and V; 2=Y and G
 *
 * Arguments in/out :
 *      Vop   *cur_vop         Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the repetitive padding of an block (Y or U and V)
 *      of a given VOP. Needed in the estimation/compensation process.
 *
 *      1) Padding is performed for an luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *      2) Padding of the block must use the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) If image_type = 0, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane.
 *         If image_type = 1, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane decimated according to VM.
 *      4) The average of the values of the differents MB/Bs used to pad is
 *         performed in the calling code (here, it is no possible to know how
 *         many reference MB/Bs will infuence the current MB/B).
 *      5) No input checkings are performed!!
 *
 * See also :
 *
 *
 * Modified :
 *      09.05.97 Minhua Zhou: changed operation "+=" to "="
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *
 ***********************************************************CommentEnd********/

Void
BlockRepetitivePadding (
   Int     v,                /* <-- vertical coord. (in block units)        */
   Int     h,                /* <-- horizontal coord. (in block units)      */
   Int     vref,             /* <-- vertical ref. coord. (in block units)   */
   Int     href,             /* <-- horizontal ref. coord. (in block units) */
   Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
   Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
   Int     image_type,       /* <-- 0=Y; 1=U and V; 2=Y and G               */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )

{
  SInt   *y_data,*y_block,*yref_block;
  SInt   *g_data[MAX_MAC], *g_block, *gref_block; /* HYUNDAI (Grayscale) */ 
  SInt   *u_data,*u_block,*uref_block;
  SInt   *v_data,*v_block,*vref_block;
  Int    direction;
  Int    size_w,pos;
  Int    i,j,n;
  Int    aux_comp_count = GetVopAuxCompCount(cur_vop);

  /* prev_rec_alpha has, depending on image_type, the right alpha plane
     subsampled for the UV data or for the Y data */

  /* block-based repetitive padding */

  y_data=(SInt*)GetImageData(GetVopY(cur_vop));
  for(n=0;n<aux_comp_count;n++) /* MAC (SB) 12-Nov-99 */
    g_data[n]=(SInt*)GetImageData(GetVopG(n,cur_vop)); /* HYUNDAI (Grayscale) */ 
  u_data=(SInt*)GetImageData(GetVopU(cur_vop));
  v_data=(SInt*)GetImageData(GetVopV(cur_vop));

  if (vref==v)
    direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
  else /* href==h */
    direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;

  if(image_type==0 || image_type==2)  /* Y */ /* HYUNDAI (Grayscale) */
  {
    size_w=GetImageSizeX(GetVopY(cur_vop));
    
    /* Y repetitive-padding */
    y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    yref_block=(SInt*)LoadArea(y_data,href*block_size,vref*block_size,
                               block_size,block_size,size_w);
    
    if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
    {
      pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
      for(j=0;j<block_size;j++)
        for(i=0;i<block_size;i++)
          y_block[j*block_size+i]=yref_block[pos+i];
    }
    else
    {
      pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
      for(j=0;j<block_size;j++)
        for(i=0;i<block_size;i++)
          y_block[j*block_size+i]=yref_block[j*block_size+pos];
    }
    
    SetArea(y_block,h*block_size,v*block_size,block_size,block_size,
            size_w,y_data);
    
    free((Char*)y_block);
    free((Char*)yref_block);
    
    /* HYUNDAI (Grayscale) */ 
    if(image_type==2) {
      for(n=0;n<aux_comp_count;n++) {/* MAC (SB) 12-Nov-99 */
        g_block=(SInt*)LoadArea(g_data[n],h*block_size,v*block_size,
                                block_size,block_size,size_w);
        gref_block=(SInt*)LoadArea(g_data[n],href*block_size,vref*block_size,
                                   block_size,block_size,size_w);

        if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
        {
          pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
          for(j=0;j<block_size;j++)
            for(i=0;i<block_size;i++)
              g_block[j*block_size+i]=gref_block[pos+i];
        }
        else
        { 
          pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
          for(j=0;j<block_size;j++)
            for(i=0;i<block_size;i++)
              g_block[j*block_size+i]=gref_block[j*block_size+pos];
        }
        
        SetArea(g_block,h*block_size,v*block_size,block_size,block_size,
                size_w,g_data[n]);
      
        free((Char*)g_block);
        free((Char*)gref_block);
      }
    }
  }
  else              /* U-V */
  {
    size_w=GetImageSizeX(GetVopU(cur_vop));
    
    /* U-V repetitive-padding */
    u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    uref_block=(SInt*)LoadArea(u_data,href*block_size,vref*block_size,
                               block_size,block_size, size_w);
    vref_block=(SInt*)LoadArea(v_data,href*block_size,vref*block_size,
                               block_size,block_size, size_w);
    
    if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
    {
      pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
      for(j=0;j<block_size;j++)
        for(i=0;i<block_size;i++)
        {
          u_block[j*block_size+i]=uref_block[pos+i];
          v_block[j*block_size+i]=vref_block[pos+i];
        }
    }
    else
    {
      pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
      for(j=0;j<block_size;j++)
        for(i=0;i<block_size;i++)
        {
          u_block[j*block_size+i]=uref_block[j*block_size+pos];
          v_block[j*block_size+i]=vref_block[j*block_size+pos];
        }
    }
    
    SetArea(u_block,h*block_size,v*block_size,
            block_size,block_size,size_w,u_data);
    SetArea(v_block,h*block_size,v*block_size,
            block_size,block_size,size_w,v_data);
    
    free((Char*)u_block);
    free((Char*)v_block);
    free((Char*)uref_block);
    free((Char*)vref_block);
  }
  
} /* BlockRepetitivePadding */

/***********************************************************CommentBegin******
 *
 * -- BlockPadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Repetitive/zero padding of an 8x8/16x16 Y or UV image blocks
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     padding_type,       0=zeroes padding; 1=repetitive padding
 *      Int     block_size,         B_SIZE or MB_SIZE
 *      Image   *prev_rec_alpha,    previous reconstructed alpha plane (Y or
 *                                  subsampled UV)
 *      Int     image_type,         0=Y; 1=U and v; 2=Y and G
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of an 8x8 block (Y or U and V) of a
 *      given VOP. Needed in the texture coding.
 *
 *      1) Padding is performed for an 8x8/16x16 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *      2) Padding of the block uses the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) If image_type = 0, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane.
 *         If image_type = 1, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane decimated according to VM.
 *      4) If all the pixels in an 8x8 block are transparent, their values
 *         are replaced by zero.
 *      5) Performs two paddings depending on the padding_type parameter:
 *           0   ==> padding with zeroes; for residue 8x8 blocks (prediction
 *                   error after the MC)
 *           1   ==> Repetitive padding as in ME/MC; for intra 8x8 blocks
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      22.10.96 A. Pacheco: modified to allow blocks of 8 and 16 units.
 *                  ATTENTION: if block_size==16 the padding_type must be 1!!
 *                  in the same way, there are combinations of parameters not
 *                  checked because they would be not used at all.
 *      27.07.97 C.S. Boon: Modified for VM8.0 padding -- simplified padding
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *
 ***********************************************************CommentEnd********/

Void
BlockPadding (
   Int     v,                /* <-- vertical coord. (in block units)        */
   Int     h,                /* <-- horizontal coord. (in block units)      */
   Int     padding_type,     /* <-- 0=zeroes padding; 1=repetitive padding  */
   Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
   Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
   Int     image_type,       /* <-- 0=Y; 1=U and V; 2=Y and G               */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )
{
  SInt   *a_data,*a_block;
  SInt   *y_data,*y_block;
  SInt   *g_data[MAX_MAC],*g_block; /* HYUNDAI (Grayscale) */ 
  SInt   *u_data,*u_block;
  SInt   *v_data,*v_block;
  SInt   *status_block;
  Int    size_w,j;

  Int    i,n;
  Byte   *ca_block;

  /* prev_rec_alpha has, depending on image_type, the right alpha plane
     subsampled for the UV data or for the Y data */

  if(padding_type==0) /* Zero pading */
  {
    ZeroFillBlock (v,h,image_type,prev_rec_alpha,cur_vop);
    return;
  }

  /* block-based repetitive padding */

  y_data=(SInt*)GetImageData(GetVopY(cur_vop));
  for(n=0;n<GetVopAuxCompCount(cur_vop);n++) 	/* MAC (SB) 12-Nov-99 */
    g_data[n]=(SInt*)GetImageData(GetVopG(n,cur_vop)); /* HYUNDAI (Grayscale) */ 
  u_data=(SInt*)GetImageData(GetVopU(cur_vop));
  v_data=(SInt*)GetImageData(GetVopV(cur_vop));
  a_data=(SInt*)GetImageData(prev_rec_alpha);

  /* creates and inits status_block */
  if ((status_block=(SInt*)malloc(block_size*block_size*sizeof(SInt)))==NULL)
  {
    fprintf(stderr,"BlockPadding: malloc failed\n");
    return;
  }
  
  for (j=0;j<(block_size*block_size);j++)
    status_block[j]=NO_PAD;

  if(image_type==0 || image_type==2)  /* Y */ /* HYUNDAI (Grayscale) */
  {
    size_w=GetImageSizeX(GetVopY(cur_vop));

    /* Y repetitive-padding */
    y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    
    ca_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);

    for(i=0;i<block_size*block_size;i++)
    {
      ca_block[i]=(Byte)a_block[i];
    }
    
    simpl_pad(y_block,ca_block,block_size,block_size);
    
    free(ca_block);

    SetArea(y_block,h*block_size,v*block_size,
            block_size,block_size,size_w,y_data);

    free((Char*)y_block);
    free((Char*)a_block);

    /* HYUNDAI (Grayscale) */ 
    if(image_type==2)
    {
      for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { 	/* MAC (SB) 12-Nov-99 */

        g_block=(SInt*)LoadArea(g_data[n],h*block_size,v*block_size,
                                block_size,block_size,size_w);
        a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        
        
        ca_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
        
        for(i=0;i<block_size*block_size;i++)
        {
          ca_block[i]=(Byte)a_block[i];
        }
        
        simpl_pad(g_block,ca_block,block_size,block_size);
        
        free(ca_block);
        
        SetArea(g_block,h*block_size,v*block_size,
                block_size,block_size,size_w,g_data[n]);

      }
      free((Char*)g_block);
      free((Char*)a_block);      
    }
    
    free((Char*)status_block);
  }
  else              /* U-V */
  {

    ca_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
    size_w=GetImageSizeX(GetVopU(cur_vop));

    /* U-V repetitive-padding */
    u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    
    for(i=0;i<block_size*block_size;i++)
    {
      ca_block[i]=(Byte)a_block[i];
    }

    simpl_pad(u_block,ca_block,block_size,block_size); /*UV simplif. padding*/
    simpl_pad(v_block,ca_block,block_size,block_size); /*UV simplif. padding*/
    
    free(ca_block);

    SetArea(u_block,h*block_size,v*block_size,
            block_size,block_size,size_w,u_data);
    SetArea(v_block,h*block_size,v*block_size,
            block_size,block_size,size_w,v_data);
    
    free((Char*)a_block);
    free((Char*)u_block);
    free((Char*)v_block);
    free((Char*)status_block);
  }
} /* BlockPadding */

/***********************************************************CommentBegin******
 *
 * -- simpl_pad --
 *
 * Authors :
 *      C. S. Boon, J. Takahashi
 *
 * Created :
 *      31.07.97
 *
 * Purpose :
 *      Simplified padding of an 8x8/16x16 Y or UV image blocks
 *
 * Arguments in :
 *      Byte *org_shape    reference shape for padding
 *      int  width         size of image
 *      int  height        size of image
 *
 * Arguments in/out :
 *      Byte *dt           in-data to be padded, out-padded data
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
simpl_pad(
   SInt   *dt,
   Byte   *org_shape,
   int    width,
   int    height
   )
{
  int    size = width*height;
  int    v, h, adr, prev_adr, tmp_adr, start_adr;
  int    cnt=0, dest;
  SInt   other[2];
  SInt   fill_dt;
  SInt   src[16];
  Byte   *hrep_shp;

  hrep_shp=(Byte *)malloc(sizeof(Byte)*size);

  /* H phase */
  memcpy(hrep_shp, org_shape, size);

  adr = 0;
  for (v=0;v<height;v++)
    {
    tmp_adr = 0;
    prev_adr = -1;
    other[1] = 0;
    while (tmp_adr < width)
      {
      cnt = 0;
      while ( (tmp_adr < width) && (*(hrep_shp+adr+tmp_adr) == 0) )
        {
        cnt++;
        tmp_adr++;
        }

      if ( (cnt > 0) && (tmp_adr < width) )
        {
        start_adr = adr + prev_adr + 1;
        if (other[1])
          {            /* left & right */
          fill_dt = (other[0] + *(dt+adr+tmp_adr) + 1 )>>1;
	  { SInt *i = dt+start_adr, *lim = dt+start_adr+cnt;
            for (; i < lim; i++) { *i = fill_dt; }
          }
          /* memset(dt+start_adr, fill_dt, cnt); */
          memset(hrep_shp+start_adr, 255, cnt);
          }
        else
          {        /* right only */
          fill_dt = *(dt+adr+tmp_adr);
	  { SInt *i = dt+start_adr, *lim = dt+start_adr+cnt;
            for (; i < lim; i++) { *i = fill_dt; }
          }
          /* memset(dt+start_adr, fill_dt, cnt); */
          memset(hrep_shp+start_adr, 255, cnt);
          }
        }

      if (tmp_adr < width)
        {
        cnt = 0;
        other[0] = *(dt + adr + tmp_adr);
        other[1] = *(hrep_shp + adr + tmp_adr);
        prev_adr = tmp_adr;
        }
      tmp_adr++;
      }

    if ((cnt > 0)&&(other[1]))
      {  /* left only */
      start_adr = adr + prev_adr + 1;
      fill_dt = other[0];
	  { SInt *i = dt+start_adr, *lim = dt+start_adr+cnt;
            for (; i < lim; i++) { *i = fill_dt; }
          }
      /* memset(dt+start_adr, fill_dt, cnt); */
      memset(hrep_shp+start_adr, 255, cnt);
      }
    adr += width;
    } /*for*/


  /* V phase */
  adr = 0;
  prev_adr = -1;
  while (adr < size)
    {
    if ( *(hrep_shp+adr) != 0 )
      {
      if (prev_adr < 0)
        {
        prev_adr = 0;
        while (prev_adr < adr)
          {
          memcpy(dt+prev_adr, dt+adr, width * sizeof(SInt));
          prev_adr += width;
          }
        }
      else
        {
        if (prev_adr+width < adr)
          {
          for (h=0;h<width;h++)
            {
            src[h] = (*(dt+prev_adr+h) + *(dt+adr+h) + 1 )/2;
            }
          prev_adr += width;
          while (prev_adr < adr)
            {
            memcpy(dt+prev_adr, src, width * sizeof(SInt));
            prev_adr += width;
            }
          }
        prev_adr = adr;
        }
      }
    adr += width;
    }

  dest = prev_adr + width;
  while (dest < adr)
    {
    memcpy(dt+dest, dt+prev_adr, width * sizeof(SInt));
    dest += width;
    }

  free(hrep_shp);
}

/***********************************************************CommentBegin******
 *
 * -- TextureBlockPadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      LPE/zero padding of an 8x8/16x16 Y or UV image blocks
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     padding_type,       0=zeroes padding; 1=repetitive padding
 *      Int     block_size,         B_SIZE or MB_SIZE
 *      Image   *prev_ori_alpha,    previous original alpha plane (Y or
 *                                  subsampled UV)
 *      Int     image_type,         0=Y; 1=U and v
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of an 8x8 block (Y or U and V) of a
 *      given VOP. Needed in the texture coding.
 *
 *      1) Padding is performed for an 8x8/16x16 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *      2) Padding of the block uses the ORIGINAL ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) If image_type = 0, the prev_rec_alpha must be the previous
 *         original alpha plane.
 *         If image_type = 1, the prev_rec_alpha must be the previous
 *         original alpha plane decimated according to VM.
 *      4) If all the pixels in an 8x8 block are transparent, their values
 *         are replaced by zero.
 *      5) Performs two paddings depending on the padding_type parameter:
 *           0   ==> padding with zeroes; for residue 8x8 blocks (prediction
 *                   error after the MC)
 *           1   ==> Low Pass Extrapolation padding
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      22.10.96 A. Pacheco: modified to allow blocks of 8 and 16 units.
 *               ATTENTION: if block_size==16 the padding_type must be 1!! in
 *               the same way, there are combinations of parameters not checked
 *               because they would be not used at all.
 *      05.02.97 A. Pacheco: modified to include the new low pass extrapolation
 *               padding technique of the VM5.1
 *      05.02.97 A. Pacheco: modified to fill the pixels acording the original
 *               alpha values (VM5.1)
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *
 ***********************************************************CommentEnd********/

Void
TextureBlockPadding (
   Int     v,                /* <-- vertical coord. (in block units)        */
   Int     h,                /* <-- horizontal coord. (in block units)      */
   Int     padding_type,     /* <-- 0=zeroes padding; 1=repetitive padding  */
   Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
   Image   *prev_ori_alpha,  /* <-- previous original alpha plane data */
   Int     image_type,       /* <-- 0=Y; 1=U and V; 2=Y and G               */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )

{
  SInt   *a_data,*a_block;
  SInt   *y_data,*y_block;
  SInt   *g_data[MAX_MAC],*g_block; /* HYUNDAI (Grayscale) */ 
  SInt   *u_data,*u_block;
  SInt   *v_data,*v_block;
  Int    size_w, n;

  /* prev_ori_alpha has, depending on image_type, the right alpha plane
     subsampled for the UV data or for the Y data */

  if(padding_type==0) /* Zero pading */
    {
    ZeroFillBlock (v,h,image_type,prev_ori_alpha,cur_vop);
    return;
    }

  /* block-based repetitive padding */

  y_data=(SInt*)GetImageData(GetVopY(cur_vop));
  for(n=0;n<GetVopAuxCompCount(cur_vop);n++) 	/* MAC (SB) 12-Nov-99 */
    g_data[n]=(SInt*)GetImageData(GetVopG(n,cur_vop)); /* HYUNDAI (Grayscale) */ 
  u_data=(SInt*)GetImageData(GetVopU(cur_vop));
  v_data=(SInt*)GetImageData(GetVopV(cur_vop));
  a_data=(SInt*)GetImageData(prev_ori_alpha);

  if(image_type==0 || image_type==2)  /* Y */ /* HYUNDAI (Grayscale) */
  {
    size_w=GetImageSizeX(GetVopY(cur_vop));
    
    /* Y repetitive-padding */
    y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);

    LPEPaddingYBlock(a_block,y_block);
    
    SetArea(y_block,h*block_size,v*block_size,
            block_size,block_size,size_w,y_data);
    
    free((Char*)y_block);
    free((Char*)a_block);
    
    /* HYUNDAI (Grayscale) */ 
    if(image_type==2)
    {
      for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { 	/* MAC (SB) 12-Nov-99 */
        g_block=(SInt*)LoadArea(g_data[n],h*block_size,v*block_size,
                                block_size,block_size,size_w);
        a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);

        LPEPaddingYBlock(a_block,g_block);

        SetArea(g_block,h*block_size,v*block_size,
                block_size,block_size,size_w,g_data[n]);
      }

      free((Char*)g_block);
      free((Char*)a_block);
    }

  }
  else              /* U-V */
  {
    size_w=GetImageSizeX(GetVopU(cur_vop));

    /* U-V repetitive-padding */
    u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    
    LPEPaddingUVBlock(a_block,u_block,v_block);
    
    SetArea(u_block,h*block_size,v*block_size,
            block_size,block_size,size_w,u_data);
    SetArea(v_block,h*block_size,v*block_size,
            block_size,block_size,size_w,v_data);
    
    free((Char*)a_block);
    free((Char*)u_block);
    free((Char*)v_block);
  }
  
} /* TextureBlockPadding */

/***********************************************************CommentBegin******
 *
 * -- ZeroFillBlock --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Fills an 8x8 block of the Y image or the 8x8 blocks of the U and V
 *      images with zeroes, depending on the prev_ori_alpha
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the image
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the image
 *      Int     image_type,         0=Y; 1=U and V; 2=Y and G
 *      Image   *prev_ori_alpha,    original reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the zeroes filling of an 8x8 block of the Y
 *      image or the 8x8 blocks of the U and V images of a given VOP, depending
 *      on the prev_ori_alpha values.
 *      Needed in the texture coding.
 *
 *      1) Filling is performed for an 8x8 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      05.02.97 A. Pacheco: modified to fill the pixels acording the original
 *               alpha values (VM5.1)
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *
 ***********************************************************CommentEnd********/

Void
ZeroFillBlock (
   Int     v,                /* <-- vertical coord. (in block units)        */
   Int     h,                /* <-- horizontal coord. (in block units)      */
   Int     image_type,       /* <-- 0=Y; 1=U and V; 2=Y and G               */
   Image   *prev_ori_alpha,  /* <-- previous original alpha plane data */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )
{
  SInt   *a_data;
  SInt   *y_data;
  SInt   *g_data; /* HYUNDAI (Grayscale) */ 
  SInt   *u_data;
  SInt   *v_data;
  Int    size_w,pos,j,n;

  if(image_type==0 || image_type==2) /* HYUNDAI (Grayscale) */
  {
    /* Y zero-padding */
    size_w=GetImageSizeX(GetVopY(cur_vop));

    y_data=(SInt*)GetImageData(GetVopY(cur_vop));
    a_data=(SInt*)GetImageData(prev_ori_alpha);

    for(j=0;j<B_SIZE;j++)
    {
      pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
      
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        y_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        y_data[pos]=0;
    }
    
    /* HYUNDAI (Grayscale) */ 
    if(image_type==2) {
      for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { 	/* MAC (SB) 12-Nov-99 */
        g_data=(SInt*)GetImageData(GetVopG(n,cur_vop));
        for(j=0;j<B_SIZE;j++)
        {
          pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0; pos++;
          if(a_data[pos]==0) g_data[pos]=0;
        }
      }  
    }
  }
  else
  {
    /* U-V zero-padding */
    size_w=GetImageSizeX(GetVopU(cur_vop));

    u_data=(SInt*)GetImageData(GetVopU(cur_vop));
    v_data=(SInt*)GetImageData(GetVopV(cur_vop));
    a_data=(SInt*)GetImageData(prev_ori_alpha);

    for(j=0;j<B_SIZE;j++)
      {
      pos=(v*B_SIZE+j)*size_w+h*B_SIZE;

      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;

      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      pos++;
      if(a_data[pos]==0)
        u_data[pos]=v_data[pos]=0;
      }
    }

} /* ZeroFillBlock */

/***********************************************************CommentBegin******
 *
 * -- AllZeroFillBlock --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Fills an 8x8 block of the Y image or the 8x8 blocks of the U and V
 *      images with zeroes.
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the image
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the image
 *      Int     image_type,         0=Y; 1=U and V; 2=Y and G
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the zeroes filling of an 8x8 block of the Y
 *      image or the 8x8 blocks of the U and V images of a given VOP.
 *      Needed in the texture coding.
 *
 *      1) Filling is performed for an 8x8 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *
 *
 ***********************************************************CommentEnd********/

Void
AllZeroFillBlock (
   Int     v,                /* <-- vertical coord. (in block units)        */
   Int     h,                /* <-- horizontal coord. (in block units)      */
   Int     image_type,       /* <-- 0=Y; 1=U and V; 2=Y and G               */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )
{
  SInt   *y_data;
  SInt   *g_data; /* HYUNDAI (Grayscale) */ 
  SInt   *u_data;
  SInt   *v_data;
  Int    size_w,pos,j,n;


  if(image_type==0 || image_type==2) /* HYUNDAI (Grayscale) */
  {
    /* Y zero-padding */
    size_w=GetImageSizeX(GetVopY(cur_vop));

    y_data=(SInt*)GetImageData(GetVopY(cur_vop));
    
    for(j=0;j<B_SIZE;j++)
    {
      pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
      
      y_data[pos++]=0;
      y_data[pos++]=0;
      y_data[pos++]=0;
      y_data[pos++]=0;
      
      y_data[pos++]=0;
      y_data[pos++]=0;
      y_data[pos++]=0;
      y_data[pos  ]=0;
    }
    
    /* HYUNDAI (Grayscale) */ 
    if(image_type==2) {
      for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { 	/* MAC (SB) 12-Nov-99 */
        g_data=(SInt*)GetImageData(GetVopG(n,cur_vop));
        
        for(j=0;j<B_SIZE;j++)
        {
          pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
          g_data[pos++]=0;
          g_data[pos++]=0;
          g_data[pos++]=0;
          g_data[pos++]=0;
          g_data[pos++]=0;
          g_data[pos++]=0;
          g_data[pos++]=0;
          g_data[pos  ]=0;
        }
      }  
    }
  }
  else
  {
    /* U-V zero-padding */
    size_w=GetImageSizeX(GetVopU(cur_vop));
    
    u_data=(SInt*)GetImageData(GetVopU(cur_vop));
    v_data=(SInt*)GetImageData(GetVopV(cur_vop));
    
    for(j=0;j<B_SIZE;j++)
    {
      pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
      
      u_data[pos]=v_data[pos]=0;
      pos++;
      u_data[pos]=v_data[pos]=0;
      pos++;
      u_data[pos]=v_data[pos]=0;
      pos++;
      u_data[pos]=v_data[pos]=0;
      pos++;
      
      u_data[pos]=v_data[pos]=0;
      pos++;
      u_data[pos]=v_data[pos]=0;
      pos++;
      u_data[pos]=v_data[pos]=0;
      pos++;
      u_data[pos]=v_data[pos]=0;
    }
  }
  
} /* AllZeroFillBlock */

/***********************************************************CommentBegin******
 *
 * -- LPEPaddingYBlock --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      5.02.97
 *
 * Purpose :
 *      Fills an 8x8 block of the Y data following the LPE technique
 *
 * Arguments in :
 *      SInt*   a_data,             Alpha block
 *
 * Arguments in/out :
 *      SInt*   y_data,             Y block to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the LPE filling of an 8x8 block of the Y
 *      data. Needed in the texture coding.
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
LPEPaddingYBlock (
   SInt*   a_data,  /* Alpha block                                      */
   SInt*   y_data   /* Y block to pad                                   */
   )
{
  Int    i,j,mean,cont;
  SInt   *a_aux,*y_aux;

  /* compute the block mean */
  mean=cont=0;
  a_aux=a_data;y_aux=y_data;

  for(i=0;i<64;i++,a_data++,y_data++)
    if (*a_data!=0)
      {
      mean+=*y_data;
      cont++;
      }

  if(cont==0)
    {
    fprintf(stdout,"\nERROR: cont==0 and the shape cannot be transparent!\n");
    fflush(stdout);
    }
  else
    {
    mean= mean/cont;
    }

  a_data=a_aux;y_data=y_aux;

  /* assign the mean value to the transparent pixels */
  for(i=0;i<64;i++,a_data++,y_data++)
    if (*a_data==0)
      *y_data=mean;

  a_data=a_aux;y_data=y_aux;

  /* perform the low pass extrapolation operator over the transparent pixels */

  /* first column, first line */
  if(a_data[0]==0)
    y_data[0]=(y_data[8]+y_data[1])/2;
  /* rest, first line */
  for(i=1;i<7;i++)
    if(a_data[i]==0)
      y_data[i]=(y_data[i-1]+y_data[i+1]+y_data[i+8])/3;
  /* last column, first line */
  if(a_data[7]==0)
    y_data[7]=(y_data[6]+y_data[15])/2;

  for(j=1;j<7;j++)
    {
    /* first column */
    if(a_data[j*8]==0)
      y_data[j*8]=(y_data[(j-1)*8]+y_data[j*8+1]+y_data[(j+1)*8])/3;

    for(i=1;i<7;i++)
      if(a_data[j*8+i]==0)
        y_data[j*8+i]=(y_data[(j-1)*8+i]+y_data[j*8+i-1]+y_data[j*8+i+1]+
                       y_data[(j+1)*8+i])/4;

    /* last column */
    if(a_data[j*8+7]==0)
      y_data[j*8+7]=(y_data[(j-1)*8+7]+y_data[j*8+6]+y_data[(j+1)*8+7])/3;
    }

  /* first column, last line*/
  if(a_data[56]==0)
    y_data[56]=(y_data[48]+y_data[57])/2;
  /* rest, last line*/
  for(i=1;i<7;i++)
    if(a_data[56+i]==0)
      y_data[56+i]=(y_data[48+i]+y_data[56+i-1]+y_data[56+i+1])/3;
  /* last column, last line*/
  if(a_data[63]==0)
    y_data[63]=(y_data[55]+y_data[62])/2;

} /* LPEPaddingYBlock */

/***********************************************************CommentBegin******
 *
 * -- LPEPaddingUVBlock --
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      5.02.97
 *
 * Purpose :
 *      Fills two 8x8 blocks of the UV data following the LPE technique
 *
 * Arguments in :
 *      SInt*   a_data,             Alpha block
 *
 * Arguments in/out :
 *      SInt*   u_data,             U block to pad
 *      SInt*   v_data,             V block to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the LPE filling of two 8x8 blocks of the UV
 *      data. Needed in the texture coding.
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
LPEPaddingUVBlock (
   SInt*   a_data,  /* Alpha block                                      */
   SInt*   u_data,  /* U block to pad                                   */
   SInt*   v_data   /* V block to pad                                   */
   )
{
  Int    i,j,mean_u, mean_v,cont;
  SInt   *a_aux,*u_aux,*v_aux;

  /* compute the block mean */
  mean_u=mean_v=cont=0;
  a_aux=a_data;
  u_aux=u_data;
  v_aux=v_data;

  for(i=0;i<64;i++,a_data++,u_data++,v_data++)
    if (*a_data!=0)
      {
      mean_u+=*u_data;
      mean_v+=*v_data;
      cont++;
      }
  if(cont==0)
    {
    fprintf(stdout,"\nERROR: cont==0 and the shape cannot be transparent!\n");
    fflush(stdout);
    }
  else
    {
    mean_u=mean_u/cont;
    mean_v=mean_v/cont;
    }
  a_data=a_aux;
  u_data=u_aux;
  v_data=v_aux;

  /* assign the mean value to the transparent pixels */
  for(i=0;i<64;i++,a_data++,u_data++,v_data++)
    if (*a_data==0)
      {
      *u_data=mean_u;
      *v_data=mean_v;
      }
  a_data=a_aux;
  u_data=u_aux;
  v_data=v_aux;

  /* perform the low pass extrapolation operator over the transparent pixels */

  /* first column, first line */
  if(a_data[0]==0)
    {
    u_data[0]=(u_data[8]+u_data[1])/2;
    v_data[0]=(v_data[8]+v_data[1])/2;
    }
  /* rest, first line */
  for(i=1;i<7;i++)
    if(a_data[i]==0)
      {
      u_data[i]=(u_data[i-1]+u_data[i+1]+u_data[i+8])/3;
      v_data[i]=(v_data[i-1]+v_data[i+1]+v_data[i+8])/3;
      }
  /* last column, first line */
  if(a_data[7]==0)
    {
    u_data[7]=(u_data[6]+u_data[15])/2;
    v_data[7]=(v_data[6]+v_data[15])/2;
    }

  for(j=1;j<7;j++)
    {
    /* first column */
    if(a_data[j*8]==0)
      {
      u_data[j*8]=(u_data[(j-1)*8]+u_data[j*8+1]+u_data[(j+1)*8])/3;
      v_data[j*8]=(v_data[(j-1)*8]+v_data[j*8+1]+v_data[(j+1)*8])/3;
      }

    for(i=1;i<7;i++)
      if(a_data[j*8+i]==0)
        {
        u_data[j*8+i]=(u_data[(j-1)*8+i]+u_data[j*8+i-1]+u_data[j*8+i+1]+
                       u_data[(j+1)*8+i])/4;
        v_data[j*8+i]=(v_data[(j-1)*8+i]+v_data[j*8+i-1]+v_data[j*8+i+1]+
                       v_data[(j+1)*8+i])/4;
        }

    /* last column */
    if(a_data[j*8+7]==0)
      {
      u_data[j*8+7]=(u_data[(j-1)*8+7]+u_data[j*8+6]+u_data[(j+1)*8+7])/3;
      v_data[j*8+7]=(v_data[(j-1)*8+7]+v_data[j*8+6]+v_data[(j+1)*8+7])/3;
      }
    }

  /* first column, last line*/
  if(a_data[56]==0)
    {
    u_data[56]=(u_data[48]+u_data[57])/2;
    v_data[56]=(v_data[48]+v_data[57])/2;
    }
  /* rest, last line*/
  for(i=1;i<7;i++)
    if(a_data[56+i]==0)
      {
      u_data[56+i]=(u_data[48+i]+u_data[56+i-1]+u_data[56+i+1])/3;
      v_data[56+i]=(v_data[48+i]+v_data[56+i-1]+v_data[56+i+1])/3;
      }
  /* last column, last line*/
  if(a_data[63]==0)
    {
    u_data[63]=(u_data[55]+u_data[62])/2;
    v_data[63]=(v_data[55]+v_data[62])/2;
    }

} /* LPEPaddingUVBlock */

/***********************************************************CommentBegin******
 *
 * -- VopPadding --
 *
 * Author :
 *      HHI - Minhua Zhou
 *
 * Created :
 *      26.05.97
 *
 * Purpose :
 *      padding a VOP
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *    14.04.1998:  Minhua Zhou added FreeImage(GetVopShapeMode(curr_vop_padded));
 *	07.05.98 Jong Deuk Kim (HYUNDAI): added RecRepetitivePadding()
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *	20.11.98 U. Benzler (University of Hannover) : Do edge pixel repetition (RecRepetetivePadding) also
 *                  for GetVopShape(curr_vop)!=0
 *	09.29.99 Dae-Sung Cho (Samsung AIT): added support for OBSS
 *
 ***********************************************************CommentEnd********/

Void
VopPadding(
   Vop   *curr_vop
   )
{
  Int   n, edge = 16;
  Vop   *curr_vop_padded;

  curr_vop_padded = AllocVop (GetVopWidth(curr_vop) + 2 * edge,
                              GetVopHeight(curr_vop) + 2 * edge,
                              GetVopAuxCompCount(curr_vop));
  curr_vop_padded->hor_spat_ref=curr_vop->hor_spat_ref-edge;
  curr_vop_padded->ver_spat_ref=curr_vop->ver_spat_ref-edge;

  /* Store width and height as integer multiples of 16 */
  PutVopWidth(GetVopWidth(curr_vop)+2*edge,curr_vop_padded);
  PutVopHeight(GetVopHeight(curr_vop)+2*edge,curr_vop_padded);

  MakeImageEdge(GetVopY(curr_vop), edge, GetVopY(curr_vop_padded));
  if (GetVopShape(curr_vop) == GREY_SCALE) /* U. Benzler, 981208 : do gray scale edge generation only if its really there! */
    for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
      MakeImageEdge(GetVopG(n,curr_vop), edge, GetVopG(n,curr_vop_padded)); /* HYUNDAI (Grayscale) */ 
  MakeImageEdge(GetVopU(curr_vop), edge/2, GetVopU(curr_vop_padded));
  MakeImageEdge(GetVopV(curr_vop), edge/2, GetVopV(curr_vop_padded));
  MakeImageEdge(GetVopA(curr_vop), edge, GetVopA(curr_vop_padded));
  MakeImageEdge(GetVopAuv(curr_vop), edge/2, GetVopAuv(curr_vop_padded));

  curr_vop_padded->arbitrary_shape=GetVopShape(curr_vop);
  PutVopBitsPerPixel(GetVopBitsPerPixel(curr_vop),curr_vop_padded);  

  /*if(GetVopShape(curr_vop)==0)*/ /* UB 981120 >> */
/* HYUNDAI 980507 */
  /*RecRepetitivePadding (edge, curr_vop_padded);
  else
    MacroblockBasedPadding (GetVopA(curr_vop_padded), curr_vop_padded);*/

  if(GetVopShape(curr_vop)!=0)
    MacroblockBasedPadding (GetVopA(curr_vop_padded), curr_vop_padded);

  RecRepetitivePadding (edge, curr_vop_padded); /* UB 981120 : do edge pixel repetition also for arbitrary shaped objects << */

  PutVopA(GetVopA(curr_vop_padded),curr_vop);
  PutVopAuv(GetVopAuv(curr_vop_padded),curr_vop);
  PutVopY(GetVopY(curr_vop_padded),curr_vop);
  for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
    PutVopG(GetVopG(n,curr_vop_padded),n,curr_vop); /* HYUNDAI (Grayscale) */ 
  PutVopU(GetVopU(curr_vop_padded),curr_vop);
  PutVopV(GetVopV(curr_vop_padded),curr_vop);
  FreeImage(GetVopQP(curr_vop_padded));
  for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
    FreeImage(GetVopGLQP(n,curr_vop_padded)); /* HYUNDAI (Grayscale) */ 
  FreeImage(GetVopShapeMode(curr_vop_padded));
/* begin : added for OBSS by Samsung AIT (1999-09-29) */
  FreeImage(GetVopAMotX(curr_vop_padded));
  FreeImage(GetVopAMotY(curr_vop_padded));
  FreeImage(GetVopASamplePos(curr_vop_padded));
/* end : added for OBSS by Samsung AIT (1999-09-29) */
  SfreeVop(curr_vop_padded);
  curr_vop->hor_spat_ref-=edge;
  curr_vop->ver_spat_ref-=edge;
  PutVopWidth(GetVopWidth(curr_vop)+2*edge,curr_vop);
  PutVopHeight(GetVopHeight(curr_vop)+2*edge,curr_vop);
}

/***********************************************************CommentBegin******
 *
 * -- CloneVop_TMP --
 *
 * Author :
 *      HHI - Minhua Zhou
 *
 * Created :
 *      29.05.97
 *
 * Purpose :
 *      Clone a VOP
 *
 * Arguments in :
 *     Vop *curr_vop
 *
 * Arguments in/out
 *
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *      17.04.1998 Minhua Zhou: added shape mode
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding 
 *	06.09.99 Eishi Morimatsu (Fujitsu Labs.): added DRC support
 *	09.29.99 Dae-Sung Cho (Samsung AIT): added support for OBSS
 *
 ***********************************************************CommentEnd********/

Vop *
CloneVop_TMP(
   Vop   *curr_vop
   )
{
  Int    n, edge = 16;
  Vop    *curr_vop_cut;
  SInt   *data_in,*data_out;
  Int    height = GetVopHeight(curr_vop),i,j;
  Int    width  = GetVopWidth(curr_vop);

  curr_vop_cut = AllocVop (GetVopWidth(curr_vop) - 2 * edge,
                           GetVopHeight(curr_vop) - 2 * edge,
                           GetVopAuxCompCount(curr_vop));
  PutVopHorSpatRef(GetVopHorSpatRef(curr_vop)+edge,curr_vop_cut);
  PutVopVerSpatRef(GetVopVerSpatRef(curr_vop)+edge,curr_vop_cut);
  PutVopArbitraryShape(GetVopArbitraryShape(curr_vop),curr_vop_cut); /* added for OBSS by Samsung AIT (1999-09-29) */

  data_in = (SInt *) GetImageData(GetVopA(curr_vop));
  data_out = (SInt *) GetImageData(GetVopA(curr_vop_cut));

  for (i=0;i<height-2*edge;i++)
    for (j=0;j<width-2*edge;j++)
      data_out[i*(width-2*edge)+j] = data_in[(i+edge)*width+j+edge];

  data_in = (SInt *) GetImageData(GetVopY(curr_vop));
  data_out = (SInt *) GetImageData(GetVopY(curr_vop_cut));

  for (i=0;i<height-2*edge;i++)
    for (j=0;j<width-2*edge;j++)
      data_out[i*(width-2*edge)+j] = data_in[(i+edge)*width+j+edge];

  /* HYUNDAI (Grayscale) */ 
  if (GetVopShape(curr_vop) == GREY_SCALE) /* U. Benzler, 981208 : gray scale cloning only if its really there! */
    {
      for(n=0;n<GetVopAuxCompCount(curr_vop);n++) { /* MAC (SB) 12-Nov-99 */
        data_in = (SInt *) GetImageData(GetVopG(n,curr_vop));
        data_out = (SInt *) GetImageData(GetVopG(n,curr_vop_cut));
        for (i=0;i<height-2*edge;i++)
          for (j=0;j<width-2*edge;j++)
            data_out[i*(width-2*edge)+j] = data_in[(i+edge)*width+j+edge];
      }
    }

  width/=2;
  height/=2;

  data_in = (SInt *) GetImageData(GetVopU(curr_vop));
  data_out = (SInt *) GetImageData(GetVopU(curr_vop_cut));

  for (i=0;i<height-edge;i++)
    for (j=0;j<width-edge;j++)
      data_out[i*(width-edge)+j] = data_in[(i+edge/2)*width+j+edge/2];

  data_in = (SInt *) GetImageData(GetVopV(curr_vop));
  data_out = (SInt *) GetImageData(GetVopV(curr_vop_cut));

  for (i=0;i<height-edge;i++)
    for (j=0;j<width-edge;j++)
      data_out[i*(width-edge)+j] = data_in[(i+edge/2)*width+j+edge/2];

  data_in = (SInt *) GetImageData(GetVopAuv(curr_vop));
  data_out = (SInt *) GetImageData(GetVopAuv(curr_vop_cut));

  for (i=0;i<height-edge;i++)
    for (j=0;j<width-edge;j++)
      data_out[i*(width-edge)+j] = data_in[(i+edge/2)*width+j+edge/2];

/* >>> added for DRC by Fujitsu (top)    <<< */
  if(GetVopReducedResolution(curr_vop)) {

  data_in = (SInt *) GetImageData(GetVopQP(curr_vop));
  data_out = (SInt *) GetImageData(GetVopQP(curr_vop_cut));

  for (i=0;i<(height-edge)/16;i++)
      for (j=0;j<(width-edge)/16;j++){
          data_out[2*i*2*(width-edge)/16+2*j] = data_in[i*(width-edge)/16+j];  
          data_out[(2*i+1)*2*(width-edge)/16+2*j] = data_in[i*(width-edge)/16+j];  
          data_out[2*i*2*(width-edge)/16+(2*j+1)] = data_in[i*(width-edge)/16+j];  
          data_out[(2*i+1)*2*(width-edge)/16+(2*j+1)] = data_in[i*(width-edge)/16+j];  
      }
      PutVopReducedResolution(GetVopReducedResolution(curr_vop),curr_vop_cut);
  } else {
/* >>> added for DRC by Fujitsu (bottom) <<< */
  CopyImage(GetVopQP(curr_vop),GetVopQP(curr_vop_cut));
/* >>> added for DRC by Fujitsu (top)    <<< */
  }
/* >>> added for DRC by Fujitsu (bottom) <<< */
  for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
    CopyImage(GetVopGLQP(n,curr_vop),GetVopGLQP(n,curr_vop_cut)); /* HYUNDAI (Grayscale) */ 
  CopyImage(GetVopShapeMode(curr_vop),GetVopShapeMode(curr_vop_cut));
/* begin : added for OBSS by Samsung AIT (1999-09-29) */
  CopyImage(GetVopAMotX(curr_vop),GetVopAMotX(curr_vop_cut));
  CopyImage(GetVopAMotY(curr_vop),GetVopAMotY(curr_vop_cut));
  CopyImage(GetVopASamplePos(curr_vop),GetVopASamplePos(curr_vop_cut));
/* end : added for OBSS by Samsung AIT (1999-09-29) */
  return curr_vop_cut;
}

/***********************************************************CommentBegin******
 *
 * -- PadMB --
 *
 * Author :
 *      Noel Brady
 *
 *
 * Created :
 *      14.11.97
 *
 * Purpose :
 *
 *
 * Arguments in :
 *      Vop     *rec_vop          alpha plane data
 *      Image   *MB_decisions     encoding mode for each MB
 *      Int     i
 *      Int     j
 *
 * Arguments in/out
 *      Vop     *cur_vop          Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      Pads the blocks of a MB. Should be used prior to encoding
 *      the MB.
 *
 *
 * See also :
 *
 *
 * Modified : 981209 U. Benzler (University of Hannover) added greyscale support
 *
 *
 ***********************************************************CommentEnd********/

Void
PadMB(
   Vop     *rec_vop,         /* <-- previous original alpha plane data */
   Image   *MB_decisions,    /* <-- encoding mode for each MB               */
   Vop     *cur_vop,         /* <-> Vop to pad                              */
   Int     i,
   Int     j
   )
{
  Image   *alpha = GetVopA(rec_vop);
  Image   *alpha_uv = GetVopAuv(rec_vop);
  Image   *block = AllocImage(8,8,SHORT_TYPE);
  SInt    *mb_modes = (SInt *)GetImageData(MB_decisions);
  SInt    *p = (SInt *)GetImageData(block);
  Int      w_size = GetImageSizeX(MB_decisions),
           mb_pos,
           padding_type,
           transparent_nb,
           mb_pos_x = i*MB_SIZE,
           mb_pos_y = j*MB_SIZE,
           type,k;
 /* U. Benzler 981209 : added Greyscale support >>*/
  Int image_type; 
  if(GetVopShape(cur_vop) == GREY_SCALE)
     image_type = 2;
   else
     image_type = 0;

 /* U. Benzler 981209 : added Greyscale support <<*/

  mb_pos=w_size*j+i;

  if (mb_modes[mb_pos]==MBM_INTRA)
    padding_type=1;
  else /* MBM_INTER16) || MBM_INTER8 or (MBM_TRANSPARENT||MBM_OUT??*/
    padding_type=0;

  /* Y blocks */

  transparent_nb = 0;
  GetSubImage(alpha,block,mb_pos_x,mb_pos_y);
  for (k=0;k<64;k++)
    if (!p[k]) transparent_nb++;
  if (transparent_nb == 64)
    type = MBM_TRANSPARENT;
  else if (transparent_nb == 0)
    type = MBM_OPAQUE;
  else
    type = MBM_BOUNDARY;

  if(type==MBM_TRANSPARENT)
    AllZeroFillBlock(j*2  ,i*2  ,image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  else if (type==MBM_BOUNDARY)
    TextureBlockPadding(j*2,i*2,padding_type,B_SIZE,alpha,
                        image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  /* else MBM_OPAQUE, do nothing */

  transparent_nb = 0;
  GetSubImage(alpha,block,mb_pos_x+B_SIZE,mb_pos_y);
  for (k=0;k<64;k++)
    if (!p[k]) transparent_nb++;
  if (transparent_nb == 64)
    type = MBM_TRANSPARENT;
  else if (transparent_nb == 0)
    type = MBM_OPAQUE;
  else
    type = MBM_BOUNDARY;

  if(type==MBM_TRANSPARENT)
    AllZeroFillBlock(j*2  ,i*2+1  ,image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  else if (type==MBM_BOUNDARY)
    TextureBlockPadding(j*2,i*2+1,padding_type,B_SIZE,alpha,
                        image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  /* else MBM_OPAQUE, do nothing */

  transparent_nb = 0;
  GetSubImage(alpha,block,mb_pos_x,mb_pos_y+B_SIZE);
  for (k=0;k<64;k++)
    if (!p[k]) transparent_nb++;
  if (transparent_nb == 64)
    type = MBM_TRANSPARENT;
  else if (transparent_nb == 0)
    type = MBM_OPAQUE;
  else
    type = MBM_BOUNDARY;

  if(type==MBM_TRANSPARENT)
    AllZeroFillBlock(j*2+1  ,i*2  ,image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  else if (type==MBM_BOUNDARY)
    TextureBlockPadding(j*2+1,i*2,padding_type,B_SIZE,alpha,
                        image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  /* else MBM_OPAQUE, do nothing */

  transparent_nb = 0;
  GetSubImage(alpha,block,mb_pos_x+B_SIZE,mb_pos_y+B_SIZE);
  for (k=0;k<64;k++)
    if (!p[k]) transparent_nb++;
  if (transparent_nb == 64)
    type = MBM_TRANSPARENT;
  else if (transparent_nb == 0)
    type = MBM_OPAQUE;
  else
    type = MBM_BOUNDARY;

  if(type==MBM_TRANSPARENT)
    AllZeroFillBlock(j*2+1  ,i*2+1  ,image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  else if (type==MBM_BOUNDARY)
    TextureBlockPadding(j*2+1,i*2+1,padding_type,B_SIZE,alpha,
                        image_type,cur_vop);/* U. Benzler 981209 : added Greyscale support <<*/
  /* else MBM_OPAQUE, do nothing */

  /* U-V blocks */
  transparent_nb = 0;
  GetSubImage(alpha_uv,block,mb_pos_x/2,mb_pos_y/2);
  for (k=0;k<64;k++)
    if (!p[k]) transparent_nb++;
  if (transparent_nb == 64)
    type = MBM_TRANSPARENT;
  else if (transparent_nb == 0)
    type = MBM_OPAQUE;
  else
    type = MBM_BOUNDARY;

  if(type==MBM_TRANSPARENT)
    AllZeroFillBlock(j,i,1,cur_vop);
  else if(type==MBM_BOUNDARY)
    /* has alpha_sub*/
    TextureBlockPadding(j,i ,padding_type,B_SIZE, alpha_uv,1,cur_vop);
  /* else MBM_OPAQUE, do nothing */

  FreeImage(block);
  return;
}



/***********************************************************CommentBegin******
 *
 * -- FieldBasedPadding --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Field based ME/MC padding
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :   U. Benzler, 981120 : added gray scale support                  
 *
 *
 ***********************************************************CommentEnd********/
Int
FieldBasedPadding (
   Image   *prev_rec_alpha,/* <-- previous reconstructed alpha plane data  */
   Vop     *cur_vop       /* <-> Vop to pad                                */
   )
{
  Image   *alpha_sub_ima,
          *y_ima, *u_ima, *v_ima;
  SInt    *alpha_modes;
  SInt    *alpha,
          *alpha_uv,
          *alpha_sub,
          *buf,
	  *bufg; /* U. Benzler, 981120 : added gray scale support */
  Int     a_w,a_h,i,j,i_ref,j_ref, padding_type, n;
  Int     h_size,w_size;
  Int     default_padding_value = 1 << (GetVopBitsPerPixel(cur_vop)-1);

#ifdef _NO_PADDING_      
  return(2);
#endif

       
  /*
   * FIELDBLOCK-BASED REPETITIVE PADDING
   */

  /* checkings */
  if ( (cur_vop==NULL) || (prev_rec_alpha==NULL) ||
       ((y_ima=GetVopY(cur_vop))==NULL) || ((u_ima=GetVopU(cur_vop))==NULL) ||
       ((v_ima=GetVopV(cur_vop))==NULL) ||
       ((a_w=GetImageSizeX(GetVopY(cur_vop)))<1)  ||
       ((a_h=GetImageSizeY(GetVopY(cur_vop)))<1)  ||
       ((alpha=(SInt*)GetImageData(prev_rec_alpha))==NULL) ||
       (a_w!=GetImageSizeX(prev_rec_alpha)) ||
       (a_h!=GetImageSizeY(prev_rec_alpha)) ||
       (GetVopShape(cur_vop)==0) )
    {
    fprintf(stderr,"FieldBasedPadding: incorrect input parameters\n\n");
    return(-1);
    }
  w_size=a_w/MB_SIZE;
  h_size=a_h/B_SIZE;

  alpha_sub_ima=GetVopAuv(cur_vop);

  /* array with MBM_TRANSPARENT/MBM_OPAQUE/MBM_BOUNDARY modes per MB */
  alpha_modes=(SInt*)malloc(sizeof(SInt)*w_size*h_size);
  subsamp_field_alpha(alpha,a_w,a_h,0/*MB-based*/,alpha_modes);

  alpha_uv= (SInt*)malloc(sizeof(SInt)*w_size*h_size);          
  alpha_sub=(SInt*)GetImageData(alpha_sub_ima);

  subsamp_field_alpha(alpha_sub,a_w/2,a_h/2,1/*B-based*/,alpha_uv);

  buf = (SInt *)GetImageData(GetVopY(cur_vop)); 
  /* U. Benzler, 981120 : added gray scale support >>*/
  if (GetVopShape(cur_vop) == GREY_SCALE) {
    for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { /* MAC (SB) 12-Nov-99 */
      bufg = (SInt *)GetImageData(GetVopG(n,cur_vop));
      for (i=0;i<a_w*a_h;i++)
        if (alpha[i]==0) {
          buf[i]=default_padding_value;
          bufg[i]=default_padding_value;
        }
    }
  } else {               /* no greyscale */
   for (i=0;i<a_w*a_h;i++)
    if (alpha[i]==0)   buf[i]=default_padding_value;
  }
  /* U. Benzler, 981120 : added gray scale support <<*/
  buf = (SInt *)GetImageData(GetVopU(cur_vop));
  for (i=0;i<a_w*a_h/4;i++)
    if (alpha_sub[i]==0)  buf[i]=default_padding_value;
  buf = (SInt *)GetImageData(GetVopV(cur_vop));
  for (i=0;i<a_w*a_h/4;i++)
    if (alpha_sub[i]==0)  buf[i]=default_padding_value;

  padding_type=1; /*==repetitive padding */
 
  /* Padding of boundary blocks */
  for(j=1;j<(h_size-1);j++)
    for(i=1;i<(w_size-1);i++)
      {
      if(alpha_modes[w_size*j+i]==MBM_BOUNDARY)
        FieldBlockPadding(j,i,padding_type, MB_SIZE, prev_rec_alpha,0,cur_vop);
      if(alpha_uv[w_size*j+i]==MBM_BOUNDARY)
        FieldBlockPadding(j,i,padding_type, B_SIZE, alpha_sub_ima,1,cur_vop);
      }
 
  /* Blocks surrounded by transparent blocks should be also
     treated as boundary blocks in the next step of block repetiting */
  for(j=1;j<(h_size-1);j++)
    for(i=1;i<(w_size-1);i++)
      {
      if ((alpha_modes[w_size*j+i]==MBM_OPAQUE)&&
          ((alpha_modes[w_size*j+i-1]==MBM_TRANSPARENT)||
          (alpha_modes[w_size*j+i+1]==MBM_TRANSPARENT)||
          (alpha_modes[w_size*(j+1)+i]==MBM_TRANSPARENT)||
          (alpha_modes[w_size*(j-1)+i]==MBM_TRANSPARENT)))
        alpha_modes[w_size*j+i]=MBM_BOUNDARY;
      if ((alpha_uv[w_size*j+i]==MBM_OPAQUE)&&
          ((alpha_uv[w_size*j+i-1]==MBM_TRANSPARENT)||
          (alpha_uv[w_size*j+i+1]==MBM_TRANSPARENT)||
          (alpha_uv[w_size*(j+1)+i]==MBM_TRANSPARENT)||
          (alpha_uv[w_size*(j-1)+i]==MBM_TRANSPARENT)))
        alpha_uv[w_size*j+i]=MBM_BOUNDARY;
      }

  /* priority order for extended padding according to VM8.0. boon */
  for (j=0;j<h_size;j++)
    for (i=0;i<w_size;i++)
      {
      if(alpha_modes[w_size*j+i]==MBM_TRANSPARENT)
        {
        i_ref=j_ref=-1;
        if ((i-1>=0)&&(alpha_modes[w_size*j+i-1]==MBM_BOUNDARY))
          {     /*LEFT*/
          j_ref =j;i_ref =i-1;
          }
        else if ((j-1>=0)&&(alpha_modes[w_size*(j-1)+i]==MBM_BOUNDARY))
          {   /*TOP*/
          j_ref =j-1;i_ref =i;
          }
        else if ((i+1<=w_size-1)&&(alpha_modes[w_size*j+i+1]==MBM_BOUNDARY))
          { /*RIGHT*/
          j_ref =j;i_ref =i+1;
          }
        else if ((j+1<=h_size-1)&&(alpha_modes[w_size*(j+1)+i]==MBM_BOUNDARY))
          { /*BOTTOM*/
          j_ref =j+1;i_ref =i;
          }
        if (i_ref!=-1)
          FieldBlockRepetitivePadding(j,i,j_ref,i_ref,MB_SIZE,
                                 prev_rec_alpha,0,cur_vop);
        }
 
      if (alpha_uv[w_size*j+i]==MBM_TRANSPARENT)
        {
        i_ref=j_ref=-1;
        if ((i-1>=0)&&(alpha_uv[w_size*j+i-1]==MBM_BOUNDARY))
          {  /*LEFT*/
          j_ref =j;i_ref =i-1;
          }
        else if ((j-1>=0)&&(alpha_uv[w_size*(j-1)+i]==MBM_BOUNDARY))
          {   /*TOP*/
          j_ref =j-1;i_ref =i;
          }
        else if ((i+1<=w_size-1)&&(alpha_uv[w_size*j+i+1]==MBM_BOUNDARY))
          { /*RIGHT*/
          j_ref =j;i_ref =i+1;
          }
        else if ((j+1<=h_size-1)&&(alpha_uv[w_size*(j+1)+i]==MBM_BOUNDARY))
          { /*BOTTOM*/
          j_ref =j+1;i_ref =i;
          }
        if (i_ref!=-1)
          FieldBlockRepetitivePadding(j,i,j_ref,i_ref,B_SIZE,
                                 alpha_sub_ima,1,cur_vop);
        }
          
      }  

  /* Free alpha_sub_ima  & alpha_uv before return - NO'C, 13/Sep/96 */
  free((Char *)alpha_uv);
  free((Char *)alpha_modes);
  return 1;
}
 
 
/***********************************************************CommentBegin******
 *
 * -- FieldBlockRepetitivePadding --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Field Block Repeatitive Padding
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :   U. Benzler, 981120 : added gray scale support
 *
 *
 ***********************************************************CommentEnd********/
Void
FieldBlockRepetitivePadding (
   Int     v,                /* <-- vertical coord. (in block units)        */
   Int     h,                /* <-- horizontal coord. (in block units)      */
   Int     vref,             /* <-- vertical ref. coord. (in block units)   */
   Int     href,             /* <-- horizontal ref. coord. (in block units) */
   Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
   Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
   Int     image_type,       /* <-- 0=Y; 1=U and V                          */
   Vop     *cur_vop          /* <-> Vop to pad                              */
   )
{ 
  SInt   *y_data,*y_block,*yref_block;
  SInt   *u_data,*u_block,*uref_block;
  SInt   *v_data,*v_block,*vref_block;
  Int    direction;
  Int    size_w,pos;
  Int    n,i,j,v_block_size=block_size/2;

  /* prev_rec_alpha has, depending on image_type, the right alpha plane
     subsampled for the UV data or for the Y data */

  /* block-based repetitive padding */               

  y_data=(SInt*)GetImageData(GetVopY(cur_vop));
  u_data=(SInt*)GetImageData(GetVopU(cur_vop));
  v_data=(SInt*)GetImageData(GetVopV(cur_vop));

  if (vref==v)                                  
    direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
  else /* href==h */
    direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;

  if(image_type==0)  /* Y */                                
  {
    size_w=GetImageSizeX(GetVopY(cur_vop));
     
    /* Y repetitive-padding */
    y_block=(SInt*)LoadArea(y_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);
    yref_block=(SInt*)LoadArea(y_data,href*block_size,vref*v_block_size,
                               block_size,v_block_size,size_w);
    
    /* U. Benzler, 981120 : added gray scale support >>*/
    if (GetVopShape(cur_vop) == GREY_SCALE) {
      for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { /* MAC (SB) 12-Nov-99 */
        u_data=(SInt*)GetImageData(GetVopG(n,cur_vop));
        u_block=(SInt*)LoadArea(u_data,h*block_size,v*v_block_size,
                                block_size,v_block_size,size_w);
        uref_block=(SInt*)LoadArea(u_data,href*block_size,vref*v_block_size,
                                   block_size,v_block_size,size_w);
        if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
        {
          pos=(direction==UPPER_REFERENCE)?block_size*(v_block_size-1):0;
          for(j=0;j<v_block_size;j++)
            for(i=0;i<block_size;i++) {
              y_block[j*block_size+i]=yref_block[pos+i];
              u_block[j*block_size+i]=uref_block[pos+i];
            }
        }
        else
        {
          pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
          for(j=0;j<v_block_size;j++)
            for(i=0;i<block_size;i++) {
              y_block[j*block_size+i]=yref_block[j*block_size+pos];
              u_block[j*block_size+i]=uref_block[j*block_size+pos];
            }
        }
        SetArea(u_block,h*block_size,v*v_block_size,block_size,v_block_size,
                size_w,u_data);
     
        free((Char*)u_block);
        free((Char*)uref_block);
      }     
   
      SetArea(y_block,h*block_size,v*v_block_size,block_size,v_block_size,
              size_w,y_data);
      free((Char*)y_block);
      free((Char*)yref_block);
    } else {
      /* U. Benzler, 981120 : added gray scale support <<*/
      if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
      {
        pos=(direction==UPPER_REFERENCE)?block_size*(v_block_size-1):0;
        for(j=0;j<v_block_size;j++)
          for(i=0;i<block_size;i++)
            y_block[j*block_size+i]=yref_block[pos+i];
      }
      else
      {
        pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
        for(j=0;j<v_block_size;j++)
          for(i=0;i<block_size;i++)
            y_block[j*block_size+i]=yref_block[j*block_size+pos];
      }
      
      SetArea(y_block,h*block_size,v*v_block_size,block_size,v_block_size,
              size_w,y_data);
      
      free((Char*)y_block);
      free((Char*)yref_block);
    } /* U. Benzler, 981120 : added gray scale support <<*/
  }
  else              /* U-V */
  {
    size_w=GetImageSizeX(GetVopU(cur_vop));
     
    /* U-V repetitive-padding */            
    u_block=(SInt*)LoadArea(u_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);
    v_block=(SInt*)LoadArea(v_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);
    uref_block=(SInt*)LoadArea(u_data,href*block_size,vref*v_block_size,
                               block_size,v_block_size, size_w);
    vref_block=(SInt*)LoadArea(v_data,href*block_size,vref*v_block_size,
                               block_size,v_block_size, size_w);
 
    if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
    {
      pos=(direction==UPPER_REFERENCE)?block_size*(v_block_size-1):0;
      for(j=0;j<v_block_size;j++)
        for(i=0;i<block_size;i++)
        {
          u_block[j*block_size+i]=uref_block[pos+i];
          v_block[j*block_size+i]=vref_block[pos+i];
        }
    }
    else
    {
      pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
      for(j=0;j<v_block_size;j++)
        for(i=0;i<block_size;i++)
        {
          u_block[j*block_size+i]=uref_block[j*block_size+pos];
          v_block[j*block_size+i]=vref_block[j*block_size+pos];
        }
    }
 
    SetArea(u_block,h*block_size,v*v_block_size,
            block_size,v_block_size,size_w,u_data);
    SetArea(v_block,h*block_size,v*v_block_size,
            block_size,v_block_size,size_w,v_data);
 
    free((Char*)u_block);
    free((Char*)v_block);
    free((Char*)uref_block);
    free((Char*)vref_block);
  }
 
} /* FieldBlockRepetitivePadding */
 
 
 
/***********************************************************CommentBegin******
 *
 * -- FieldBlockPadding --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Field Block Padding
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :   U. Benzler, 981120 : added gray scale support
 *
 *
 ***********************************************************CommentEnd********/
Void
FieldBlockPadding (
  Int     v,                /* <-- vertical coord. (in block units)        */
  Int     h,                /* <-- horizontal coord. (in block units)      */
  Int     padding_type,     /* <-- 0=zeroes padding; 1=repetitive padding  */
  Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
  Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
  Int     image_type,       /* <-- 0=Y; 1=U and V                          */
  Vop     *cur_vop          /* <-> Vop to pad                              */
  )
{
  SInt   *a_data,*a_block;
  SInt   *y_data,*y_block;
  SInt   *g_block; /* U. Benzler, 981120 : added gray scale support */
  SInt   *u_data,*u_block;
  SInt   *v_data,*v_block;
  SInt   *status_block;
  Int    size_w,j,n;

  Int    i, v_block_size=block_size/2;
  Byte   *ca_block;

  /* prev_rec_alpha has, depending on image_type, the right alpha plane
     subsampled for the UV data or for the Y data */

  if(padding_type==0) /* Zero pading */              
  {
    ZeroFillBlock (v,h,image_type,prev_rec_alpha,cur_vop);
    return;
  }

  /* block-based repetitive padding */
     
  y_data=(SInt*)GetImageData(GetVopY(cur_vop));
/*  g_data=(SInt*)GetImageData(GetVopG(cur_vop)); (SB)*/ /* U. Benzler, 981120 : added gray scale support */
  u_data=(SInt*)GetImageData(GetVopU(cur_vop));
  v_data=(SInt*)GetImageData(GetVopV(cur_vop));
  a_data=(SInt*)GetImageData(prev_rec_alpha);
 
  /* creates and inits status_block */
  if ((status_block=(SInt*)malloc(block_size*v_block_size*sizeof(SInt)))==NULL)
  {
    fprintf(stderr,"BlockPadding: malloc failed\n");
    return;
  }
  for (j=0;j<(block_size*v_block_size);j++)
    status_block[j]=NO_PAD;
 
 
  if(image_type==0)  /* Y */
  {
    size_w=GetImageSizeX(GetVopY(cur_vop));
 
    /* Y repetitive-padding */
    y_block=(SInt*)LoadArea(y_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);

    a_block=(SInt*)LoadArea(a_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);


    ca_block=(Byte *)malloc(sizeof(Byte)*block_size*v_block_size);
 
    for(i=0;i<block_size*v_block_size;i++)
    {
      ca_block[i]=(Byte)a_block[i];
    }

    simpl_pad(y_block,ca_block,block_size,v_block_size);

    if (GetVopShape(cur_vop) == GREY_SCALE) { /* U. Benzler, 981120 : added gray scale support */
      for(n=0;n<GetVopAuxCompCount(cur_vop);n++) { /* MAC (SB) 12-Nov-99 */

        g_block=(SInt*)LoadArea((SInt*)GetImageData(GetVopG(n,cur_vop)),h*block_size,v*v_block_size,
                                block_size,v_block_size,size_w);
        simpl_pad(g_block,ca_block,block_size,v_block_size);
        SetArea(g_block,h*block_size,v*v_block_size,
                block_size,v_block_size,size_w,(SInt*)GetImageData(GetVopG(n,cur_vop)));
        free((Char*)g_block);
      }
    }
    

    free(ca_block);
 
    SetArea(y_block,h*block_size,v*v_block_size,
            block_size,v_block_size,size_w,y_data);
 
    free((Char*)y_block);
    free((Char*)status_block);
    free((Char*)a_block);
  }
  else              /* U-V */
  {
    ca_block=(Byte *)malloc(sizeof(Byte)*block_size*v_block_size);
    size_w=GetImageSizeX(GetVopU(cur_vop));
 
    /* U-V repetitive-padding */
    u_block=(SInt*)LoadArea(u_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);
    v_block=(SInt*)LoadArea(v_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);
    a_block=(SInt*)LoadArea(a_data,h*block_size,v*v_block_size,
                            block_size,v_block_size,size_w);

    for(i=0;i<block_size*v_block_size;i++)
    {
      ca_block[i]=(Byte)a_block[i];
    }

    simpl_pad(u_block,ca_block,block_size,v_block_size); /*UV simplif. padding*/
    simpl_pad(v_block,ca_block,block_size,v_block_size); /*UV simplif. padding*/

    free(ca_block);
 
    SetArea(u_block,h*block_size,v*v_block_size,
            block_size,v_block_size,size_w,u_data);
    SetArea(v_block,h*block_size,v*v_block_size,
            block_size,v_block_size,size_w,v_data);
 
    free((Char*)a_block);
    free((Char*)u_block);
    free((Char*)v_block);
    free((Char*)status_block);
  }
} /* FieldBlockPadding */
 
/***********************************************************CommentBegin******
 *
 * -- VopFieldPadding --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      padding a VOP with field base
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *	20.11.98 U. Benzler (University of Hannover) : Do edge pixel repetition (RecRepetetivePadding) also
 *                  for GetVopShape(curr_vop)!=0
 *	09.29.99 Dae-Sung Cho (Samsung AIT): added support for OBSS
 *
 ***********************************************************CommentEnd********/
Void
VopFieldPadding(
   Vop   *curr_vop
   )
{
  Int   n, edge = 16;
  Vop   *curr_vop_padded;
  Vop   *curr_top_padded;
  Vop   *curr_bot_padded;
 
  curr_vop_padded = AllocVop (GetVopWidth(curr_vop) + 2 * edge,
                              GetVopHeight(curr_vop) + 2 * edge,
                              GetVopAuxCompCount(curr_vop) );
  curr_vop_padded->hor_spat_ref=curr_vop->hor_spat_ref-edge;
  curr_vop_padded->ver_spat_ref=curr_vop->ver_spat_ref-edge;
 
  /* Store width and height as integer multiples of 16 */
  PutVopWidth(GetVopWidth(curr_vop)+2*edge,curr_vop_padded);
  PutVopHeight(GetVopHeight(curr_vop)+2*edge,curr_vop_padded);

  MakeImageEdge(GetVopY(curr_vop), edge, GetVopY(curr_vop_padded));
  if (GetVopShape(curr_vop) == GREY_SCALE) /* U. Benzler, 981208 : do gray scale edge generation only if its really there! */
    for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
      MakeImageEdge(GetVopG(n,curr_vop), edge, GetVopG(n,curr_vop_padded)); /* Gray-Scale, 16.09.98 SB */
  MakeImageEdge(GetVopU(curr_vop), edge/2, GetVopU(curr_vop_padded));
  MakeImageEdge(GetVopV(curr_vop), edge/2, GetVopV(curr_vop_padded));
  MakeImageEdge(GetVopA(curr_vop), edge, GetVopA(curr_vop_padded));
  MakeImageEdge(GetVopAuv(curr_vop), edge/2, GetVopAuv(curr_vop_padded));
 
  curr_vop_padded->arbitrary_shape=GetVopShape(curr_vop);

  curr_top_padded = AllocVop(GetVopWidth(curr_vop_padded),
                             GetVopHeight(curr_vop_padded)/2,
                             GetVopAuxCompCount(curr_vop));
  curr_bot_padded = AllocVop(GetVopWidth(curr_vop_padded),
                             GetVopHeight(curr_vop_padded)/2,
                             GetVopAuxCompCount(curr_vop));
 
  PutVopWidth(GetVopWidth(curr_vop_padded),curr_top_padded);
  PutVopHeight(GetVopHeight(curr_vop_padded)/2,curr_top_padded);
  PutVopInterlaced(GetVopInterlaced(curr_vop), curr_top_padded);
  PutVopBitsPerPixel(GetVopBitsPerPixel(curr_vop), curr_top_padded);
  curr_top_padded->arbitrary_shape=GetVopShape(curr_vop);
 
  PutVopWidth(GetVopWidth(curr_vop_padded),curr_bot_padded);
  PutVopHeight(GetVopHeight(curr_vop_padded)/2,curr_bot_padded);
  PutVopInterlaced(GetVopInterlaced(curr_vop), curr_bot_padded);
  PutVopBitsPerPixel(GetVopBitsPerPixel(curr_vop), curr_bot_padded);
  curr_bot_padded->arbitrary_shape=GetVopShape(curr_vop);
 
  /*
   * Shuffle curr_vop_padded to curr_top_padded and curr_bot_padded.
   */
/*   ShuffleVop(curr_vop_padded, curr_top_padded, curr_bot_padded); */  /* HYUNDAI (jdkim) deleted : 981020 */
   
   
  if(GetVopShape(curr_vop)==0)
  {
/* HYUNDAI (jdkim) deleted : 981020 
    RecRepetitivePadding (edge, curr_top_padded);
    RecRepetitivePadding (edge, curr_bot_padded);
 HYUNDAI (jdkim) end : 981020 */
    RecRepetitivePadding (edge, curr_vop_padded);   /* HYUNDAI (jdkim) 981020 */

  }
  else
  {
  /*
   * Shuffle curr_vop_padded to curr_top_padded and curr_bot_padded.
   */

    ShuffleVop(curr_vop_padded, curr_top_padded, curr_bot_padded);  /* HYUNDAI (jdkim) 981020 */

    FieldBasedPadding (GetVopA(curr_top_padded), curr_top_padded);
    FieldBasedPadding (GetVopA(curr_bot_padded), curr_bot_padded);

  /*
   * Deshuffle ..
   */
    DeshuffleVop(curr_top_padded, curr_bot_padded, curr_vop_padded);  /* HYUNDAI (jdkim) 981020 */
    RecRepetitivePadding (edge, curr_vop_padded);   /* HYUNDAI (jdkim) 981020 */
  }

  /*
   * Deshuffle ..
   */
/*    DeshuffleVop(curr_top_padded, curr_bot_padded, curr_vop_padded);  */  /* HYUNDAI (jdkim) deleted : 981020 */

                                                                   
  PutVopA(GetVopA(curr_vop_padded),curr_vop);
  PutVopAuv(GetVopAuv(curr_vop_padded),curr_vop);
  PutVopY(GetVopY(curr_vop_padded),curr_vop);
  PutVopU(GetVopU(curr_vop_padded),curr_vop);
  PutVopV(GetVopV(curr_vop_padded),curr_vop);
  for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
    PutVopG(GetVopG(n,curr_vop_padded),n,curr_vop);	/* Gray-Scale, 16.09.98 SB */
  FreeImage(GetVopQP(curr_vop_padded));
  for(n=0;n<GetVopAuxCompCount(curr_vop);n++) /* MAC (SB) 12-Nov-99 */
    FreeImage(GetVopGLQP(n,curr_vop_padded));
  FreeImage(GetVopShapeMode(curr_vop_padded));
/* begin : added for OBSS by Samsung AIT (1999-09-29) */
  FreeImage(GetVopAMotX(curr_vop_padded));
  FreeImage(GetVopAMotY(curr_vop_padded));
  FreeImage(GetVopASamplePos(curr_vop_padded));
/* end : added for OBSS by Samsung AIT (1999-09-29) */
  SfreeVop(curr_vop_padded);
  FreeVop(curr_top_padded);
  FreeVop(curr_bot_padded);
  curr_vop->hor_spat_ref-=edge;
  curr_vop->ver_spat_ref-=edge;
  PutVopWidth(GetVopWidth(curr_vop)+2*edge,curr_vop);
  PutVopHeight(GetVopHeight(curr_vop)+2*edge,curr_vop);
}
 
 
 
/***********************************************************CommentBegin******
 *
 * -- subsamp_field_alpha --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Check field block(16x8) mode.
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
Void
subsamp_field_alpha(
  SInt   *alpha,      /* <-- alpha plane data                     */
  Int    br_width,    /* <-- width of the alpha plane data        */
  Int    br_height,   /* <-- height  of the alpha plane data      */
  Int    block_based, /* <-- ==1/0 block/MB-based subsampling     */
  SInt   *alpha_sub   /* --> subsampled alpha data                */
  )
{
  Int   h, v;
  Int   br_h;
  Int   br_w;
  Int   v_size,h_size,cells;

  h_size=(block_based==1)?B_SIZE:MB_SIZE;
  v_size=h_size/2;

  br_h=br_height/v_size;
  br_w=br_width/h_size;
  cells=h_size*v_size;

  for (v=0;v<br_h;v++) 
    for (h=0;h<br_w;h++)
      alpha_sub[v*br_w+h]=0;
 
  for (v=0;v<br_height;v++)
    for (h=0;h<br_width;h++)
      if (alpha[v*br_width+h])
        alpha_sub[(v/v_size)*br_w+(h/h_size)] ++;

  for (v=0;v<br_h;v++)                            
    for (h=0;h<br_w;h++)
      if (alpha_sub[v*br_w+h]==0)
        alpha_sub[v*br_w+h]=MBM_TRANSPARENT;
      else if (alpha_sub[v*br_w+h]==cells)    /* v_size x h_size */
        alpha_sub[v*br_w+h]=MBM_OPAQUE;
      else
        alpha_sub[v*br_w+h]=MBM_BOUNDARY;
}        
 
 
/***********************************************************CommentBegin******
 *
 * -- ShuffleVop --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Shuffle Vop to Top-field Vop and Bottom-field Vop.
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *      Vop *curr_top
 *      Vop *curr_bot
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
Void
ShuffleVop(Vop *curr_vop,
           Vop *curr_top,
           Vop *curr_bot)
{
    SInt  *pCurrY, *pTopY, *pBotY;
    SInt  *pCurrG, *pTopG, *pBotG;  /* added Gray-Level, 16.09.98 SB */
    SInt  *pCurrU, *pTopU, *pBotU;
    SInt  *pCurrV, *pTopV, *pBotV;
    SInt  *pCurrA, *pTopA, *pBotA;
    SInt  *pCurrAuv, *pTopAuv, *pBotAuv;
    Int x,y, l_width, l_height, c_width,c_height, n;

    l_width = GetVopWidth(curr_vop);              
    l_height = GetVopHeight(curr_vop);
    c_width = l_width/2;
    c_height = l_height/2;

    pCurrY = (SInt *)GetImageData(GetVopY(curr_vop));

    pCurrU = (SInt *)GetImageData(GetVopU(curr_vop));
    pCurrV = (SInt *)GetImageData(GetVopV(curr_vop));
    pCurrA = (SInt *)GetImageData(GetVopA(curr_vop));
    pCurrAuv = (SInt *)GetImageData(GetVopAuv(curr_vop));

    pTopY = (SInt *)GetImageData(GetVopY(curr_top));      
   
    pTopU = (SInt *)GetImageData(GetVopU(curr_top));
    pTopV = (SInt *)GetImageData(GetVopV(curr_top));
    pTopA = (SInt *)GetImageData(GetVopA(curr_top));
    pTopAuv = (SInt *)GetImageData(GetVopAuv(curr_top));

    pBotY = (SInt *)GetImageData(GetVopY(curr_bot));
    pBotU = (SInt *)GetImageData(GetVopU(curr_bot));
    pBotV = (SInt *)GetImageData(GetVopV(curr_bot));
    pBotA = (SInt *)GetImageData(GetVopA(curr_bot));
    pBotAuv = (SInt *)GetImageData(GetVopAuv(curr_bot));
 
    for(y=0; y<l_height/2; y++) for(x=0; x<l_width; x++)
    {
        pTopY[y*l_width+x] = pCurrY[2*y*l_width+x];
        pBotY[y*l_width+x] = pCurrY[(2*y+1)*l_width+x];
    }

    /* added Gray-Level, 16.09.98 SB */ 
    if (GetVopShape(curr_vop) == GREY_SCALE) { /* U. Benzler, 981208 : gray scale shuffle only if its really there! */

      for(n=0;n<GetVopAuxCompCount(curr_vop);n++) {/* MAC (SB) 12-Nov-99 */
        pCurrG = (SInt *)GetImageData(GetVopG(n,curr_vop));
        pTopG = (SInt *)GetImageData(GetVopG(n,curr_top));
        pBotG = (SInt *)GetImageData(GetVopG(n,curr_bot));

        for(y=0; y<l_height/2; y++) for(x=0; x<l_width; x++)
	{
	  pTopG[y*l_width+x] = pCurrG[2*y*l_width+x];
	  pBotG[y*l_width+x] = pCurrG[(2*y+1)*l_width+x];
	}
      }
    }

    for(y=0; y<l_height/2; y++) for(x=0; x<l_width; x++)
    {
        pTopA[y*l_width+x] = pCurrA[2*y*l_width+x];
        pBotA[y*l_width+x] = pCurrA[(2*y+1)*l_width+x];
    }
 
    for(y=0; y<c_height/2; y++) for(x=0; x<c_width; x++)
    {
        pTopU[y*c_width+x] = pCurrU[2*y*c_width+x];
        pBotU[y*c_width+x] = pCurrU[(2*y+1)*c_width+x];
    }
 
    for(y=0; y<c_height/2; y++) for(x=0; x<c_width; x++)
    {
        pTopV[y*c_width+x] = pCurrV[2*y*c_width+x];
        pBotV[y*c_width+x] = pCurrV[(2*y+1)*c_width+x];
    }

    for(y=0; y<c_height/2; y++) for(x=0; x<c_width; x++)
    {
        pTopAuv[y*c_width+x] = pCurrAuv[2*y*c_width+x];
        pBotAuv[y*c_width+x] = pCurrAuv[(2*y+1)*c_width+x];
    }
}
 
 
 
 

/***********************************************************CommentBegin******
 *
 * -- DeshuffleVop --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Deshuffle Top-field Vop and Botttom-field Vop to Current Vop.
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_top
 *      Vop *curr_bot
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
Void
DeshuffleVop( Vop *curr_top,
              Vop *curr_bot,
              Vop *curr_vop )
{
    SInt  *pCurrY, *pTopY, *pBotY;
    SInt  *pCurrG, *pTopG, *pBotG;    /* added Gray-Level, 16.09.98 SB */
    SInt  *pCurrU, *pTopU, *pBotU;
    SInt  *pCurrV, *pTopV, *pBotV;
    SInt  *pCurrA, *pTopA, *pBotA;
    SInt  *pCurrAuv, *pTopAuv, *pBotAuv;
    Int x,y, l_width, l_height, c_width,c_height,n;

    l_width = GetVopWidth(curr_vop);              
    l_height = GetVopHeight(curr_vop);
    c_width = l_width/2;
    c_height = l_height/2;

    pCurrY = (SInt *)GetImageData(GetVopY(curr_vop));
    pCurrU = (SInt *)GetImageData(GetVopU(curr_vop));
    pCurrV = (SInt *)GetImageData(GetVopV(curr_vop));
    pCurrA = (SInt *)GetImageData(GetVopA(curr_vop));
    pCurrAuv = (SInt *)GetImageData(GetVopAuv(curr_vop));
 
    pTopY = (SInt *)GetImageData(GetVopY(curr_top));
    pTopU = (SInt *)GetImageData(GetVopU(curr_top));
    pTopV = (SInt *)GetImageData(GetVopV(curr_top));
    pTopA = (SInt *)GetImageData(GetVopA(curr_top));
    pTopAuv = (SInt *)GetImageData(GetVopAuv(curr_top));
 
    pBotY = (SInt *)GetImageData(GetVopY(curr_bot));
    pBotU = (SInt *)GetImageData(GetVopU(curr_bot));
    pBotV = (SInt *)GetImageData(GetVopV(curr_bot));
    pBotA = (SInt *)GetImageData(GetVopA(curr_bot));
    pBotAuv = (SInt *)GetImageData(GetVopAuv(curr_bot));
 
    for(y=0; y<l_height/2; y++) for(x=0; x<l_width; x++)
    {
        pCurrY[2*y*l_width+x] = pTopY[y*l_width+x];
        pCurrY[(2*y+1)*l_width+x] = pBotY[y*l_width+x];
    }
    /* added Gray-Level, 16.09.98 SB */
    if (GetVopShape(curr_vop) == GREY_SCALE) { /* U. Benzler, 981208 : gray scale deshuffle only if its really there! */

      for(n=0;n<GetVopAuxCompCount(curr_vop);n++) {/* MAC (SB) 12-Nov-99 */
        pCurrG = (SInt *)GetImageData(GetVopG(n,curr_vop));
        pTopG = (SInt *)GetImageData(GetVopG(n,curr_top));
        pBotG = (SInt *)GetImageData(GetVopG(n,curr_bot));
        
        for(y=0; y<l_height/2; y++) for(x=0; x<l_width; x++)
        {
          pCurrG[2*y*l_width+x] = pTopG[y*l_width+x];
          pCurrG[(2*y+1)*l_width+x] = pBotG[y*l_width+x];
        }
      }
    }


    for(y=0; y<l_height/2; y++) for(x=0; x<l_width; x++)
    {
        pCurrA[2*y*l_width+x] = pTopA[y*l_width+x];
        pCurrA[(2*y+1)*l_width+x] = pBotA[y*l_width+x];
    }
 
    for(y=0; y<c_height/2; y++) for(x=0; x<c_width; x++)
    {
        pCurrU[2*y*c_width+x] = pTopU[y*c_width+x];
        pCurrU[(2*y+1)*c_width+x] = pBotU[y*c_width+x];
    }
    for(y=0; y<c_height/2; y++) for(x=0; x<c_width; x++)
    {
        pCurrV[2*y*c_width+x] = pTopV[y*c_width+x];
        pCurrV[(2*y+1)*c_width+x] = pBotV[y*c_width+x];
    }

    for(y=0; y<c_height/2; y++) for(x=0; x<c_width; x++)
    {
        pCurrAuv[2*y*c_width+x] = pTopAuv[y*c_width+x];
        pCurrAuv[(2*y+1)*c_width+x] = pBotAuv[y*c_width+x];
    }
}
 
 
 
 
 
/***********************************************************CommentBegin******
 *
 * -- RecRepetitivePadding --
 *
 * Author :
 *      HYUNDAI - Jong Deuk Kim
 *
 * Created :
 *      07.05.98
 *
 * Purpose :
 *      Repetetive Padding for Rectangle shape.
 *
 * Arguments in :
 *      Int edge
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified : 981208 U. Benzler : added greyscale support
 *
 *
 ***********************************************************CommentEnd********/
Void RecRepetitivePadding (
  Int edge,
  Vop *curr_vop)
{
  Int x,y, width, height,field_edge,n;
  SInt left_value,right_value;
 
  SInt *pCurrY, *pCurrG, *pCurrU, *pCurrV;
 
  pCurrY = (SInt *)GetImageData(GetVopY(curr_vop));
  pCurrU = (SInt *)GetImageData(GetVopU(curr_vop));
  pCurrV = (SInt *)GetImageData(GetVopV(curr_vop));
  width =  GetVopWidth(curr_vop);
  height = GetVopHeight(curr_vop);
 
 
  /*if(GetVopInterlaced(curr_vop))
    field_edge = edge >>  1;
    else*/ /* UB 981130 : switch off field based edge pixel repetition, always do it frame based */
  field_edge = edge;
 
 
 
  /*
   *  Horizontal padding...
   */
 
  for( y=field_edge; y<height-field_edge; y++)
  {
    left_value = pCurrY[y*width+edge];
    right_value = pCurrY[y*width+width-edge-1];
    for(x=0; x<edge; x++)
    {
      pCurrY[y*width+x] = left_value;
      pCurrY[y*width+width-edge+x] = right_value;
    }
  }

  if (GetVopShape(curr_vop) == GREY_SCALE) /* U. Benzler, 981208 : added gray scale support */
    for(n=0;n<GetVopAuxCompCount(curr_vop);n++) {/* MAC (SB) 12-Nov-99 */
      pCurrG = (SInt *)GetImageData(GetVopG(n,curr_vop));
      for( y=field_edge; y<height-field_edge; y++)
      {
        left_value = pCurrG[y*width+edge];
        right_value = pCurrG[y*width+width-edge-1];
        for(x=0; x<edge; x++)
        {
          pCurrG[y*width+x] = left_value;
          pCurrG[y*width+width-edge+x] = right_value;
        }
      }
    }

  /*
   *  Vertical padding...
   */
  for(y=0; y<field_edge; y++)
    for(x=0; x<width; x++)
      pCurrY[y*width+x] = pCurrY[field_edge*width+x];
 
  for(y=height-field_edge; y<height; y++)
    for(x=0; x<width; x++)
      pCurrY[y*width+x] = pCurrY[(height-1-field_edge)*width+x];

  if (GetVopShape(curr_vop) == GREY_SCALE) /* U. Benzler, 981208 : added gray scale support */
  {
    for(n=0;n<GetVopAuxCompCount(curr_vop);n++) {/* MAC (SB) 12-Nov-99 */
      pCurrG = (SInt *)GetImageData(GetVopG(n,curr_vop));

      for(y=0; y<field_edge; y++)
        for(x=0; x<width; x++)
          pCurrG[y*width+x] = pCurrG[field_edge*width+x];
      
      for(y=height-field_edge; y<height; y++)
        for(x=0; x<width; x++)
          pCurrG[y*width+x] = pCurrG[(height-1-field_edge)*width+x];
    }
  }
                                                                       
/*  Chrominance padding..... */

  /*                          
   *  Horizontal padding...
   */
  edge >>= 1;
  field_edge >>= 1;
  width >>= 1;
  height >>= 1;
  for( y=field_edge; y<height-field_edge; y++)
  {
    left_value = pCurrU[y*width+edge];
    right_value = pCurrU[y*width+width-edge-1];
    for(x=0; x<edge; x++)
    {
      pCurrU[y*width+x] = left_value;
      pCurrU[y*width+width-edge+x] = right_value;
    }

    left_value = pCurrV[y*width+edge];
    right_value = pCurrV[y*width+width-edge-1];
    for(x=0; x<edge; x++)
    {
      pCurrV[y*width+x] = left_value;
      pCurrV[y*width+width-edge+x] = right_value;
    }
  }
 
  /*
   *  Vertical padding...
   */
  for(y=0; y<field_edge; y++)
  {
    for(x=0; x<width; x++)
    {
      pCurrU[y*width+x] = pCurrU[field_edge*width+x];
      pCurrV[y*width+x] = pCurrV[field_edge*width+x];
    }
  }

  for(y=height-field_edge; y<height; y++)
  {
    for(x=0; x<width; x++)
    {
      pCurrU[y*width+x] = pCurrU[(height-1-field_edge)*width+x];
      pCurrV[y*width+x] = pCurrV[(height-1-field_edge)*width+x];
    }
  }
}

/* 01.04.01: Studio Profile */
/***********************************************************CommentBegin******
 *
 * -- MacroblockBasedPadding_Frame_Studio --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      21.10.96
 *
 * Purpose :
 *      Macroblock-based padding of the Vop images, according to the
 *      VM 5.1
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   Previous reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Vop     *cur_vop,          Vop to pad
 *
 * Arguments out :
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the YUV images of a given VOP
 *      needed in the motion estimation and compensation (see section 3.3.1).
 *
 *      1) The padded Vop is NOT extended 16 pixels in all directions (it is
 *         supposed that this extension has been previously done).
 *      2) Padding is performed SEPARATELY FOR EACH of the Y, U/V bands, but
 *         grouped in MB/B units.
 *      3) Padding of the MB/B uses the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block (input parameter).
 *      4) Padding of chroma 8x8 block is performed by decimating the 16x16
 *         alpha block (by descarding every other row and column of pixels,
 *         starting from the second row and column, with an OR operation
 *         over the alpha pixel values).
 *
 * See also :
 *
 *
 * Modified :
 *      28-01-97 Robert Danielsen: Fix for freeing alpha_modes. Fix by
 *                Noel O'Connor.
 *      06.02.97 Angel Pacheco: the padding of the transparent blocks
 *                 is made looking at the most recently padded boundary block.
 *      17.04.97 Minhua Zhou: rewrote this program
 *      27.07.97 C.S. Boon: Changed to VM8.0 padding -- simplified padding
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding
 *      01.04.01 Osamu Sunohara (SONY): Function for Studio Profile
 *
 ***********************************************************CommentEnd********/
Int
MacroblockBasedPadding_Frame_Studio (
   Vop     *cur_vop       /* <-> Vop to pad                                */
   )
{
  
  Int vopstr = GetVopVopStructure_st(cur_vop);
  Int vop_w = GetVopWidth_st(cur_vop, 0);
  Int vop_h = GetVopHeight_st(cur_vop, 0);
  Int w_mbunit = vop_w/MB_SIZE;
  Int h_mbunit = vop_h/MB_SIZE;
  
  Int binum;
  Image **alpha_ima;
  Image **alpha_uv_ima;
  SInt *alpha_data[2];
  SInt *alpha_uv_data[2];
  SInt *alpha_mode;
  SInt *alpha_uv_mode;
  
  SInt *y_data[2];
  SInt *u_data[2];
  SInt *v_data[2];
  SInt *g_data[MAX_AUX_COMP_NUM][2];

  Int i,j,i_ref,j_ref;
  Int k;
  
  Int chroma_format = GetVopChromaFormat_st(cur_vop);
  Int bits_per_pixel = GetVopBitsPerPixel(cur_vop);
  Int interlaced = GetVopInterlaced(cur_vop);
  
  Int cdivx, cdivy;
  if(chroma_format == CHROMA420) {
    cdivx = 2; cdivy = 2;
  } else if(chroma_format == CHROMA422) {
    cdivx = 2; cdivy = 1;
  } else {
    cdivx = 1; cdivy = 1;
  }
  
  if(vopstr != FRAME_VOP) {
    fprintf(stderr, "EXit\n");
    exit(0);
  }

  binum = 0;
  if(GetVopShape(cur_vop)==GREY_SCALE) {
    for(i=0;i<MAX_AUX_COMP_NUM;i++) {
      if(GetVopAuxCompType_st(cur_vop,i)==AUX_COMP_ALPHA) {
        binum = i;
        break;
      }
    }
  }
  
  alpha_ima = GetVopA2_st(cur_vop, binum);
  alpha_uv_ima = GetVopAuv2_st(cur_vop, binum);
  for(i=0;i<2;i++) {
    alpha_data[i] = (SInt *)GetImageData(alpha_ima[i]);
    alpha_uv_data[i] = (SInt *)GetImageData(alpha_uv_ima[i]);
  }				     
  alpha_mode = (SInt *)malloc(sizeof(SInt)*w_mbunit*h_mbunit);
  alpha_uv_mode = (SInt *)malloc(sizeof(SInt)*w_mbunit*h_mbunit);

  subsamp_alpha_Studio(alpha_data, FRAME_VOP, 0, vop_w, vop_h, 0/*MBbased*/, alpha_mode);
  subsamp_alpha_chroma_Studio(alpha_uv_data, FRAME_VOP, 0, vop_w/cdivx, vop_h/cdivy, 0/*MBbased*/, alpha_uv_mode, chroma_format);

  
  for(i=0;i<2;i++) {
    y_data[i] = (SInt *)GetImageData(GetVopY_st(cur_vop,i));
    u_data[i] = (SInt *)GetImageData(GetVopU_st(cur_vop,i));
    v_data[i] = (SInt *)GetImageData(GetVopV_st(cur_vop,i));
    if(GetVopShape(cur_vop)==GREY_SCALE)
      for(k=0;k<GetVopAuxCompCount_st(cur_vop);k++) {
	g_data[k][i] = (SInt *)GetImageData(GetVopG_st(cur_vop,k,i));
      }
  }
	
  /*Repetitive Padding*/
  for(i=0;i<h_mbunit;i++)
    for(j=0;j<w_mbunit;j++) {
      if(alpha_mode[w_mbunit*i+j] == MBM_BOUNDARY) {

	BlockRepetitivePadding_Frame_Studio(j, i, MB_SIZE, MB_SIZE, vop_w, vop_h, alpha_data, y_data, interlaced, bits_per_pixel);

	if(GetVopShape(cur_vop)==GREY_SCALE)
	  for(k=0;k<GetVopAuxCompCount_st(cur_vop);k++) {
	    BlockRepetitivePadding_Frame_Studio(j, i, MB_SIZE, MB_SIZE, vop_w, vop_h, alpha_data, g_data[k], interlaced, bits_per_pixel);

	  }

      }
      if(alpha_uv_mode[w_mbunit*i+j] == MBM_BOUNDARY) {

	BlockRepetitivePadding_Frame_Studio(j, i, MB_SIZE/cdivx, MB_SIZE/cdivy, vop_w/cdivx, vop_h/cdivy, alpha_uv_data, u_data, interlaced, bits_per_pixel);

	BlockRepetitivePadding_Frame_Studio(j, i, MB_SIZE/cdivx, MB_SIZE/cdivy, vop_w/cdivx, vop_h/cdivy, alpha_uv_data, v_data, interlaced, bits_per_pixel);

      }
    }

  /*Extended Padding*/
  for(i=0;i<h_mbunit;i++)
    for(j=0;j<w_mbunit;j++) {
      if(alpha_mode[i*w_mbunit+j] == MBM_TRANSPARENT) {
	j_ref = i_ref = -1;
	if(j>0 && (alpha_mode[i*w_mbunit+j-1]==MBM_BOUNDARY||alpha_mode[i*w_mbunit+j-1]==MBM_OPAQUE)) {
	  j_ref = j-1;
	  i_ref = i;
	} else if(i>0 && (alpha_mode[(i-1)*w_mbunit+j]==MBM_BOUNDARY||alpha_mode[(i-1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i-1;
	} else if(j<w_mbunit-1 && (alpha_mode[i*w_mbunit+j+1]==MBM_BOUNDARY||alpha_mode[i*w_mbunit+j+1]==MBM_OPAQUE)) {
	  j_ref = j+1;
	  i_ref = i;
	} else if(i<h_mbunit-1 && (alpha_mode[(i+1)*w_mbunit+j]==MBM_BOUNDARY||alpha_mode[(i+1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i+1;
	}

	BlockExtendedPadding_Frame_Studio(j, i, j_ref, i_ref, MB_SIZE, MB_SIZE, vop_w, vop_h, y_data, interlaced, bits_per_pixel);
	if(GetVopShape(cur_vop)==GREY_SCALE)
	  for(k=0;k<GetVopAuxCompCount_st(cur_vop);k++) {
	    BlockExtendedPadding_Frame_Studio(j, i, j_ref, i_ref, MB_SIZE, MB_SIZE, vop_w, vop_h, g_data[k], interlaced, bits_per_pixel);

	  }

      }

      if(alpha_uv_mode[i*w_mbunit+j] == MBM_TRANSPARENT) {
	j_ref = i_ref = -1;
	if(j>0 && (alpha_uv_mode[i*w_mbunit+j-1]==MBM_BOUNDARY||alpha_uv_mode[i*w_mbunit+j-1]==MBM_OPAQUE)) {
	  j_ref = j-1;
	  i_ref = i;
	} else if(i>0 && (alpha_uv_mode[(i-1)*w_mbunit+j]==MBM_BOUNDARY||alpha_uv_mode[(i-1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i-1;
	} else if(j<w_mbunit-1 && (alpha_uv_mode[i*w_mbunit+j+1]==MBM_BOUNDARY||alpha_uv_mode[i*w_mbunit+j+1]==MBM_OPAQUE)) {
	  j_ref = j+1;
	  i_ref = i;
	} else if(i<h_mbunit-1 && (alpha_uv_mode[(i+1)*w_mbunit+j]==MBM_BOUNDARY||alpha_uv_mode[(i+1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i+1;
	}

	BlockExtendedPadding_Frame_Studio(j, i, j_ref, i_ref, MB_SIZE/cdivx, MB_SIZE/cdivy, vop_w/cdivx, vop_h/cdivy, u_data, interlaced, bits_per_pixel);
	BlockExtendedPadding_Frame_Studio(j, i, j_ref, i_ref, MB_SIZE/cdivx, MB_SIZE/cdivy, vop_w/cdivx, vop_h/cdivy, v_data, interlaced, bits_per_pixel);

      }
      
    }
	
  free((Char *)alpha_mode);
  free((Char *)alpha_uv_mode);
}
      
Void
BlockRepetitivePadding_Frame_Studio(Int mb_x, Int mb_y,
				    Int mb_w, Int mb_h,
				    Int vop_w, Int vop_h,
				    SInt *alpha[2],
				    SInt *data[2],
				    Int field,
				    Int bits_per_pixel) {
  Int a_block[256], block[256];
  Byte ca_block[256];
  SInt si_block[256];
  Byte ca_fblock[2][128];
  SInt si_fblock[2][128];
  Int i, j, k;

  Int empty[2];
  
  GetBlockFromFrame_Studio2(alpha[0], alpha[1], vop_w, vop_h,
			    mb_x*mb_w, mb_y*mb_h, mb_w, mb_h, a_block);
  GetBlockFromFrame_Studio2(data[0], data[1], vop_w, vop_h,
			    mb_x*mb_w, mb_y*mb_h, mb_w, mb_h, block);

  if(!field) {
  
    for(i=0;i<mb_w*mb_h;i++) {
      ca_block[i] = (Byte)a_block[i];
      si_block[i] = (SInt)block[i];
    }
    
    simpl_pad_Studio(si_block, ca_block, mb_w, mb_h);
    
    for(i=0;i<mb_w*mb_h;i++)
      block[i] = (Int) si_block[i];
  } else {
    
    for(k=0;k<2;k++)
      for(j=0;j<mb_h/2;j++)
	for(i=0;i<mb_w;i++) {
	  ca_fblock[k][j*mb_w+i] = (Byte)a_block[(2*j+k)*mb_w+i];
	  si_fblock[k][j*mb_w+i] = (SInt)block[(2*j+k)*mb_w+i];
	}

    for(k=0;k<2;k++) {
      empty[k] = 0;
      for(i=0;i<mb_w*mb_h/2;i++) {
	if(ca_fblock[k][i])
	  empty[k]++;
      }
    }
    
    for(k=0;k<2;k++) {
      if(empty[k]==0)
	for(i=0;i<mb_w*mb_h/2;i++)
	  si_fblock[k][i] = 1 << bits_per_pixel-1;
      else if(empty[k]<mb_w*mb_h/2)
	simpl_pad_Studio(si_fblock[k], ca_fblock[k], mb_w, mb_h/2);
    }

    for(j=0;j<mb_h;j++)
      for(i=0;i<mb_w;i++)
	block[j*mb_w+i] = (Int) si_fblock[j%2][(j/2)*mb_w+i];
  }
    
  PutBlockToFrame_Studio2(data[0], data[1], vop_w, vop_h,
			  mb_x*mb_w, mb_y*mb_h, mb_w, mb_h, block);

}

Void
simpl_pad_Studio(SInt *padded, Byte *alpha,
		 Int width, Int height) {
  
  int    size = width*height;
  int    v, h, adr, prev_adr, tmp_adr, start_adr;
  int    cnt=0, dest;
  SInt   other[2];
  SInt   fill_dt;
  SInt   src[16];
  Byte   *hrep_shp;

  hrep_shp=(Byte *)malloc(sizeof(Byte)*size);

  /* H phase */
  memcpy(hrep_shp, alpha, size);

  adr = 0;
  for (v=0;v<height;v++)
    {
    tmp_adr = 0;
    prev_adr = -1;
    other[1] = 0;
    while (tmp_adr < width)
      {
      cnt = 0;
      while ( (tmp_adr < width) && (*(hrep_shp+adr+tmp_adr) == 0) )
        {
        cnt++;
        tmp_adr++;
        }

      if ( (cnt > 0) && (tmp_adr < width) )
        {
        start_adr = adr + prev_adr + 1;
        if (other[1])
          {            /* left & right */
          fill_dt = (other[0] + *(padded+adr+tmp_adr) + 1 )>>1;
	  { SInt *i = padded+start_adr, *lim = padded+start_adr+cnt;
            for (; i < lim; i++) { *i = fill_dt; }
          }
          /* memset(dt+start_adr, fill_dt, cnt); */
          memset(hrep_shp+start_adr, 255, cnt);
          }
        else
          {        /* right only */
          fill_dt = *(padded+adr+tmp_adr);
	  { SInt *i = padded+start_adr, *lim = padded+start_adr+cnt;
            for (; i < lim; i++) { *i = fill_dt; }
          }
          /* memset(dt+start_adr, fill_dt, cnt); */
          memset(hrep_shp+start_adr, 255, cnt);
          }
        }

      if (tmp_adr < width)
        {
        cnt = 0;
        other[0] = *(padded + adr + tmp_adr);
        other[1] = *(hrep_shp + adr + tmp_adr);
        prev_adr = tmp_adr;
        }
      tmp_adr++;
      }

    if ((cnt > 0)&&(other[1]))
      {  /* left only */
      start_adr = adr + prev_adr + 1;
      fill_dt = other[0];
	  { SInt *i = padded+start_adr, *lim = padded+start_adr+cnt;
            for (; i < lim; i++) { *i = fill_dt; }
          }
      /* memset(dt+start_adr, fill_dt, cnt); */
      memset(hrep_shp+start_adr, 255, cnt);
      }
    adr += width;
    } /*for*/


  /* V phase */
  adr = 0;
  prev_adr = -1;
  while (adr < size)
    {
    if ( *(hrep_shp+adr) != 0 )
      {
      if (prev_adr < 0)
        {
        prev_adr = 0;
        while (prev_adr < adr)
          {
          memcpy(padded+prev_adr, padded+adr, width * sizeof(SInt));
          prev_adr += width;
          }
        }
      else
        {
        if (prev_adr+width < adr)
          {
          for (h=0;h<width;h++)
            {
            src[h] = (*(padded+prev_adr+h) + *(padded+adr+h) + 1 )/2;
            }
          prev_adr += width;
          while (prev_adr < adr)
            {
            memcpy(padded+prev_adr, src, width * sizeof(SInt));
            prev_adr += width;
            }
          }
        prev_adr = adr;
        }
    }
    adr += width;
  }

  dest = prev_adr + width;
  while (dest < adr)
    {
    memcpy(padded+dest, padded+prev_adr, width * sizeof(SInt));
    dest += width;
    }

  free(hrep_shp);
}

Void
BlockExtendedPadding_Frame_Studio (Int h,
				   Int v,
				   Int href,
				   Int vref,
				   Int b_w,/*block_size*/
				   Int b_h,/*block*/
				   Int width,/*frame*/
				   Int height,/*frame*/
				   SInt *data[2],
				   Int field,
				   Int bits_per_pixel) {

  Int block[256], ref_block[256];
  Int direction;
  Int pos;
  Int i,j;

  if(vref!=-1) { /*no reference -> an adjacent block*/
  
    if (vref==v)
      direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
    else /* href==h */
      direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;

    GetBlockFromFrame_Studio2(data[0], data[1], width, height,
			      href*b_w, vref*b_h, b_w, b_h, ref_block);
    
    if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
      {
	if(!field) {
	  pos=(direction==UPPER_REFERENCE)?b_w*(b_h-1):0;
	  for(j=0;j<b_h;j++)
	    for(i=0;i<b_w;i++)
	      block[j*b_w+i]=ref_block[pos+i];
	} else {
	  /*top*/
	  pos=(direction==UPPER_REFERENCE)?b_w*(b_h-2):0;
	  for(j=0;j<b_h/2;j++)
	    for(i=0;i<b_w;i++)
	      block[2*j*b_w+i]=ref_block[pos+i];
	  /*bot*/
	  pos=(direction==UPPER_REFERENCE)?b_w*(b_h-1):1;
	  for(j=0;j<b_h/2;j++)
	    for(i=0;i<b_w;i++)
	      block[(2*j+1)*b_w+i]=ref_block[pos+i];
	}
      }
    else
      {
      pos=(direction==LEFT_REFERENCE)?(b_w-1):0;
      for(j=0;j<b_h;j++)
        for(i=0;i<b_w;i++)
          block[j*b_w+i]=ref_block[j*b_w+pos];
      }
    
  } else { /*an remaining block*/

    for(j=0;j<b_h;j++)
      for(i=0;i<b_w;i++)
	block[j*b_w+i]= 1 << bits_per_pixel-1;
  }

  PutBlockToFrame_Studio2(data[0], data[1], width, height,
			  h*b_w, v*b_h, b_w, b_h, block);

} /* BlockExtendedPadding_Frame_Studio */

/*field*/
/***********************************************************CommentBegin******
 *
 * -- MacroblockBasedPadding_Field_Studio --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      21.10.96
 *
 * Purpose :
 *      Macroblock-based padding of the Vop images, according to the
 *      VM 5.1
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   Previous reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Vop     *cur_vop,          Vop to pad
 *
 * Arguments out :
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the YUV images of a given VOP
 *      needed in the motion estimation and compensation (see section 3.3.1).
 *
 *      1) The padded Vop is NOT extended 16 pixels in all directions (it is
 *         supposed that this extension has been previously done).
 *      2) Padding is performed SEPARATELY FOR EACH of the Y, U/V bands, but
 *         grouped in MB/B units.
 *      3) Padding of the MB/B uses the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block (input parameter).
 *      4) Padding of chroma 8x8 block is performed by decimating the 16x16
 *         alpha block (by descarding every other row and column of pixels,
 *         starting from the second row and column, with an OR operation
 *         over the alpha pixel values).
 *
 * See also :
 *
 *
 * Modified :
 *      28-01-97 Robert Danielsen: Fix for freeing alpha_modes. Fix by
 *                Noel O'Connor.
 *      06.02.97 Angel Pacheco: the padding of the transparent blocks
 *                 is made looking at the most recently padded boundary block.
 *      17.04.97 Minhua Zhou: rewrote this program
 *      27.07.97 C.S. Boon: Changed to VM8.0 padding -- simplified padding
 *	21.05.98 Ji Heon Kweon (HYUNDAI): support for grayscale coding
 *      01.04.01 Osamu Sunohara (SONY): Function for Studio Profile
 *
 ***********************************************************CommentEnd********/
Int
MacroblockBasedPadding_Field_Studio2 (Vop *cur_vop, Int half_mb)
{
  
  Int vopstr = GetVopVopStructure_st(cur_vop);
  Int fip = CalcFieldParity_Studio(vopstr,
				   GetVopVopCounter_st(cur_vop),
				   GetVopTopFieldFirst(cur_vop));
  Int vop_w = GetVopWidth_st(cur_vop, fip);
  Int vop_h = GetVopHeight_st(cur_vop, fip);
  Int w_mbunit;
  Int h_mbunit;
  Int h_mb_size;
  Int v_mb_size;
  
  Int binum;
  SInt *alpha_data[2];
  SInt *alpha_uv_data[2];
  SInt *alpha_mode;
  SInt *alpha_uv_mode;
  
  SInt *y_data[2];
  SInt *u_data[2];
  SInt *v_data[2];
  SInt *g_data[MAX_AUX_COMP_NUM][2];

  Int i,j,i_ref,j_ref;
  Int k;
  
  Int chroma_format = GetVopChromaFormat_st(cur_vop);
  Int bits_per_pixel = GetVopBitsPerPixel(cur_vop);
  
  Int cdivx, cdivy;

  if(half_mb) {
    h_mb_size = MB_SIZE;
    v_mb_size = MB_SIZE/2;
  } else {
    h_mb_size = MB_SIZE;
    v_mb_size = MB_SIZE;
  }    
  w_mbunit = vop_w/h_mb_size;
  h_mbunit = vop_h/v_mb_size;

  
  if(chroma_format == CHROMA420) {
    cdivx = 2; cdivy = 2;
  } else if(chroma_format == CHROMA422) {
    cdivx = 2; cdivy = 1;
  } else {
    cdivx = 1; cdivy = 1;
  }
  
  if(vopstr != FIELD_VOP) {
    fprintf(stderr, "EXit\n");
    exit(0);
  }

  binum = 0;
  if(GetVopShape(cur_vop)==GREY_SCALE) {
    for(i=0;i<MAX_AUX_COMP_NUM;i++) {
      if(GetVopAuxCompType_st(cur_vop,i)==AUX_COMP_ALPHA) {
        binum = i;
        break;
      }
    }
  }
  
  alpha_data[fip] = (SInt *)GetImageData(GetVopA_st(cur_vop, binum, fip));
  alpha_uv_data[fip] = (SInt *)GetImageData(GetVopAuv_st(cur_vop, binum, fip));

  alpha_mode = (SInt *)malloc(sizeof(SInt)*w_mbunit*h_mbunit);
  alpha_uv_mode = (SInt *)malloc(sizeof(SInt)*w_mbunit*h_mbunit);

  if(half_mb) {
    subsamp_alpha_Studio2(alpha_data, FIELD_VOP, fip, vop_w, vop_h, 2/*MBbased*/, alpha_mode);
    subsamp_alpha_chroma_Studio2(alpha_uv_data, FIELD_VOP, fip, vop_w/cdivx, vop_h/cdivy, 2/*MBbased*/, alpha_uv_mode, chroma_format);
  } else {
    subsamp_alpha_Studio2(alpha_data, FIELD_VOP, fip, vop_w, vop_h, 0/*MBbased*/, alpha_mode);
    subsamp_alpha_chroma_Studio2(alpha_uv_data, FIELD_VOP, fip, vop_w/cdivx, vop_h/cdivy, 0/*MBbased*/, alpha_uv_mode, chroma_format);
  }    

  
  y_data[fip] = (SInt *)GetImageData(GetVopY_st(cur_vop,fip));
  u_data[fip] = (SInt *)GetImageData(GetVopU_st(cur_vop,fip));
  v_data[fip] = (SInt *)GetImageData(GetVopV_st(cur_vop,fip));
  if(GetVopShape(cur_vop)==GREY_SCALE)
    for(k=0;k<GetVopAuxCompCount_st(cur_vop);k++)
      g_data[k][fip] = (SInt *)GetImageData(GetVopG_st(cur_vop,k,fip));
	
  /*Repetitive Padding*/
  for(i=0;i<h_mbunit;i++)
    for(j=0;j<w_mbunit;j++) {
      if(alpha_mode[w_mbunit*i+j] == MBM_BOUNDARY) {

	BlockRepetitivePadding_Field_Studio(j, i, h_mb_size, v_mb_size, vop_w, vop_h, alpha_data[fip], y_data[fip], bits_per_pixel);

	if(GetVopShape(cur_vop)==GREY_SCALE)
	  for(k=0;k<GetVopAuxCompCount_st(cur_vop);k++)
	    BlockRepetitivePadding_Field_Studio(j, i, h_mb_size, v_mb_size, vop_w, vop_h, alpha_data[fip], g_data[k][fip], bits_per_pixel);

      }
      if(alpha_uv_mode[w_mbunit*i+j] == MBM_BOUNDARY) {

	BlockRepetitivePadding_Field_Studio(j, i, h_mb_size/cdivx, v_mb_size/cdivy, vop_w/cdivx, vop_h/cdivy, alpha_uv_data[fip], u_data[fip], bits_per_pixel);

	BlockRepetitivePadding_Field_Studio(j, i, h_mb_size/cdivx, v_mb_size/cdivy, vop_w/cdivx, vop_h/cdivy, alpha_uv_data[fip], v_data[fip], bits_per_pixel);

      }
    }

  /*Extended Padding*/
  for(i=0;i<h_mbunit;i++)
    for(j=0;j<w_mbunit;j++) {
      if(alpha_mode[i*w_mbunit+j] == MBM_TRANSPARENT) {

	j_ref = i_ref = -1;
	if(j>0 && (alpha_mode[i*w_mbunit+j-1]==MBM_BOUNDARY||alpha_mode[i*w_mbunit+j-1]==MBM_OPAQUE)) {
	  j_ref = j-1;
	  i_ref = i;
	} else if(i>0 && (alpha_mode[(i-1)*w_mbunit+j]==MBM_BOUNDARY||alpha_mode[(i-1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i-1;
	} else if(j<w_mbunit-1 && (alpha_mode[i*w_mbunit+j+1]==MBM_BOUNDARY||alpha_mode[i*w_mbunit+j+1]==MBM_OPAQUE)) {
	  j_ref = j+1;
	  i_ref = i;
	} else if(i<h_mbunit-1 && (alpha_mode[(i+1)*w_mbunit+j]==MBM_BOUNDARY||alpha_mode[(i+1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i+1;
	}


	BlockExtendedPadding_Field_Studio(j, i, j_ref, i_ref, h_mb_size, v_mb_size, vop_w, vop_h, y_data[fip], bits_per_pixel);
	if(GetVopShape(cur_vop)==GREY_SCALE)
	  for(k=0;k<GetVopAuxCompCount_st(cur_vop);k++)
	    BlockExtendedPadding_Field_Studio(j, i, j_ref, i_ref, h_mb_size, v_mb_size, vop_w, vop_h, g_data[k][fip], bits_per_pixel);

      }

      if(alpha_uv_mode[i*w_mbunit+j] == MBM_TRANSPARENT) {
	j_ref = i_ref = -1;
	if(j>0 && (alpha_uv_mode[i*w_mbunit+j-1]==MBM_BOUNDARY||alpha_uv_mode[i*w_mbunit+j-1]==MBM_OPAQUE)) {
	  j_ref = j-1;
	  i_ref = i;
	} else if(i>0 && (alpha_uv_mode[(i-1)*w_mbunit+j]==MBM_BOUNDARY||alpha_uv_mode[(i-1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i-1;
	} else if(j<w_mbunit-1 && (alpha_uv_mode[i*w_mbunit+j+1]==MBM_BOUNDARY||alpha_uv_mode[i*w_mbunit+j+1]==MBM_OPAQUE)) {
	  j_ref = j+1;
	  i_ref = i;
	} else if(i<h_mbunit-1 && (alpha_uv_mode[(i+1)*w_mbunit+j]==MBM_BOUNDARY||alpha_uv_mode[(i+1)*w_mbunit+j]==MBM_OPAQUE)) {
	  j_ref = j;
	  i_ref = i+1;
	}

	BlockExtendedPadding_Field_Studio(j, i, j_ref, i_ref, h_mb_size/cdivx, v_mb_size/cdivy, vop_w/cdivx, vop_h/cdivy, u_data[fip], bits_per_pixel);
	BlockExtendedPadding_Field_Studio(j, i, j_ref, i_ref, h_mb_size/cdivx, v_mb_size/cdivy, vop_w/cdivx, vop_h/cdivy, v_data[fip], bits_per_pixel);

      }
      
    }
	
  free((Char *)alpha_mode);
  free((Char *)alpha_uv_mode);
}

Void
BlockRepetitivePadding_Field_Studio(Int mb_x, Int mb_y,
				    Int mb_w, Int mb_h,
				    Int vop_w, Int vop_h,
				    SInt *alpha,
				    SInt *data,
				    Int bits_per_pixel) {
  Int a_block[256], block[256];
  Byte ca_block[256];
  SInt si_block[256];
  Byte ca_fblock[2][128];
  SInt si_fblock[2][128];
  Int i, j, k;

  Int empty[2];
  
  GetBlockFromField_Studio2(alpha, vop_w, vop_h,
			    mb_x*mb_w, mb_y*mb_h, mb_w, mb_h, a_block);
  GetBlockFromField_Studio2(data, vop_w, vop_h,
			    mb_x*mb_w, mb_y*mb_h, mb_w, mb_h, block);

  for(i=0;i<mb_w*mb_h;i++) {
    ca_block[i] = (Byte)a_block[i];
    si_block[i] = (SInt)block[i];
  }
  
  simpl_pad_Studio(si_block, ca_block, mb_w, mb_h);
  
  for(i=0;i<mb_w*mb_h;i++)
    block[i] = (Int) si_block[i];
  
  PutBlockToField_Studio2(data, vop_w, vop_h,
			  mb_x*mb_w, mb_y*mb_h, mb_w, mb_h, block);

}

Void
BlockExtendedPadding_Field_Studio (Int h,
				   Int v,
				   Int href,
				   Int vref,
				   Int b_w,/*block_size*/
				   Int b_h,/*block*/
				   Int width,/*frame*/
				   Int height,/*frame*/
				   SInt *data,
				   Int bits_per_pixel) {

  Int block[256], ref_block[256];
  Int direction;
  Int pos;
  Int i,j;

  if(vref!=-1) { /*no reference -> an adjacent block*/
  
    if (vref==v)
      direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
    else /* href==h */
      direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;

    GetBlockFromField_Studio2(data, width, height,
			      href*b_w, vref*b_h, b_w, b_h, ref_block);
    
    if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
      {
	pos=(direction==UPPER_REFERENCE)?b_w*(b_h-1):0;
	for(j=0;j<b_h;j++)
	  for(i=0;i<b_w;i++)
	    block[j*b_w+i]=ref_block[pos+i];
      }
    else
      {
      pos=(direction==LEFT_REFERENCE)?(b_w-1):0;
      for(j=0;j<b_h;j++)
        for(i=0;i<b_w;i++)
          block[j*b_w+i]=ref_block[j*b_w+pos];
      }
    
  } else { /*an remaining block*/

    for(j=0;j<b_h;j++)
      for(i=0;i<b_w;i++)
	block[j*b_w+i]= 1 << bits_per_pixel-1;
  }

  PutBlockToField_Studio2(data, width, height,
			  h*b_w, v*b_h, b_w, b_h, block);

} /* BlockExtendedPadding_Frame_Studio */
/*field*/

Void
FrameVopPadding_Studio(Vop *curr_vop)
{
  Int   edge = 16;
  Vop   *curr_vop_padded;

  int orig_width  = GetVopWidth_st (curr_vop,0);
  int orig_height = GetVopHeight_st(curr_vop,0);
  int chroma_format = GetVopChromaFormat_st(curr_vop);
  int vop_structure = GetVopVopStructure_st(curr_vop);
  int cdivx = 1, cdivy = 1;

  int i,j;
  
  if(vop_structure != FRAME_VOP) {
    fprintf(stderr, "\nFrameVopPadding\n");
    exit(0);
  }

  if(chroma_format==CHROMA422)
    cdivx <<= 1;
  else if(chroma_format==CHROMA420) {
    cdivx <<= 1;
    cdivy <<= 1;
  }

  curr_vop_padded = AllocFrameVop_Studio();
  AllocVopChannels_Studio(curr_vop_padded, FRAME_VOP, 0,
			  orig_width +2*edge,
			  orig_height+2*edge,
			  chroma_format);
  curr_vop_padded->hor_spat_ref=curr_vop->hor_spat_ref-edge;
  curr_vop_padded->ver_spat_ref=curr_vop->ver_spat_ref-edge;

  /* Store width and height as integer multiples of 16 */
  PutVopWidth_st (orig_width +2*edge, curr_vop_padded, 0);
  PutVopHeight_st(orig_height+2*edge, curr_vop_padded, 0);

  MakeImageEdge_Studio(GetVopY2_st(curr_vop),
		       FRAME_VOP, 0,
		       orig_width, orig_height, edge, edge,
		       GetVopY2_st(curr_vop_padded));
  if (GetVopShape(curr_vop) == GREY_SCALE) {
    for(i=0; i<MAX_AUX_COMP_NUM; i++)
      MakeImageEdge_Studio(GetVopG2_st(curr_vop,i),
			   FRAME_VOP, 0,
			   orig_width, orig_height, edge, edge,
			   GetVopG2_st(curr_vop_padded,i));
  }
  MakeImageEdge_Studio(GetVopU2_st(curr_vop),
		       FRAME_VOP, 0,
		       orig_width/cdivx, orig_height/cdivy, edge/cdivx, edge/cdivy,
		       GetVopU2_st(curr_vop_padded));
  MakeImageEdge_Studio(GetVopV2_st(curr_vop),
		       FRAME_VOP, 0,
		       orig_width/cdivx, orig_height/cdivy, edge/cdivx, edge/cdivy,
		       GetVopV2_st(curr_vop_padded));
  for(i=0; i<MAX_AUX_COMP_NUM; i++) {
    MakeImageEdge_Studio(GetVopA2_st(curr_vop,i),
			 FRAME_VOP, 0,
			 orig_width, orig_height, edge, edge,
			 GetVopA2_st(curr_vop_padded,i));
    MakeImageEdge_Studio(GetVopAuv2_st(curr_vop,i),
			 FRAME_VOP, 0,
			 orig_width/cdivx, orig_height/cdivy, edge/cdivx, edge/cdivy,
			 GetVopAuv2_st(curr_vop_padded,i));
  }
    
  curr_vop_padded->arbitrary_shape=GetVopShape(curr_vop);
  PutVopBitsPerPixel(GetVopBitsPerPixel(curr_vop),curr_vop_padded);  
  for(i=0; i<MAX_AUX_COMP_NUM; i++)
    PutVopGrayBitsPerPixel_st(GetVopGrayBitsPerPixel_st(curr_vop,i),curr_vop_padded,i);  
  
#if 0
  if(GetVopShape(curr_vop)!=0)
    MacroblockBasedPadding (GetVopA(curr_vop_padded), curr_vop_padded);
#endif 0
  
  RecRepetitivePadding_Studio (curr_vop_padded, edge);

  for(i=0; i<2; i++) {
    for(j=0; j<MAX_AUX_COMP_NUM; j++) {
      PutVopA_st(GetVopA_st(curr_vop_padded,j,i),curr_vop,j,i);
      PutVopAuv_st(GetVopAuv_st(curr_vop_padded,j,i),curr_vop,j,i);
      PutVopG_st(GetVopG_st(curr_vop_padded,j,i),curr_vop,j,i);
    }
    PutVopY_st(GetVopY_st(curr_vop_padded,i),curr_vop,i);
    PutVopU_st(GetVopU_st(curr_vop_padded,i),curr_vop,i);
    PutVopV_st(GetVopV_st(curr_vop_padded,i),curr_vop,i);
    FreeImage(GetVopQP_st(curr_vop_padded,i));
    FreeImage(GetVopGLQP_st(curr_vop_padded,0,i));
    FreeImage(GetVopShapeMode_st(curr_vop_padded,i));
  }
  SfreeVop(curr_vop_padded);
  curr_vop->hor_spat_ref_st[0] -= edge;
  curr_vop->ver_spat_ref_st[0] -= edge;
  PutVopWidth_st (GetVopWidth_st (curr_vop,0)+2*edge,curr_vop,0);
  PutVopHeight_st(GetVopHeight_st(curr_vop,0)+2*edge,curr_vop,0);
}

Void RecRepetitivePadding_Studio(Vop *curr_vop, int edge)
{
  Int x,y, width, height,field_edge;
  SInt left_value,right_value;
  SInt top_value, bottom_value;
  
  SInt *pCurrY[2], *pCurrG[2], *pCurrU[2], *pCurrV[2];
  
  int vop_structure = GetVopVopStructure_st(curr_vop);
  int field_parity = CalcFieldParity_Studio(vop_structure, GetVopVopCounter_st(curr_vop),
					    GetVopTopFieldFirst(curr_vop));
  int chroma_format = GetVopChromaFormat_st(curr_vop);
  int cdivx=1, cdivy=1;
  int i;
  
  if(chroma_format==CHROMA422)
    cdivx <<= 1;
  else if(chroma_format==CHROMA420) {
    cdivx <<= 1;
    cdivy <<= 1;
  }
    

    for(i=0; i<2; i++) {
      pCurrY[i] = (SInt *)GetImageData(GetVopY_st(curr_vop,i));
      pCurrG[i] = (SInt *)GetImageData(GetVopG_st(curr_vop,0,i));
      pCurrU[i] = (SInt *)GetImageData(GetVopU_st(curr_vop,i));
      pCurrV[i] = (SInt *)GetImageData(GetVopV_st(curr_vop,i));
    }
    width  = GetVopWidth_st (curr_vop,field_parity);
    height = GetVopHeight_st(curr_vop,field_parity);
 
 
    /*if(GetVopInterlaced(curr_vop))
        field_edge = edge >>  1;
	else*/ /* UB 981130 : switch off field based edge pixel repetition, always do it frame based */
        field_edge = edge;
 
    /*
     *  Horizontal padding...
     */
 
    for( y=field_edge; y<height-field_edge; y++)
      {
	left_value  = GetImageDataFromFieldMemory(pCurrY, edge, y, width,
						 vop_structure, field_parity);
	right_value = GetImageDataFromFieldMemory(pCurrY, width-edge-1, y, width,
						  vop_structure, field_parity);
	for(x=0; x<edge; x++)
	  {
	    PutImageDataToFieldMemory(pCurrY, x, y, width,
				      vop_structure, field_parity,
				      left_value);
	    PutImageDataToFieldMemory(pCurrY, width-edge+x, y, width,
				      vop_structure, field_parity,
				      right_value);
	  }
      }
    
    if (GetVopShape(curr_vop) == GREY_SCALE)
      for( y=field_edge; y<height-field_edge; y++)
	{
	  left_value  = GetImageDataFromFieldMemory(pCurrG, edge, y, width,
						    vop_structure, field_parity);
	  right_value = GetImageDataFromFieldMemory(pCurrG, width-edge-1, y, width,
						    vop_structure, field_parity);
	  for(x=0; x<edge; x++)
	    {
	      PutImageDataToFieldMemory(pCurrG, x, y, width,
					vop_structure, field_parity,
					left_value);
	      PutImageDataToFieldMemory(pCurrG, width-edge+x, y, width,
					vop_structure, field_parity,
					right_value);
	    }
	  /*
	  left_value = pCurrG[y*width+edge];
	  right_value = pCurrG[y*width+width-edge-1];
	  for(x=0; x<edge; x++)
	  {
	  pCurrG[y*width+x] = left_value;
	  pCurrG[y*width+width-edge+x] = right_value;
	  }
	  */
	}
 
    /*
     *  Vertical padding...
     */
    for(y=0; y<field_edge; y++)
      for(x=0; x<width; x++) {
	top_value  = GetImageDataFromFieldMemory(pCurrY, x, field_edge, width,
						  vop_structure, field_parity);
	PutImageDataToFieldMemory(pCurrY, x, y, width,
				  vop_structure, field_parity,
				  top_value);
      }
 
    for(y=height-field_edge; y<height; y++)
      for(x=0; x<width; x++) {
	bottom_value  = GetImageDataFromFieldMemory(pCurrY, x, height-field_edge-1, width,
						    vop_structure, field_parity);
	PutImageDataToFieldMemory(pCurrY, x, y, width,
				  vop_structure, field_parity,
				  bottom_value);
      }

    if (GetVopShape(curr_vop) == GREY_SCALE)
      {
	for(y=0; y<field_edge; y++)
	  for(x=0; x<width; x++) {
	    top_value  = GetImageDataFromFieldMemory(pCurrG, x, field_edge, width,
						     vop_structure, field_parity);
	    PutImageDataToFieldMemory(pCurrG, x, y, width,
				      vop_structure, field_parity,
				      top_value);
	  }
	
	for(y=height-field_edge; y<height; y++)
	  for(x=0; x<width; x++) {
	    bottom_value  = GetImageDataFromFieldMemory(pCurrG, x, height-field_edge-1, width,
							vop_structure, field_parity);
	    PutImageDataToFieldMemory(pCurrG, x, y, width,
				      vop_structure, field_parity,
				      bottom_value);
	  }
      }
    
/*  Chrominance padding..... */

    /*                          
     *  Horizontal padding...
     */
    edge /= cdivx;
    field_edge /= cdivy;
    width /= cdivx;
    height /= cdivy;
    
    for( y=field_edge; y<height-field_edge; y++)
      {
	left_value  = GetImageDataFromFieldMemory(pCurrU, edge, y, width,
						 vop_structure, field_parity);
	right_value = GetImageDataFromFieldMemory(pCurrU, width-edge-1, y, width,
						  vop_structure, field_parity);
	for(x=0; x<edge; x++)
	  {
	    PutImageDataToFieldMemory(pCurrU, x, y, width,
				      vop_structure, field_parity,
				      left_value);
	    PutImageDataToFieldMemory(pCurrU, width-edge+x, y, width,
				      vop_structure, field_parity,
				      right_value);
	  }
	
	left_value  = GetImageDataFromFieldMemory(pCurrV, edge, y, width,
						 vop_structure, field_parity);
	right_value = GetImageDataFromFieldMemory(pCurrV, width-edge-1, y, width,
						  vop_structure, field_parity);
	for(x=0; x<edge; x++)
	  {
	    PutImageDataToFieldMemory(pCurrV, x, y, width,
				      vop_structure, field_parity,
				      left_value);
	    PutImageDataToFieldMemory(pCurrV, width-edge+x, y, width,
				      vop_structure, field_parity,
				      right_value);
	  }

    }
 
    /*
     *  Vertical padding...
     */
    for(y=0; y<field_edge; y++)
      {
        for(x=0; x<width; x++)
	  {
	    top_value  = GetImageDataFromFieldMemory(pCurrU, x, field_edge, width,
						     vop_structure, field_parity);
	    PutImageDataToFieldMemory(pCurrU, x, y, width,
				      vop_structure, field_parity,
				      top_value);
	    top_value  = GetImageDataFromFieldMemory(pCurrV, x, field_edge, width,
						     vop_structure, field_parity);
	    PutImageDataToFieldMemory(pCurrV, x, y, width,
				      vop_structure, field_parity,
				      top_value);
	  }
      }
    
    for(y=height-field_edge; y<height; y++)
      {
        for(x=0; x<width; x++)
	  {
	    bottom_value  = GetImageDataFromFieldMemory(pCurrU, x, height-field_edge-1, width,
							vop_structure, field_parity);
	    PutImageDataToFieldMemory(pCurrU, x, y, width,
				      vop_structure, field_parity,
				      bottom_value);
	    bottom_value  = GetImageDataFromFieldMemory(pCurrV, x, height-field_edge-1, width,
							vop_structure, field_parity);
	    PutImageDataToFieldMemory(pCurrV, x, y, width,
				      vop_structure, field_parity,
				      bottom_value);
        }
      }
}
/* 01.04.01: Studio Profile */

